# -*- encoding: cp1250 -*-

"""Bikubicka interpolace

Implementovano podle Numerical Recipes in C, napr.
http://www.library.cornell.edu/nr/bookcpdf/c3-6.pdf

#Dalsi zajimavy odkaz:
http://astronomy.swin.edu.au/~pbourke/colour/bicubic/
"""

import cache
import misc_num

from numpy import *

class BicubicError(Exception):
	pass

# Bicubic ###################################################################

class Bicubic(object):
	"""Bikubicka interpolace bodu v obdelniku"""
	_wt_int = (
		( 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0), 
		( 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0), 
		(-3, 0, 0, 3, 0, 0, 0, 0,-2, 0, 0,-1, 0, 0, 0, 0), 
		( 2, 0, 0,-2, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0), 
		( 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0), 
		( 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0), 
		( 0, 0, 0, 0,-3, 0, 0, 3, 0, 0, 0, 0,-2, 0, 0,-1), 
		( 0, 0, 0, 0, 2, 0, 0,-2, 0, 0, 0, 0, 1, 0, 0, 1), 
		(-3, 3, 0, 0,-2,-1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0), 
		( 0, 0, 0, 0, 0, 0, 0, 0,-3, 3, 0, 0,-2,-1, 0, 0), 
		( 9,-9, 9,-9, 6, 3,-3,-6, 6,-6,-3, 3, 4, 2, 1, 2), 
		(-6, 6,-6, 6,-4,-2, 2, 4,-3, 3, 3,-3,-2,-1,-1,-2), 
		( 2,-2, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0), 
		( 0, 0, 0, 0, 0, 0, 0, 0, 2,-2, 0, 0, 1, 1, 0, 0), 
		(-6, 6,-6, 6,-3,-3, 3, 3,-4, 4, 2,-2,-2,-2,-1,-1), 
		( 4,-4, 4,-4, 2, 2,-2,-2, 2,-2,-2, 2, 1, 1, 1, 1), 
		)

	def __init__(self, dtype = 'f'):
		self.dtype = dtype
		self.wt = array(self._wt_int, self.dtype)

	def _bcucof(self, y, y1, y2, y12, d1, d2):
		"""bcucof z numerical recipes
		
		body je pole ctyr hodnot pro body (postupne): vlevo dole, vpravo dole,
		vpravo nahore, vlevo nahore
		Argumenty jsou nasledujici (v poradi od args[0] do args[5])
		y: body, hodnoty funkce
		y1: body, hodnoty derivace po sloupcich
		y2: body, hodnoty derivace po radcich
		y12: body, hodnoty krizove derivace
		d1: delka obdelnika ve smeru po sloupcich
		d2: delka obdelnika ve smeru po radcich
		vraci: numpy.array((4,4)), narozdil od numerical recipes
			neni <1..4>, ale <0..3>

		"""
		d12 = d1*d2
		x = zeros((16,), self.dtype)
		for i in xrange(1, 5):
			i1 = i - 1
			x[i-1] = y[i1]
			x[i+3] = y1[i1]*d1
			x[i+7] = y2[i1]*d2
			x[i+11] = y12[i1]*d12

		cl = dot(self.wt, x)
		return resize(cl, (4,4))

	def interpolate(self, bl, br, tr, tl, l, r, t, b, x, y, compute_gradients = True):
		"""Bikubicka interpolace

		predpokladam souradny system, kde x roste zleva doprava a y zespoda 
		nahoru. 
		tl, tr, bl, br: levy horni, pravy horni, levy spodni, pravy spodni roh
			ctverce ve kterem interpoluji. Kazdy bod je tuple
			(val, xd, yd, cd), kde xd je derivace ve smeru x, yd derivace ve
			smeru y, cd je krizova derivace, val hodnota bodu
		l, r, t, b: souradnice vymezujici obdelnik, left, right, top, bottom
		x, y: Souradnice bodu ve kterem me zajima interpolovana hodnota
		compute_gradients: pokud je False, pocita se pouze hodnota v
			interpolovanem bode, gradienty ne a maji nedefinovanou hodnotu.  
			Pokud nejsou potreba, vypocet se urychli.
		vraci: tuple (hodnota, gradient ve smeru x, gradient ve smeru y)
		"""
		_y  = (bl[0], br[0], tr[0], tl[0])
		_y1 = (bl[1], br[1], tr[1], tl[1])
		_y2 = (bl[2], br[2], tr[2], tl[2])
		_y12= (bl[3], br[3], tr[3], tl[3])
		result = self._recipes_interpolate(_y, _y1, _y2, _y12, l, r, b, t,
				x, y, compute_gradients)
		return result

	def _recipes_interpolate(self, y, y1, y2, y12, x1l, x1u, x2l, x2u, x1, x2, 
	                         compute_gradients = True):
		"""Interpolace presne podle numerical recipes
		
		compute_gradients: pokud je False, pocita se pouze hodnota v
			interpolovanem bode, gradienty ne a maji nedefinovanou hodnotu.  
			Pokud nejsou potreba, vypocet se urychli.
		"""
		if x1l >= x1u or x2l >= x2u:
			raise BicubicError("Not a rectangle")
		d1, d2 = x1u - x1l, x2u - x2l
		c = self._bcucof(y, y1, y2, y12, d1, d2)
		t, u = (x1 - x1l)/d1, (x2 - x2l)/d2
		if compute_gradients:
			ansy = ansy1 = ansy2 = 0
			for i in xrange(3, -1, -1):
				ansy = t*ansy + ((c[i][3]*u + c[i][2])*u + c[i][1])*u + c[i][0]
				ansy2 = t*ansy2 + (3.0*c[i][3]*u + 2.0*c[i][2])*u + c[i][1]
				ansy1 = u*ansy1 + (3.0*c[3][i]*t + 2.0*c[2][i])*t + c[1][i]
			ansy1, ansy2 = ansy1/d1, ansy2/d2
			return (ansy, ansy1, ansy2)
		else:
			ansy = 0
			for i in xrange(3, -1, -1):
				ansy = t*ansy + ((c[i][3]*u + c[i][2])*u + c[i][1])*u + c[i][0]
			return (ansy, None, None)

# BicubicMesh ###############################################################

def interpolate_mesh(mesh, cols, rows, new_cols, new_rows):
	"""Provede bikubickou interpolaci mrizky mesh do nove mrizky

	cols: numpy.array, setrizene pole souradnic sloupcu mrizky mesh
	rows: numpy.array, setrizene pole souradnic radku mrizky mesh
	new_cols: numpy.array, setrizene pole souradnic pozadovanych sloupcu
	new_rows: numpy.array, setrizene pole souradnic pozadovanych radku
	vraci: numpy.array, interpolovanou mrizku

	pri chybe hazi BicubicError
	"""
	b = BicubicMesh(mesh, cols, rows)
	return b.interpolate_mesh(new_cols, new_rows)

class BicubicMesh(Bicubic):
	"""Bikubicka interpolace mrizky"""

	#-- public
	
	def __init__(self, mesh, cols, rows):
		"""konstruktor
		cols: numpy.array, pole souradnic sloupcu
		rows: numpy.array, pole souradnic radku
		mesh: numpy.array, hodnoty mrizky
		"""
		self.dtype = mesh.dtype
		Bicubic.__init__(self, self.dtype)
		self.cols = cols
		self.rows = rows
		self.mesh = mesh
		self._compute_derivations()
		self._bcucof_cache = None #Nastavi se v self.interpolate_mesh

	def interpolate_point(self, x, y):
		""" Nalezne interpolovanou hodnotu bodu v souradnicich x, y 

		Vzdy interpoluje, tj. interpolace se provadi i v bodech mimo puvodni
		mrizku (extrapolaci), ale samozrejme cim dal od mrizky, tim mensi smysl
		interpolovana hodnota ma.. a to tak ze se odvraci kubickou rychlosti!!
		
		x, y: float
		vraci: float
		"""

		srows, scols = self.rows, self.cols
		pt = self.pt

		self._bcucof_cache = cache.Cache(limit = 3*len(srows))

		#indexy ctverce v puvodni mrizce
		l, r, _ = misc_num.vec_find_neighbourgs(self.cols, x)
		b, t, _ = misc_num.vec_find_neighbourgs(self.rows, y)
		lb, rb, rt, lt = pt(l, b), pt(r, b), pt(r, t), pt(l, t)
		lval, rval, tval, bval = scols[l], scols[r], srows[t], srows[b],
				
		return self.interpolate(lb, rb, rt, lt, lval, rval, tval, bval, x, y, False)[0]

	def interpolate_mesh(self, cols, rows, epsilon = 1e-5):
		"""Provede bikubickou interpolaci aktualni mrizky do nove mrizky

		Pokud zjisti, ze nema cenu interpolovat (zdrojove a cilove
		radky/sloupce se shoduji s presnosti lepsi nez epsilon), vrati primo
		kopii puvodni mrizky

		Vzdy interpoluje, tj. interpolace se provadi i v bodech mimo puvodni
		mrizku (extrapolaci), ale samozrejme cim dal od mrizky, tim mensi smysl
		interpolovana hodnota ma.. a to tak ze se odvraci kubickou rychlosti!!

		cols: numpy.array, pozadovane sloupce
		rows: numpy.array, pozadovane radky
		vraci: numpy.array, interpolovanou mrizku
		"""
		if (misc_num.vec_equals(self.rows, rows, epsilon) and
		    misc_num.vec_equals(self.cols, cols, epsilon)):
			return array(self.mesh, self.dtype)
		return self.do_interpolate_mesh(cols, rows)

	#-- private
		
	def _compute_derivations(self):
		"""Pripravi self.xd, self.yd, self.cd"""
		def _y(c, r):
			"""Bod z radku r, sloupce c, zariznuti pri prekroceni mezi
			
			Vraci: (souradnice sloupce, souradnice radku, hodnota v techto souradnicich)
			"""
			rows, cols = self.rows, self.cols
			if r < 0:
				r = 0
			elif r >= len(rows):
				r = len(rows) - 1
			if c < 0:
				c = 0
			elif c >= len(cols):
				c = len(cols) - 1
			return (cols[c], rows[r], self.mesh[r][c])

		n_rows, n_cols = len(self.rows), len(self.cols)
		xd = zeros((n_rows, n_cols), self.dtype)
		yd = zeros((n_rows, n_cols), self.dtype)
		cd = zeros((n_rows, n_cols), self.dtype)
		#Spoctu derivace po radcich, po sloupcich a krizove derivace
		for ridx, rval in enumerate(self.rows):
			for cidx, cval in enumerate(self.cols):
				l, r = _y(cidx - 1, ridx), _y(cidx + 1, ridx)
				xd[ridx][cidx] = (l[2] - r[2]) / (l[0] - r[0])
				t, b = _y(cidx, ridx + 1), _y(cidx, ridx - 1)
				yd[ridx][cidx] = (t[2] - b[2]) / (t[1] - b[1])
				tl, tr = _y(cidx - 1, ridx + 1), _y(cidx + 1, ridx + 1)
				bl, br = _y(cidx - 1, ridx - 1), _y(cidx + 1, ridx - 1)
				cd[ridx][cidx] = (tr[2] - br[2] - tl[2] + bl[2]) /       \
				                 ((r[0] - l[0]) * (t[1] - b[1]))
		self.xd, self.yd, self.cd = xd, yd, cd

	def _bcucof(self, *args):
		"""bcucof z numerical recipes

		tahle verze jen cachuje vysledky Bicubic._bcucof
		"""
		try:
			return self._bcucof_cache[args]
		except KeyError:
			pass

		result = Bicubic._bcucof(self, *args)

		self._bcucof_cache.add(args, result)
		return result

	def pt(self, col, row):
		y = self.mesh[row][col]
		y1 = self.xd[row][col]
		y2 = self.yd[row][col]
		y12 = self.cd[row][col]
		return (y, y1, y2, y12)

	def do_interpolate_mesh(self, cols, rows):
		"""Provede bikubickou interpolaci aktualni mrizky do nove mrizky

		cols: numpy.array, pozadovane sloupce
		rows: numpy.array, pozadovane radky
		vraci: numpy.array, interpolovanou mrizku
		"""

		srows, scols = self.rows, self.cols
		pt = self.pt
		n_rows, n_cols = len(srows), len(scols)
		m = zeros((len(rows), len(cols)), self.dtype) # vysledna mrizka

		self._bcucof_cache = cache.Cache(limit = 3*n_rows)

		#Jo, je to duplicitni a nelibi se mi to.
		l, r, b, t = 0, 1, 0, 1 #indexy ctverce v puvodni mrizce
		lb, rb, rt, lt = pt(l, b), pt(r, b), pt(r, t), pt(l, t)
		lval, rval, tval, bval = scols[l], scols[r], srows[t], srows[b],

		for rowidx, rowval in enumerate(rows):
			points_changed = False
			while t < n_rows - 1 and rowval > srows[t]:
				t += 1
				if t >= 2:
					b += 1
				points_changed = True
			l, r = 0, 1
			for colidx, colval in enumerate(cols):
				while r < n_cols - 1 and colval > scols[r]:
					r += 1
					if r >= 2:
						l += 1
				points_changed = True

				if points_changed:
					lb, rb, rt, lt = pt(l, b), pt(r, b), pt(r, t), pt(l, t)
					lval, rval, tval, bval = scols[l], scols[r], srows[t], srows[b],
				
				m[rowidx][colidx] = self.interpolate(
					lb, rb, rt, lt, lval, rval, tval, bval, 
					colval, rowval, False)[0]

		return m

