#!/usr/bin/python
# -*- encoding: cp1250 -*-

import time
from misc import Timer, Counter
import struct
import registry
from error_logging import *

import serial

import crc16
import debug

from ckdm_interface_common import devices, raise_error, log_event, CKDM_Error
from mvc import ask_yes_no

"""
- Komunikace s ramenem -

Pokud jsem rameno prave zapnul, samo se uvede do lokalniho rezimu. Jedina dalsi
moznost, jak rameno do lokalniho rezimu uvest, je prikazem Remote Disable
(A_RDis). Pokud je rameno v lokalnim rezimu, uzivatel mu muze bez vedomi PC
provadet ruzna zverstva, predevsim prenastavit nulu. Proto se pred kazdym
prikazem, ktery pocita s tim ze nula je nastavena korektne, podivam jestli
nahodou rameno neni v lokalnim rezimu, a pokud ano, pro jistotu nastavim nulu
jeste jednou. Dalsi problemy ktere mohou nastat a netykaji se primo problemu s
komunikaci, jsou ztrata vazby a dojeti na koncove spinace. Obe maji za dusledek
nulovani bitu ARM_BIT_REFERENCED. Pred kazdym prikazem ktery pocita s tim, ze
rameno ma definovanou polohu (tj. bylo referencovano), se tedy podivam zda
skutecne referencovano bylo, a pokud ne, nejprve jej zreferencuji. Po provedeni
prikazu se pro jistotu podivam jeste jednou.


- Komunikace se stolem -

Je o neco obtiznejsi nez s ramenem, protoze u ramene mam implementovane vsechny
potrebne prikazy primo v CKDM, tedy jsem schopen se CKDM primo otazat zda je v
referencovamem stavu atd... u stolu to neni mozne, ale logika je stejna. Zde v
programu tedy pouzivam stejny algoritmus jako pro ovladani ramene a doufam, ze
CKDM implementovala vse potrebne pokud mozno co nejlepe.
"""

# Sercom ####################################################################

class Sercom(object):
	"""Komunikace s CKDM po seriovem portu"""
	TIMEOUT = 1
	DEBUG_SEND = 0
	DEBUG_RECV = 0
	
	CMD_OK = 0
	CMD_TERMINATED = 1
	CMD_TIMEOUT = 2
	CMD_INVALID = 3
	CMD_COMM = 4
	CMD_WAIT = 5
	CMD_LOCAL = 6
	CMD_STOP_PRESSED = 7
	CMD_ILLEGAL_MOVEMENT = 8

	def __init__(self):
		self.ser = serial.Serial()
		self.ser.port = 0
		self.ser.baudrate = 38400
		self.ser.timeout = self.TIMEOUT
		self.last_action = None # 'send' nebo 'recv', pro debug

	@classmethod
	def str_cerr(cls, err):
		try:
			return {
				cls.CMD_OK: 'OK',
				cls.CMD_TERMINATED: 'Terminated',
				cls.CMD_TIMEOUT: 'Timeout',
				cls.CMD_INVALID: 'Invalid',
				cls.CMD_COMM: 'Comm',
				cls.CMD_WAIT: 'Wait',
				cls.CMD_LOCAL: 'Local',
				cls.CMD_ILLEGAL_MOVEMENT: 'Illegal movement'
			}[int(err)]
		except KeyError:
			return u"Neznm chyba(%i)!!" % err

	#-- Low level read/write

	def connect(self, port = 0):
		if not self.ser.isOpen():
			try:
				self.ser.open()
			except serial.SerialException, e:
				raise_error(uni(str(e)), u"Chyba pri otevrn sriovho portu")

		if not self.ser.isOpen():
			raise_error(u"Chyba pri otevrn sriovho portu")


	def disconnect(self):
		if self.ser.isOpen():
			self.ser.close()

	def iflush(self):
		self.ser.flushInput()

	def _dbg(self, action, data):
		dbgdata = []
		for d in data:
			if ord(d) < 32 or ord(d) > 127:
				dbgdata.append('[%02x]'%ord(d))
			else:
				dbgdata.append(d)
		if action != self.last_action:
			print
			print {'send': '>> ', 'recv': '<< '}[action],
			self.last_action = action
		print dbgdata

	def ser_write(self, data):	
		self.ser.write(data)
		if self.DEBUG_SEND:
			self._dbg('send', data)

	def ser_read(self, size):	
		data = self.ser.read(size)
		if self.DEBUG_RECV:
			self._dbg('recv', data)
		return data

	#-- Read/write podle sercom protokolu

	@staticmethod
	def format2cmd(format, *args):
		size = 0
		s_format = ['<']
		for f, a in zip(format, args):
			if   f == 'c':	#char -> uint32
				size += 1
				s_format.append('I')
			elif f == 'i':	#unsigned int -> uint32
				size += 4
				s_format.append('I')
			elif f == 'I':	#signed int -> int32
				size += 4
				s_format.append('i')
			elif f == 's':	#string -> string
				l = len(a) + 1
				size += l
				s_format.append('%is' % l)
		return (struct.pack(''.join(s_format), *args), size)

	def send(self, format, *args):
		# Pred kazdym sendem pro jistotu flushnu vstupni buffer.
		self.iflush()
		cmd, size = self.format2cmd(format, *args)
		#Poslu kanal a delku
		self.ser_write("#")
		self.ser_write(struct.pack('<H', size))
		#Poslu data
		self.ser_write(cmd)
		#Poslu crc
		self.ser_write(struct.pack('<H', crc16.crc16(cmd)))
		#Poslu 0x0a
		self.ser_write(chr(0x0a))

	def recv(self):
		err = u"Chyba pi pjmu odpovdi: "
		b = self.ser_read(3)
		if len(b) != 3:
			raise_error(err + u"Timeout")
		if b[0] != '#':
			raise_error(err + u"Prvn znak zprvy nen #")
		size = struct.unpack('<H', b[1:3])[0]
		msg = self.ser_read(size)
		if len(msg) != size:
			raise_error(err + u"Pijat chyb poet znak (mlo bt %i, bylo %i)"%(size, len(msg)))
		tail = self.ser_read(3)
		if len(tail) != 3:
			raise_error(err + u"Nepilo zakonen zprvy")
		if ord(tail[2]) != 0x0a:
			raise_error(err + u"Posledn znak zprvy nen 0x0A")
		crc = struct.unpack('<H', tail[0:2])[0]
		if crc16.crc16(msg) != crc:
			raise_error(err + u"Chybn kontroln souet(CRC) zprvy")
		return msg

	def format(self, msg, format):
		retval = []
		try:
			for f in format:
				if f in 'ic':
					retval.append(struct.unpack('<I', msg[0:4])[0])
					msg = msg[4:]
				elif f == 'I':
					retval.append(struct.unpack('<i', msg[0:4])[0])
					msg = msg[4:]
				elif f == 's':
					i = msg.index('\x00')
					retval.append(msg[0:i])
					msg = msg[i+1:]
		except (struct.error, ValueError):
			raise_error("Chyba pi zpracovn zprvy")
		return tuple(retval)

# Commander #################################################################

class Commander(object):
	def __init__(self, sercom):
		self.sercom = sercom

	def cmd(self, oformat, iformat, *args):
		"""Posle prikaz a vrati odpoved.
		
		Pokud je iformat jednoznakovy, vraci primo prevedenou hodnotu,
		pokud je viceznakovy, vrati tuple hodnot.

		V teto rutine, i kdyz by to asi melo byt nekde jinde, take osetruji
		stlaceni havarijni stopky.  Pocitam s tim, ze odpoved je budto nejaka
		obecna odpoved ve forme uint32(status), ..., nebo, pri stlaceni
		havarijni stopky, je to pouze status. To znamena ze se nejprve pokusim
		naformatovat zpravu podle oformatu, pokud se to nepovede tak ji
		naformatuji pouze na prijem jednoho uint32. Pokud je status stisk
		stopky, hodim chybu.  Pokud ne, hazim puvodni chybu (pro formatovani
		zadanym iformat-em.)
		"""

		class StopButtonPressedError(Exception):
			pass

		errors = []
		# print '$', oformat, '$', args, '$'
		for tries in xrange(3):
			try:	
				self.sercom.send(oformat, *args)
				msg = self.sercom.recv()
				try:
					result = self.sercom.format(msg, iformat)
				except CKDM_Error, orig_error:
					try:
						result = self.sercom.format(msg, 'i')
					except CKDM_Error:
						raise orig_error
					if result[0] == Sercom.CMD_STOP_PRESSED:
						raise StopButtonPressedError
					raise orig_error

				if len(result) == 1:
					return result[0]
				return result
			except CKDM_Error, e:
				errors.append(e.msg)
				time.sleep(1)
			except StopButtonPressedError:
				raise_error(u"Stisknuta havarijni stopka. Vyete nastalou situaci a vypnte a zapnte panel.")

		errors.append(u"Pkaz: %s" % Sercom.format2cmd(oformat, *args)[0])
		errors.append(u"Chyba pi komunikaci s CKDM")
		raise_error(*tuple(errors))

	def check_cerr(self, cerr, err_msg):
		if cerr == Sercom.CMD_OK:
			return
		raise_error(u"Chyba: %s, kd chyby:%s" % (err_msg, Sercom.str_cerr(cerr)))

	def checked_cmd(self, err_msg, oformat, iformat, *args):
		"""Provede prikaz pro nejaky pristroj a osetri pripadnou chybu.

		iformat: Vzdy se implicitne pripojuje na zacatek 'i', tedy
		status. Pokud je iformat None, nevraci se nic, pokud je
		jednopismenny, vraci se primo hodnota, pokud je vicepismenny,
		vraci se seznam hodnot.

		oformat: Vzdy na zacatek pripoji 's' ktere ma formatovat jmeno
		vystupniho prikazu 
		"""
		if oformat:
			oformat = 's'+oformat
		else:
			oformat = 's'

		if iformat:
			retval = self.cmd(oformat, 'i'+iformat, *args)
			self.check_cerr(retval[0], err_msg)
			if len(iformat) == 1:
				return retval[1]
			else:
				return retval[1:]
		else:
			retval = self.cmd(oformat, 'i', *args)
			self.check_cerr(retval, err_msg)

# CKDM_Interface ############################################################

class CKDM_Interface(object):
	def __init__(self):
		self.commander = Commander(Sercom())

	def connect(self, port = None):
		"""Pripoji se k seriovemu portu. port je 'COM1 .. 'COM4', pokud je
		None, zjisti se z registry.settings. """ 
		if port == None:
			port = registry.settings.value.ckdm_com.value
		try:
			port_idx = {'COM1': 0, 'COM2': 1, 'COM3': 2, 'COM4': 3}[port]
		except KeyError:
			raise_error(u"Chyba pi otevrn sriovho portu: Neznm port %s" % str(port))
		log_event('notice', 'connect to %s' % port)
		self.commander.sercom.connect(port_idx)

	def disconnect(self):
		log_event('notice', 'disconnect')
		self.commander.sercom.disconnect()

	def measure_prepare(self):
		self.table_get_pos()
		self.table_get_zero()
	
	#-- Konzole

	def ckdm_cmd(self, oformat, iformat, *args):
		return self.commander.checked_cmd(u"Chyba pi provdn pkazu na CKDM: ", oformat, iformat, *args)

	def ckdm_id(self):
		log_event('notice', 'ckdm_id')
		devices.ckdm_id.value = self.ckdm_cmd(None, 's', 'C_ID')
	
	def ckdm_gpib_send(self, addr, data):
		log_event('notice', 'ckdm_gpib_send: addr %i, data %s' % (addr, data))
		self.ckdm_cmd('is', None, 'C_Send', addr, data)
	
	def ckdm_gpib_recv(self, addr):
		log_event('notice', 'ckdm_gpib_recv from %i' % addr)
		return self.ckdm_cmd('i', 's', 'C_Recv')
	
	def ckdm_gpib_ifc(self):
		log_event('notice', 'ckdm_gpib_ifc')
		return self.ckdm_cmd(None, None, 'C_IFC')
	
	#-- Stul
	
	TABLE_BIT_ALIVE = (1 << 0)
	TABLE_BIT_REFERENCED = (1 << 1)
	TABLE_BIT_BUSY = (1 << 2)
	TABLE_BIT_LOCAL = (1 << 3)

	TABLE_ERR_OK = 0
	TABLE_ERR_SYNTAX = 1
	TABLE_ERR_BADPARAM = 2
	TABLE_ERR_END = 3
	TABLE_ERR_RANGE = 4
	TABLE_ERR_DIST = 5
	TABLE_ERR_POWER = 6
	TABLE_ERR_MOTORS = 7
	TABLE_ERR_POS = 8
	TABLE_ERR_MACRO = 9
	TABLE_ERR_STOP = 10

	TABLE_ERR_RECV = 50
	TABLE_ERR_SEND = 51

	TABLE_ERR_CMOS = 90
	TABLE_ERR_RAM = 91
	TABLE_ERR_ROM = 92
	TABLE_ERR_201 = 93
	TABLE_ERR_301 = 94
	TABLE_ERR_221 = 95
	TABLE_ERR_311 = 96

	@classmethod
	def table_strerr(cls, err):
		try:
			return {
				cls.TABLE_ERR_OK: u"dn chyba",
				cls.TABLE_ERR_SYNTAX: u"Neznm nebo chybn pkaz",
				cls.TABLE_ERR_BADPARAM: u"patn parametr pkazu",
				cls.TABLE_ERR_END: u"patn ukonen bloku dat",
				cls.TABLE_ERR_RANGE: u"Zadan data mimo rozsah",
				cls.TABLE_ERR_DIST: u"Pli velk drha",
				cls.TABLE_ERR_POWER:u"Porucha napjen koncovch stup", 
				cls.TABLE_ERR_MOTORS: u"Porucha vazby krokovch motor",
				cls.TABLE_ERR_POS: u"Najet na koncov spna",
				cls.TABLE_ERR_MACRO: u"Makro nen povoleno",
				cls.TABLE_ERR_STOP: u"Zastaveno tlatkem Stop",

				cls.TABLE_ERR_RECV: u"Chyba pi pjmu pes sbrnici GPIB",
				cls.TABLE_ERR_SEND: u"Chyba pi odesln pes sbrnici GPIB",

				cls.TABLE_ERR_CMOS: u"Porucha pamti CMOS",
				cls.TABLE_ERR_RAM: u"Porucha pamti RAM(RWM)",
				cls.TABLE_ERR_ROM: u"Porucha pamti ROM(EPROM)",
				cls.TABLE_ERR_201: u"Porucha jednotky MS80-201",
				cls.TABLE_ERR_301: u"Porucha jednotky MS80-301",
				cls.TABLE_ERR_221: u"Porucha jednotky MS80-221",
				cls.TABLE_ERR_311: u"Porucha jednotky MS80-311",
			}[err]
		except KeyError:
			return u"Neznm chyba stolu: %s" % str(err)

	def table_cmd(self, oformat, iformat, *args):
		errors = []
		for i in xrange(3):
			try:
				return self.commander.checked_cmd(u"pi komunikaci se stolem", oformat, iformat, *args)
			except CKDM_Error, e:
				errors.append(e.msg)
		raise_error(*tuple(errors))

	def table_check_last_error(self):
		"""Precte posledni chybu stolu, pokud nejaka nastala tak hodi vyjimku."""
		last_error = self.table_cmd('s', 'i', 'T_LastErr')
		if last_error != self.TABLE_ERR_OK:
			raise_error(self.table_strerr(last_error), u"Chyba pi provdn pkazu stolu.")

	def table_busy_wait(self, timeout = None):
		"""Ceka dokud je rameno busy (tj. provadi presun). Vraci flags."""
		t = Timer(timeout)
		while not t.expired():
			t.sleep(0.3)
			flags = self.table_get_flags()
			if not (flags & self.TABLE_BIT_BUSY):
				return flags
		raise_error(u"Timeout pi ekn na stl")

	def exec_and_check_lasterr(self, f, log_entry, *args, **kwargs):
		log_event('notice', log_entry)
		errors = []

		for i in xrange(3):
			retval = f(*args, **kwargs)
			last_error = self.table_cmd('s', 'i', 'T_LastErr')
			if last_error == self.TABLE_ERR_OK:
				for e in errors:
					log_event('error', e)
				return retval
			errors.append(self.table_strerr(last_error))

		errors.append(u"Chyba pi provdn pkazu stolu %s." % log_entry)
		raise_error(*errors)

	def do_table_id(self):
		devices.table_id.value = self.table_cmd(None, 's', 'T_ID')
		last_error = self.table_cmd('s', 'i', 'T_LastErr')
	def table_id(self):
		self.exec_and_check_lasterr(self.do_table_id, 'table_id')

	def table_get_flags(self):
		log_event('notice', 'table_get_flags')
		flags = self.table_cmd(None, 'i', 'T_Flags') 
		return flags

	def do_table_manual_enable(self):
		self.table_cmd(None, None, 'T_RDis')
	def table_manual_enable(self):
		self.exec_and_check_lasterr(self.do_table_manual_enable, 'table_manual_enable')
		# exec_and_check_lasterr posle zas nejake povely stolu, coz by
		# zrusilo lok. ovladani - tak ho nakonec dam jeste rucne bez
		# testu.
		self.do_table_manual_enable()
	  
	def do_table_manual_disable(self): 
		devices.table_id.value = self.table_cmd(None, 's', 'T_ID') # treba...
	def table_manual_disable(self): 
		self.exec_and_check_lasterr(self.do_table_manual_disable, 'table_manual_disable')

	def do_table_goto(self, rot, tilt):
		self.table_cmd('II', None, 'T_Goto', int(rot*1000), int(tilt*1000))
		flags = self.table_busy_wait(30)
		if not (flags & self.TABLE_BIT_REFERENCED):
			raise_error(u"Pesun stolu se nezdail, stl vyjel z reference")
	def table_goto(self, rot, tilt):
		# stul ma takovou specialitku, ze nechce tocit vic nez 270 deg (hlasi
		# 05 Error). Co se da delat jineho nez rozdelit pohyb na dva a
		# nadavat..
		self.table_reference()
		current_rot = devices.rot.value
		if abs(current_rot - rot) > 180:
			interm_rot = (rot+current_rot)/2
			self.exec_and_check_lasterr(self.do_table_goto, 'table_goto %f, %f'%(interm_rot, tilt), interm_rot, tilt)
		self.exec_and_check_lasterr(self.do_table_goto, 'table_goto %f, %f'%(rot, tilt), rot, tilt)
		self.table_get_pos()

	def do_table_reference(self):
		flags = self.table_busy_wait()
		if (flags & self.TABLE_BIT_REFERENCED):
			return
		if not ask_yes_no(u"Referencovn stolu", 
				u"Poloha stolu nen definovan, je poteba jej zreferencovat. " + 
				u"Mm provst automatick referencovn stolu?"):
			raise_error(u"Stl nebyl podle volby uivatele zreferencovn.")
		self.table_cmd(None, None, 'T_Reference')
		flags = self.table_busy_wait(60)
		if not (flags & self.TABLE_BIT_REFERENCED):
			raise_error(u"Referencovn stolu se nezdailo.")
		return True
	def table_reference(self):
		"""Provede referenci, pouze pokud je to nutne."""
		if self.exec_and_check_lasterr(self.do_table_reference, 'table_reference'):
			cal = registry.calibration_table.value
			self.table_set_zero(cal.rot_zero.value, cal.tilt_zero.value)
			self.table_get_pos()
			self.table_check_last_error()

	def do_table_get_pos(self):
		r, t = self.table_cmd(None, 'II', 'T_Pos')
		devices.rot.value = r/1000.0
		devices.tilt.value = t/1000.0
	def table_get_pos(self):
		self.exec_and_check_lasterr(self.do_table_get_pos, 'table_get_pos')

	def do_table_get_zero(self):
		r, t = self.table_cmd(None, 'II', 'T_GetZero')
		devices.rot_zero.value = r/1000.0
		devices.tilt_zero.value = t/1000.0
	def table_get_zero(self):
		self.exec_and_check_lasterr(self.do_table_get_zero, 'table_get_zero')

	def do_table_set_zero(self, rot, tilt):
		log_event('notice', 'table_set_zero %f, %f' % (rot, tilt))
		self.table_cmd('II', None, 'T_SetZero', int(rot*1000), int(tilt*1000))
		self.table_get_zero()
	def table_set_zero(self, rot, tilt):
		self.exec_and_check_lasterr(self.do_table_set_zero, 'table_set_zero', rot, tilt)

	def do_table_invert(self, invert):
		log_event('notice', 'table_invert %s' % str(invert))
		if not invert:
			invert = 0
		else:
			invert = 1
		self.table_cmd('i', None, 'T_Invert', invert)
	def table_invert(self, invert):
		self.exec_and_check_lasterr(self.do_table_invert, 'table_invert', invert)

	#-- Rameno

	ARM_BIT_ALIVE = (1 << 0)
	ARM_BIT_REFERENCED = (1 << 1)
	ARM_BIT_BUSY = (1 << 2)
	ARM_BIT_LOCAL = (1 << 3)
	ARM_BIT_RANGE = (1 << 10)
	ARM_BIT_RANGE_POSITIVE = (1 << 11)
	ARM_BIT_RANGE_NEGATIVE = (1 << 12)
	ARM_BIT_BLOCKED = (1 << 13)

	def arm_cmd(self, oformat, iformat, *args):
		return self.commander.checked_cmd(u"Chyba pi komunikaci s ramenem: ", oformat, iformat, *args)

	def arm_busy_wait(self, timeout = None):
		"""Ceka dokud je rameno busy (tj. provadi presun). Vraci flags."""
		t = Timer(timeout)
		while not t.expired():
			flags = self.arm_get_flags()
			if not (flags & self.ARM_BIT_BUSY):
				return flags
			t.sleep(0.3)
		raise_error(u"Timeout pi ekn na rameno")

	def arm_id(self):
		log_event('notice', 'arm_id')
		devices.arm_id.value = self.arm_cmd(None, 's', 'A_ID')
		devices.arm_manual_enable.value = False

	def arm_get_flags(self):
		log_event('notice', 'arm_get_flags')
		devices.arm_manual_enable.value = False
		return self.arm_cmd(None, 'i', 'A_Flags') 

	def arm_manual_enable(self):
		log_event('notice', 'arm_manual_enable')
		self.arm_cmd(None, None, 'A_RDis')

		devices.arm_manual_enable.value = True
		# arm_get_flags manualni ovladani opet zrusi :-)
		# devices.arm_manual_enable.value = bool(self.arm_get_flags() & self.ARM_BIT_LOCAL)
		# if not devices.arm_manual_enable.value:
		#	raise_error("Nepodailo se uvst rameno do reimu runho ovldn.")

	def arm_manual_disable(self):
		log_event('notice', 'arm_manual_disable')
		devices.arm_manual_enable.value = False
		self.arm_cmd(None, 'i', 'A_Flags') 

	def arm_goto(self, pos):
		log_event('notice', 'arm_goto %f' % pos)
		flags = self.arm_busy_wait()
		self.arm_reference()
		if (flags & self.ARM_BIT_LOCAL):
			self.arm_set_zero(registry.calibration_arm.value.arm_zero.value)
		self.arm_cmd('I', None, 'A_Goto', int(pos*1000))
		flags = self.arm_busy_wait()
		if not (flags & self.ARM_BIT_REFERENCED):
			raise_error(u"Pesun ramene se nezdail.")
		devices.arm.value = pos
		devices.arm_manual_enable.value = False

	def arm_reference(self):
		"""Provede referenci, pouze pokud je to nutne."""
		log_event('notice', 'arm_reference')
		flags = self.arm_busy_wait()
		if (flags & self.ARM_BIT_REFERENCED):
			return
		if not ask_yes_no(u"Referencovn ramene", 
				u"Poloha ramene nen definovan, je poteba jej zreferencovat. " + 
				u"Mm provst automatick referencovn ramene?"):
			raise_error(u"Rameno nebylo podle volby uivatele zreferencovno.")
		self.arm_cmd(None, None, 'A_Reference')
		flags = self.arm_busy_wait(60)
		if not (flags & self.ARM_BIT_REFERENCED):
			raise_error(u"Referencovn ramene se nezdailo.")
		devices.arm.value = 0.0
		devices.arm_manual_enable.value = False

	def arm_set_zero(self, zero):
		log_event('notice', 'arm_set_zero %f' % zero)
		self.arm_cmd('I', None, 'A_SetZero', int(zero*1000))
		self.arm_get_zero()
		devices.arm_manual_enable.value = False

	def arm_get_zero(self):
		log_event('notice', 'arm_get_zero')
		devices.arm_zero.value = self.arm_cmd(None, 'I' , 'A_GetZero')/1000.0
		devices.arm_manual_enable.value = False

	def arm_get_pos(self):
		"""Tady by bylo sakra dobry nejak overovat, ze uz je referencovano a
		nastavena nula. TODO i u table_get_pos."""

		log_event('notice', 'arm_get_pos')
		devices.arm.value = self.arm_cmd(None, 'I', 'A_Pos')/1000.0
		devices.arm_manual_enable.value = False

	#-- Luxmetr

	LUX_BIT_ALIVE = (1 << 0)
	LUX_BIT_MEASURING = (1 << 1)
	LUX_BIT_MEASURED = (1 << 2)

	def lux_cmd(self, oformat, iformat, *args):
		return self.commander.checked_cmd(u"Chyba pi komunikaci s luxmetrem: ", oformat, iformat, *args)

	def lux_init(self):
		log_event('notice', 'lux_init')
		self.lux_cmd(None, None, 'L_Reset')
		devices.luxmeter_manual_enable.value = False
	
	def lux_busy_wait(self, timeout = None):
		"""Ceka dokud je rameno busy (tj. provadi presun). Vraci flags."""
		t = Timer(timeout)
		while not t.expired():
			flags = self.lux_get_flags()
			if not (flags & self.LUX_BIT_MEASURING):
				return flags
			t.sleep(0.3)
		raise_error(u"Timeout pi ekn na luxmetr")

	def lux_manual_enable(self):
		log_event('notice', 'lux_manual_enable')
		self.lux_cmd(None, None, 'L_RDis')
		devices.luxmeter_manual_enable.value = True
	  
	def lux_manual_disable(self):
		log_event('notice', 'lux_manual_disable')
		self.lux_cmd(None, None, 'L_Flags')
		devices.luxmeter_manual_enable.value = False
	  
	def lux_get_flags(self):
		log_event('notice', 'lux_get_flags')
		devices.luxmeter_manual_enable.value = False
		return self.lux_cmd(None, 'i', 'L_Flags') 

	def lux_do_get_val(self):
		self.lux_cmd(None, None, 'L_Start')
		self.lux_busy_wait()
		val = self.lux_cmd(None, 'I', 'L_Get')
		devices.luxmeter_manual_enable.value = False
		if val < 0:
			try:
				msg = {
					-1: u"Chyba komunikace s luxmetrem",
					-2: u"Odmr peruen",
					-3: u"Luxmetr pesvcen",
					-4: u"Osvtlenost pli kols",
					}[val]
			except KeyError:
				msg = u"Neznm chyba luxmetru (%i), kontaktujte vrobce software." % val
			raise_error(msg)
		devices.luxmeter.value = val/1000.0 # Luxmetr vraci hodnotu v mililuxech

	def lux_get_val(self):
		# Pokud se to poprve nezdari, resetuju luxmetr a zkusim jeste jednou.
		log_event('notice', 'lux_get_val')
		try:
			self.lux_do_get_val()
		except CKDM_Error:
			self.lux_init()
			self.lux_do_get_val()

def _test():
	i = CKDM_Interface()
	i.connect()
	print i.ckdm_id()
	i.ckdm_remote_enable()
	print i.lux_get_val()

if __name__ == '__main__':
	_test()
