# -*- encoding: cp1250 -*-

"""Modul pro vyhodnocovani merenych dat. 

Je obecny v tom smyslu, ze nevyhodnocuje konkretni typove zkousky, ma pouze
algoritmy, ktere transformuji nejaka merena data spolu s informaci o tom, co se
v nich ma vyhodnotit, do nejakeho vysledku vyhodnoceni (instance EvalResult).
Konkretni aplikace, tj vyhodnoceni typovych zkousek nebo vyhodnocovani pri
vyvoji potom pouze pouziva tyto algoritmy pro vlastni ucely.

Trida Eval je bezestavova, obsahuje pouze tridni metody.
"""

import wx

from mvc import *
import plot
import storage
import typetests
from misc import any, ternary, pairwise, SubclassResponsibilityError
import misc_num
from meas_corrections import MeasCorrector, MeasCorrectorError
from numpy import sin, cos
from error_logging import *

import debug

epsilon = storage.epsilon

_defined_colors =  [((0xff-x) << 8) | (x << 16) for x in (0xff, 0xbb, 0x88, 0x44, 0x00)]
_measured_colors = [((0xff-x) << 0) | (x << 16) for x in (0xff, 0xbb, 0x88, 0x44, 0x00)]

def _get_color_from_scheme(scheme, i):
	col = Color()
	try:
		hexcode = scheme[i]
	except KeyError:
		hexcode = scheme[-1]
	col.set_hex(hexcode)
	return col

def color_measured(i):
	return _get_color_from_scheme(_measured_colors, i)

def set_color_measured(col, i):
	col.copy_from(color_measured(i))

def color_defined(i):
	return _get_color_from_scheme(_defined_colors, i)

def set_color_defined(col, i):
	col.copy_from(color_defined(i))

# pomocna HTMLizace

def html_os(omega, sigma):
	html_omega = u'<font face="Symbol">w</font>'
	html_sigma = u'<font face="Symbol">s</font>'
	return u'%s = %.3f&deg;, %s = %.3f&deg;' % (html_omega, omega, html_sigma, sigma)

def html_osi(pt):
	if pt == None:
		return u"<font color=#ff1010>Bod nenalezen</font>"
	return html_os(pt.omega, pt.sigma) + u', I = %.0f cd' % pt.intensity

def html_compliant(compliant):
	"""Vraci html VYHOVUJE nebo NEVYHOVUJE"""
	if compliant:
		return u"<b><font color=#008000>VYHOVUJE</font></b>"
	return u"<b><font color=#800000>NEVYHOVUJE</font></b>"

# pomocna Plotizace

def plot_measpt(plot_model, pt, text, color = None, movegroup = ''):
	plot_pt = plot.PlotPoint()
	plot_pt.pos.set_xy(pt.omega, pt.sigma)
	plot_pt.color.set(color)
	plot_pt.movegroup.value = '%s_pts' % movegroup
	plot_model.plottables.append(plot_pt)

	plot_text = plot.PlotText()
	plot_text.text.value = text
	plot_text.movable.value = True
	plot_pt.set_text_props(plot_text)
	plot_text.color.set(color)
	plot_text.bound(plot_model.extent.copy().expand_ratio(-0.05, -0.05))
	plot_text.movegroup.value = '%s_texts' % movegroup
	plot_model.plottables.append(plot_text)

def plot_iso_text(plot_model, text, lines, position, color = None, movegroup = ''):
	"""Pokud jsou zobrazeny nejake cary, hodi do grafu popisek
	
	position: 'top' nebo 'bottom' (viz plot.PlotLines.set_text_props())
	"""
	if lines.is_empty():
		return
	plot_text = plot.PlotText(text)
	plot_text.color.set(color)
	plot_text.movable.value = True
	plot_text.movegroup.value = u'%s_texts' % movegroup
	lines.set_text_props(plot_text, position)
	plot_text.bound(plot_model.extent.copy().expand_ratio(-0.05, -0.05))
	plot_model.plottables.append(plot_text)

# dalsi pomocne veci ########################################################

def str_omega_sigma(omega, sigma):
	return u"omega: %.3f, sigma: %.3f" % (omega, sigma)

def raise_point_not_found(omega, sigma):
	log_error(u"V namench datech se nepodailo nalzt bod o souadnicch:\n" + str_omega_sigma(omega, sigma))
	raise EvalError

# WhiteRedMesh ##############################################################

class WhiteRedMesh(object):
	"""Pomocna trida vytvarejici bilou/cervenou mrizku, cachuje vysledky vytvareni."""
	orig_mesh = storage.Mesh()
	wrmesh = storage.Mesh()
	last_wr_coef = None

	@classmethod
	def get_wrmesh(cls, mesh, white, red):
		""" Vytvori pomocnou mrizku pro pocitani sest. izokandel
		
		U sestupovych navestidel neni situace uplne jednoducha. Kazda
		izokandela je vlastne rozdelena na dve pulky, nad sigma=0 a pod. To
		odpovida atributum white_intensity(nahore) a red_intensity(dole). Toto
		se projevi pouze pri vyhodnocovani kontur merenych izokandel, ja
		situaci resim tak ze si vygeneruji pomocnou mrizku, ktera ma v casti
		sigma > 0 nezmenene hodnoty a v casti sigma < 0 hodnoty vynasobene
		koeficientem white_intensity/red_intensity. Kvuli zubum, ktere by
		vznikaly v sigma = 0, radek sigma = 0 pocitam jako prumer nejblizsi
		vyssi a nejblizsi nizsi hodnoty sigma. Tato funkce udela hlubokou kopii
		mesh a provede s ni popsane zvrhlosti. Vraci novou pomocnou mesh.

		Mrizka je vytvorena z mrizky mesh, pokud ma byt svitivost v bile casti
		white a svitivost v cervene casti red.

		Uklada si referenci na mesh a nekontroluje, zda se mezitim mesh
		nezmenila! To znamena, ze pokud se mezitim zmeni mesh, bude tato metoda
		vracet neplatny vysledek.
		"""
		try:
			wr_coef = white/red
		except ZeroDivisionError:
			log_error(u"erven intenzita je nulov, nelze vytvoit pomocnou bloervenou mku.")
			raise EvalError
		if cls.orig_mesh != mesh or cls.last_wr_coef == None or abs(wr_coef - cls.last_wr_coef) > 0.01:
			cls.orig_mesh = mesh
			cls.wrmesh.copy_from(cls.orig_mesh)
			cls.wrmesh.multiply_bottom(wr_coef)
			cls.last_wr_coef = wr_coef
		return cls.wrmesh

#############################################################################
# Predci mereni a vysledku ##################################################
#############################################################################

# EvalError #################################################################

class EvalError(Exception):
	"""Vyjimka hazena pri chybe vyhodnocovani """
	pass

# Eval ######################################################################

class Eval(object):
	"""Vyhodnocovac. Nema stav, obsahuje pouze tridni metody."""
	meas_corrector_settings = {'apply_simfilter': False, 'apply_calibration_matrix': False}

# EvalResult ################################################################

class EvalResult(CompositeModel):
	"""Vysledek vyhodnoceni"""
	_models = {
		# Kdy bylo vyhodnoceno
		'eval_timestamp': TimeModel(),
		# Korekce mereni (simfilter, kalibracni matice pripadne dalsi)
		'meas_corrector': MeasCorrector(),
		# Vysledek vyhodnoceni (vyhovuje/nevyhovuje)
		'verdict': BoolModel(),
	}

	def just_evaluated(self):
		"""Updatne verdict a timestamp"""
		self.verdict.value = self.check_verdict()
		self.eval_timestamp.set_now()

	def html_report_meas_corrections(self):
		"""HTML report korekci mereni"""
		return self.meas_corrector.html_report()

	def clear(self):
		"""Ma za ukol vycistit mereni - smazat body, cary nebo cokoliv."""
		pass

#############################################################################
#############################################################################
# Smerove a sestupove typove zkousky ########################################
#############################################################################
#############################################################################

#############################################################################
# Vyhodnocene definovane izokandely #########################################
#############################################################################

# EvaluatedDefIsocandel #####################################################

class EvaluatedDefIsocandel(CompositeModel):
	"""Predek vyhodnocenych definovanych izokandel."""
	_models = {
		 # dodrzena podminka min. svitivosti. Zaroven implikuje, zda je
		 # definovana izokandela uvnitr namerene.
		'min_intensity_satisfied': BoolModel(False),
		'min': storage.MeasuredPoint(), # min. a max. intenzita na def. izokandele
		'max': storage.MeasuredPoint(),
		'abbr': StrModel(),
		'descr': StrModel(),
		'group': StrModel(), # kvuli vyvoji
		'color': Color(), # taky kvuli vyvoji, chtelo by to rozumnejc rozdelit..
		# 'isocandel': typetests.IsocandelOval | typetests.IsocandelRectangle | ... v potomcich
	}

	def get_description(self):
		"""Ma vratit str, informace o izokandele(pro statusbar, tooltip)""" 
		raise SubclassResponsibilityError

	#-- vyhodnocovani

	def eval_extremes(self, mesh):
		"""nalezne extremy (min, max) definovane izokandely"""
		minmax = self.isocandel.find_contour_minmax(mesh)
		if minmax == None:
			log_error(u"Nepodailo se nalzt kritick body izokandely.")
			raise EvalError
		self.min.copy_from(minmax[0])
		self.max.copy_from(minmax[1])

	def eval_min_intensity_satisfied(self, mesh, intensity):
		inner_points = self.isocandel.find_inner_points(mesh)
		if len(inner_points):
			min_intensity = min(p.intensity for p in inner_points)
			self.min_intensity_satisfied.value = bool(min_intensity >= intensity)
		else:
			self.min_intensity_satisfied.value = False

	def eval(self, mesh, typetest_isocandel, abbr, descr):
		"""Vyhodnoti se podle izokandely typove zkousky."""
		raise SubclassResponsibilityError

	def eval_meas(self, mesh):
		"""Vyhodnoti merenou izokandelu.

		Predpoklada, ze v self.isocandel jiz je spravne vyplnena typova
		zkouska. Vytvori prislusnou novou vyhodnocenou merenou izokandelu
		(EvaluatedMeasIsocandel*), a vyhodnoti ji na mrizce mesh.
		"""
		raise SubclassResponsibilityError

	#-- Pridavani do grafu

	def add_labels_to_plot(self, plot_model, lines, color = None, movegroup = ''):
		raise SubclassResponsibilityError

	def add_lines_to_plot(self, plot_model, color = None, movegroup = ''):
		abbr = self.abbr.value
		i = self.isocandel
		lines = plot.PlotLines()
		lines.color.set(color)
		lines.width.value = 1
		lines.lines.copy_from(i.get_lines())
		lines.description.value = self.get_description()
		lines.movegroup.value = u'%s_lines' % movegroup
		plot_model.plottables.append(lines)
		return lines

	def add_minmax_to_plot(self, plot_model, color = None, movegroup = ''):
		plot_measpt(plot_model, self.min, u'Min %s:%.0f' % (self.abbr, self.min.intensity), color, movegroup)
		plot_measpt(plot_model, self.max, u'Max %s:%.0f' % (self.abbr, self.max.intensity), color, movegroup)

	def add_to_plot(self, plot_model, color = None, movegroup = ''):
		""" Do plot_model vykresli definovanou izokandelu.
		
		Vykresli cary, minimum, maximum, popisky
		"""
		lines = self.add_lines_to_plot(plot_model, color, movegroup)
		plot_model.freeze_changes()
		try:
			self.add_labels_to_plot(plot_model, lines, color, movegroup)
			self.add_minmax_to_plot(plot_model, color, movegroup)
		finally:
			plot_model.thaw_changes()

# EvaluatedDefIsocandelDirectional ###########################################

class EvaluatedDefIsocandelDirectional(EvaluatedDefIsocandel):
	"""Predek vyhodnocenych definovanych izokandel pro smerova navestidla"""

	def get_description(self):
		"""Ma vratit str, informace o izokandele(pro statusbar, tooltip)""" 
		return u"Definovan %s (I = %.0f)" % (self.descr.value, self.isocandel.intensity.value)

	def html_report(self):
		return u"""
			<b>%s: </b>%.0f cd<br>
			<b>Kritick body na poadovan izokandele: </b><br>
			<b>Minimum: </b>%s<br>
			<b>Maximum: </b>%s<br>
			<b>%s: </b>%s<br>
		""" % (
				self.descr.value,
				self.intensity.value,
				html_osi(self.min),
				html_osi(self.max),
				self.descr.value, html_compliant(self.min_intensity_satisfied.value)
				)
	#-- vyhodnocovani

	def _get_intensity(self):
		return self.isocandel.intensity
	intensity = property(_get_intensity)

	def eval(self, mesh, typetest_isocandel, abbr, descr):
		self.abbr.value = abbr
		self.descr.value = descr
		self.isocandel.copy_from(typetest_isocandel)
		self.eval_extremes(mesh)
		self.eval_min_intensity_satisfied(mesh, self.intensity.value)

	def eval_meas(self, mesh):
		meas_isocandel = EvalIso.get_meas_isocandel(self.isocandel)()
		meas_isocandel.eval(mesh, self.intensity.value, self.abbr.value)
		return meas_isocandel

	#-- pridavani do grafu

	def add_labels_to_plot(self, plot_model, lines, color = None, movegroup = ''):
		plot_iso_text(plot_model, u'%s:%.0f' % (self.abbr, self.intensity.value), 
		              lines, 'top', color, movegroup = 'def')

# EvaluatedDefIsocandelApproach #############################################

class EvaluatedDefIsocandelApproach(EvaluatedDefIsocandel):
	"""Predek vyhodnocenych definovanych izokandel pro sestupova navestidla"""

	def get_description(self):
		"""Ma vratit str, informace o izokandele(pro statusbar, tooltip)""" 
		i = self.isocandel
		return u"Definovan %s (B/: %.0f/%.0f)" % \
				(self.descr.value, i.white_intensity.value, i.red_intensity.value)

	def _get_white_intensity(self):
		return self.isocandel.white_intensity
	white_intensity = property(_get_white_intensity)

	def _get_red_intensity(self):
		return self.isocandel.red_intensity
	red_intensity = property(_get_red_intensity)

	def html_report(self):
		return u"""
			<b>%s: </b>bl %.0f cd, erven %.0f cd<br>
			<b>Kritick body na poadovan izokandele: </b><br>
			<b>Minimum: </b>%s<br>
			<b>Maximum: </b>%s<br>
			<b>%s: </b>%s<br>
		""" % (
				self.descr.value,
				self.white_intensity.value, self.red_intensity.value,
				html_osi(self.min),
				html_osi(self.max),
				self.descr.value, html_compliant(self.min_intensity_satisfied.value)
				)

	#-- vyhodnocovani

	def eval(self, mesh, typetest_isocandel, abbr, descr):
		self.abbr.value = abbr
		self.descr.value = descr
		self.isocandel.copy_from(typetest_isocandel)
		self.eval_extremes(mesh)
		wrmesh = WhiteRedMesh.get_wrmesh(mesh, self.white_intensity.value, self.red_intensity.value)
		self.eval_min_intensity_satisfied(wrmesh, self.white_intensity.value)

	def eval_meas(self, mesh):
		meas_isocandel = EvalIso.get_meas_isocandel(self.isocandel)()
		i = self.isocandel
		meas_isocandel.eval(mesh, i.white_intensity.value, i.red_intensity.value, self.abbr.value)
		return meas_isocandel

	#-- pridavani do grafu

	def add_labels_to_plot(self, plot_model, lines, color = None, movegroup = ''):
		i = self.isocandel
		plot_iso_text(plot_model, u'%s:%.0f' % (self.abbr, i.white_intensity.value), lines, 'top', color, movegroup)
		plot_iso_text(plot_model, u'%s:%.0f' % (self.abbr, i.red_intensity.value), lines, 'bottom', color, movegroup)

# EvaluatedDefIsocandelRectangle ############################################

class EvaluatedDefIsocandelRectangle(EvaluatedDefIsocandelDirectional):
	_models = {'isocandel': typetests.IsocandelRectangle()}
	_models.update(EvaluatedDefIsocandelDirectional._models)

# EvaluatedDefIsocandelEllipse ##############################################

class EvaluatedDefIsocandelEllipse(EvaluatedDefIsocandelDirectional):
	_models = {'isocandel': typetests.IsocandelEllipse()}
	_models.update(EvaluatedDefIsocandelDirectional._models)

# EvaluatedDefIsocandelOval #################################################

class EvaluatedDefIsocandelOval(EvaluatedDefIsocandelApproach):
	_models = {'isocandel': typetests.IsocandelOval()}
	_models.update(EvaluatedDefIsocandelApproach._models)

##############################################################################
# Vyhodnocene merene izokandely ##############################################
##############################################################################

# EvaluatedMeasIsocandel ####################################################

class EvaluatedMeasIsocandel(CompositeModel):
	"""Predek vyhodnocenych namerenych izokandel"""
	_models = {
		'lines': ArrayModel(Line()), # cary merene izokandely
		'abbr': StrModel(''), # Zkratka izokandely (pro graf a report)
		'points_found': StrEnumModel(( # Vysledek hledani car na mrizce
			'all_higher', # vsechny body maji vyssi intenzitu.
			'some_between', # Nektere nizsi, nektere vyssi
			'all_lower', # Vsechny body maji nizsi intenzitu nez pozadovana
		), 'some_between'),
	}

	def get_description(self):
		"""Ma vratit text, informace o izokandele""" 
		raise SubclassResponsibilityError

	#-- vyhodnocovani

	def read_from_mesh(self, mesh, intensity):
		"""Nalezne self.lines na dane mrizce a dane urovni (izocara).
		
		mesh: instance storage.Mesh
		intensity: FloatModel
		vraci: True pokud pro nejaky bod mrizky plati ze ma vyssi hodnotu nez
			intensity.  To jeste nemusi nutne znamenat, ze vysledna kontura obsahuje
			nejakou caru, (muze lezet cela okolo mrizky)
		"""
		def get_vpoint(row, col):
			"""Nalezne souradnice bodu s danou hodnotou na vertikale

			bod musi lezet nekde na vertikalni care spojujici m[row][col] a
			m[row+1][col]. Ve vetsine pripadu takovy bod neexistuje a vracim
			None, jinak vratim tuple (x, y) .. souradnice nalezeneho bodu.
			Souradnice pocitam linearni interpolaci.
			"""
			if row < 0 or row >= rows - 1 or col < 0 or col >= cols:
				return None
			tmp = 0
			if m[row    ][col] >= intensity: tmp += 1
			if m[row + 1][col] >= intensity: tmp += 1
			if tmp != 1:
				return None
			(x, y1) = mesh.point_coords(col, row)
			(x, y2) = mesh.point_coords(col, row + 1)
			slope = (m[row+1][col] - m[row][col]) / (y2 - y1)
			return (x, y1 + ((intensity - m[row][col]) / slope))

		def get_hpoint(row, col):
			"""Nalezne souradnice bodu s danou hodnotou na horizontale

			bod musi lezet nekde na horizontalni care spojujici m[row][col] a
			m[row+1][col]. Ve vetsine pripadu takovy bod neexistuje a vracim
			None, jinak vratim tuple (x, y) .. souradnice nalezeneho bodu.
			Souradnice pocitam linearni interpolaci.
			"""
			if row < 0 or row >= rows or col < 0 or col >= cols - 1:
				return None
			tmp = 0
			if m[row][col  ] >= intensity: tmp += 1
			if m[row][col+1] >= intensity: tmp += 1
			if tmp != 1:
				return None
			(x1, y) = mesh.point_coords(col, row)
			(x2, y) = mesh.point_coords(col + 1, row)
			slope = (m[row][col+1] - m[row][col]) / (x2 - x1)
			return (x1 + ((intensity - m[row][col]) / slope), y)
		
		m = mesh.mesh.value
		rows, cols = numpy.shape(m)
		lines = self.lines
		lines.empty()

		min, max = mesh.find_minmax()
		if intensity < min:
			self.points_found.value = 'all_higher'
			return False
		if intensity > max:
			self.points_found.value = 'all_lower'
			return True
		self.points_found.value = 'some_between'
		
		# A naleznu vsechny cary tvorici konturu na urovni intensity
		for r in xrange(rows):
			for c in xrange(cols):
				p1 = get_vpoint(r, c)
				p2 = get_vpoint(r, c+1)
				p3 = get_hpoint(r, c)
				p4 = get_hpoint(r+1, c)
				for a, b in ((p1, p2), (p1, p3), (p1, p4), 
							 (p2, p3), (p2, p4), (p3, p4)):
					if not (a and b):
						continue
					lines.append(Line().set_tuples(a, b))
		return True

	#-- Pridavani do grafu

	def add_labels_to_plot(self, plot_model, lines, color = None, movegroup = ''):
		raise SubclassResponsibilityError

	def add_lines_to_plot(self, plot_model, color = None, movegroup = ''):
		""" Do plot_model vykresli merenou izokandelu.

		vraci: [PlotLines(), .... (popisky atd) ]
		"""
		abbr = self.abbr.value
		lines = plot.PlotLines()
		if self.points_found.value != 'some_between':
			if self.points_found.value == 'all_higher':
				msg = u"%s cel le mimo menou oblast" % self.get_description()
			elif self.points_found.value == 'all_lower':
				msg = u"%s nenalezena" % self.get_description()
			else:
				error('Should not happen')
			plot_model.notes.append(StrModel(msg))

		lines.movegroup.value = '%s_lines' % movegroup
		lines.description.value = self.get_description()
		lines.lines.copy_from(self.lines)
		lines.width.value = 2
		lines.color.set(color)
		plot_model.plottables.append(lines)
		return lines

	def add_to_plot(self, plot_model, color = None, movegroup = ''):
		""" Do plot_model vykresli definovanou izokandelu.
		
		Vykresli cary, minimum, maximum, popisky
		"""
		plot_model.freeze_changes()
		try:
			lines = self.add_lines_to_plot(plot_model, color, movegroup)
			self.add_labels_to_plot(plot_model, lines, color, movegroup)
		finally:
			plot_model.thaw_changes()

# EvaluatedMeasIsocandelDirectional #########################################

class EvaluatedMeasIsocandelDirectional(EvaluatedMeasIsocandel):
	_models = {
		'intensity': FloatModel(1.0),
	}
	_models.update(EvaluatedMeasIsocandel._models)

	def get_description(self):
		return u"Men izokandela %s (I = %.0f)" % (self.abbr.value, self.intensity.value)

	#-- vyhodnocovani

	def eval(self, mesh, intensity, abbr):
		self.intensity.value = intensity
		self.abbr.value = abbr
		self.read_from_mesh(mesh, intensity)

	#-- pridavani do grafu

	def add_labels_to_plot(self, plot_model, lines, color = None, movegroup = ''):
		plot_iso_text(plot_model, u'%s:%.0f' % (self.abbr, self.intensity.value), lines, 'top', color, movegroup)

# EvaluatedMeasIsocandelApproach ############################################

class EvaluatedMeasIsocandelApproach(EvaluatedMeasIsocandel):
	_models = {
		'white_intensity': FloatModel(1.0),
		'red_intensity': FloatModel(1.0),
		}
	_models.update(EvaluatedMeasIsocandel._models)

	def get_description(self):
		"""Ma vratit str, informace o izokandele(pro statusbar, tooltip)""" 
		return u"Men izokandela %s (B/ = %.0f/%.0f)" % \
				(self.abbr.value, self.white_intensity.value, self.red_intensity.value)

	#-- vyhodnocovani

	def verticalize_at_zero(self, zero, lower, upper):
		""" Zesvisli vsechny cary, ktere zacinaji nebo konci v sigma = 0


		Dalsi trik co se musi provest s izokandelami. Vyse popsana uprava
		tvori iluzi, ze prechod mezi cervenou a bilou neni dost ostry, dochazi
		k zhruba nasledujicimu jevu:
              ________
             |        |
		  --/----------\------ sigma = 0
           |____________| 

			==>
		a lepe (z pohledu funkce sestupoveho navestidla, tedy co nejostrejsi
		prechod mezi bilou a cervenou) by vypadalo nasledujici:
              ________
             |        |
		  -------------------- sigma = 0
           |____________| 

		takze pri vyhodnocovani kontury izokandely uplne nakonec zjistim, 
		ktere useky konci v sigma = 0 a "zesvislim" je.
		"""
		if len(self.lines) == 0:
			return

		orig_lines = ArrayModel(Line())
		orig_lines.copy_from(self.lines)
		lines = self.lines
		lines.empty()
		for l in orig_lines:
			x1, y1, x2, y2 = l.get()

			# prvni bod at je vzdycky ten dole, zjednodusim si praci
			if (y1 > y2):
				(x1, y1), (x2, y2) = (x2, y2), (x1, y1)

			if y1 >= upper - epsilon or y2 <= lower + epsilon: 
				lines.append(Line().set(x1, y1, x2, y2)) #mimo oblast okolo nuly
				continue

			if -epsilon <= y2 - upper <= epsilon:
				lines.append(Line().set(x2, zero, x2, y2))
				continue

			if -epsilon <= y1 - lower <= epsilon:
				lines.append(Line().set(x1, y1, x1, zero))
				continue

	def eval(self, mesh, white_intensity, red_intensity, abbr):
		self.white_intensity.value = white_intensity
		self.red_intensity.value = red_intensity
		self.abbr.value = abbr
		wrmesh = WhiteRedMesh.get_wrmesh(mesh, white_intensity, red_intensity)
		self.read_from_mesh(wrmesh, white_intensity)
		avg_row_dist = misc_num.avg_dist(wrmesh.rows.value)
		self.verticalize_at_zero(0, -avg_row_dist, avg_row_dist)

	#-- pridavani do grafu

	def add_labels_to_plot(self, plot_model, lines, color = None, movegroup = ''):
		plot_iso_text(plot_model, u'%s:%.0f' % (self.abbr, self.white_intensity.value), lines, 'top', color, movegroup)
		plot_iso_text(plot_model, u'%s:%.0f' % (self.abbr, self.red_intensity.value), lines, 'bottom', color, movegroup)

# EvaluatedMeasIsocandelRectangle ###########################################

class EvaluatedMeasIsocandelRectangle(EvaluatedMeasIsocandelDirectional):
	pass

# EvaluatedMeasIsocandelEllipse #############################################

class EvaluatedMeasIsocandelEllipse(EvaluatedMeasIsocandelDirectional):
	pass

# EvaluatedMeasIsocandelOval ################################################

class EvaluatedMeasIsocandelOval(EvaluatedMeasIsocandelApproach):
	pass

#############################################################################
# Vyhodnocovani #############################################################
#############################################################################

# EvalIso ###################################################################

class EvalIso(Eval):
	"""Obecne funkce pro vyhodnoceni izokandelovych typovych zkousek """

	isocandel_associations = {
		typetests.IsocandelRectangle: 
			(EvaluatedDefIsocandelRectangle, EvaluatedMeasIsocandelRectangle),
		typetests.IsocandelEllipse: 
			(EvaluatedDefIsocandelEllipse, EvaluatedMeasIsocandelEllipse),
		typetests.IsocandelOval: 
			(EvaluatedDefIsocandelOval, EvaluatedMeasIsocandelOval),
	}

	@classmethod
	def get_def_isocandel(cls, typetests_isocandel):
		try:
			return cls.isocandel_associations[type(typetests_isocandel)][0]
		except KeyError:
			error(u"Neznm typ izokandely: %s." % str(typetests_isocandel)) 

	@classmethod
	def get_meas_isocandel(cls, typetests_isocandel):
		try:
			return cls.isocandel_associations[type(typetests_isocandel)][1]
		except KeyError:
			error(u"Neznm typ izokandely: %s." % str(typetests_isocandel)) 

	@classmethod
	def eval_main_isocandel(cls, min, max, brightness, mesh, main_isocandel):
		"""Vyhodnoti extremy hlavniho svazku (instance typetests.Isocandel).
		
		main_isocandel: hlavni svazek, instance typetests.Isocandel*. Pokud
			neobsahuje zadne body, hodi EvalError.

		Pokud jsou min a max jine nez None, nakopiruje do nich hodnoty minima a
			maxima hlavniho svazku.
		Pokud je brightness(FloatModel) jina nez None, ulozi do ni hodnotu
			stredni svitivosti hlavniho svazku.

		Vraci seznam bodu obsazenych v hlavnim svazku. ([MeasuredPoint, ... ]).

		Predpoklada ze na mesh uz byly aplikovany pripadne korekce mereni.
		"""
		# vyhodnotim extremy hlavniho svazku, svitivost a rovnomernost
		main_points = main_isocandel.find_inner_points(mesh)
		if len(main_points) == 0:
			log_error(u"Hlavn svazek neobsahuje dn body.")
			raise EvalError
		if min != None:
			min.copy_from(reduce(lambda a, b: ternary(a.intensity < b.intensity, a, b), main_points))
		if max != None:
			max.copy_from(reduce(lambda a, b: ternary(a.intensity > b.intensity, a, b), main_points))
		if brightness != None:
			brightness.value = sum(p.intensity for p in main_points)/len(main_points)

	@staticmethod
	def str_isocandel(i, n):
		"""Textova reprezentace n-te izokandely.

		nulta izokandela je vzdy vnitrni, (n-1)-ta vnejsi pokud n != 1, mezitim
		jsou stredni izokandely.
		"""
		if i == 0:       name = u"Vnitn izokandela"
		elif i == n - 1: name = u"Vnj izokandela"
		elif n == 3:     name = u"Stedn izokandela"
		else:            name = u"Stedn izokandela %i" % i
		return name + (' E%i' % (i+1))

	@classmethod
	def eval_isocandels(cls, result, mesh, typetest_isocandels):
		"""Vyhodnoti izokandely typove zkousky. 

		Vyhodnoti izokandely na mrizce mesh. Predpoklada, ze na mesh uz byly
		aplikovany pripadne korekce. Pro kazdou izokandelu vytvori vyhodnocenou
		definovanou a merenou izokandelu, vyhodnotije a prida je do result.

		typetest_isocandels: sekvence typetests.Isocandel*
		"""
		n_isocandels = len(typetest_isocandels)
		for idx, i_typetest in enumerate(typetest_isocandels):
			i_def = cls.get_def_isocandel(i_typetest)()
			descr = cls.str_isocandel(idx, n_isocandels)
			i_def.eval(mesh, i_typetest, u'E%i' % (idx+1), descr)
			i_meas = i_def.eval_meas(mesh)
			result.def_isocandels.append(i_def)
			result.meas_isocandels.append(i_meas)

# def_isocandels_templates ###################################################

def_isocandels_templates = ListModel([
	EvaluatedDefIsocandelRectangle(),
	EvaluatedDefIsocandelEllipse(),
	EvaluatedDefIsocandelOval(),
	])

# meas_isocandels_templates ##################################################

meas_isocandels_templates =  ListModel([
	EvaluatedMeasIsocandelRectangle(),
	EvaluatedMeasIsocandelEllipse(),
	EvaluatedMeasIsocandelOval(),
	])

# EvalResultIso #############################################################

class EvalResultIso(EvalResult):
	"""Vysledek vyhodnoceni izokandelove typove zkousky (TypetestIso)

	To, ze tu mam jak MeasData, tak Mesh neni duplicita, Mesh se sice da z
	MeasData odvodit, ale pokud provedu zmenu kodu tak se muze odvodit jinak. V
	EvalResult musi byt zaznamenano vsechno tak jak se to vyhodnotilo, aby se
	na to dalo podivat pozdeji a vse zustalo zachovano.
	"""
	_models = {
		'min': storage.MeasuredPoint(), # min. v hlavnim svazku
		'max': storage.MeasuredPoint(), # max. v hlavnim svazku
		'def_isocandels': def_isocandels_templates,
		'meas_isocandels': meas_isocandels_templates,
	}
	_models.update(EvalResult._models)
	default_eval_class = EvalIso

	def clear(self):
		self.freeze_changes()
		try:
			self.def_isocandels.empty()
			self.meas_isocandels.empty()
		finally:
			self.thaw_changes()

	def check_verdict(self):
		return self.check_isocandels()

	def check_isocandels(self):
		retval = True
		for i in self.def_isocandels:
			if not i.min_intensity_satisfied.value:
				retval = False
		return retval

	def html_report_isocandels(self):
		return u'<p>\n'.join([i.html_report() for i in self.def_isocandels])

	#-- pridavani do grafu

	def get_lower_lumi_range(self):
		if len(self.def_isocandels):
			return min(i.isocandel.intensity.value for i in self.def_isocandels)
		else:
			return 0.0

	def prepare_plot(self, plot_model):
		plot_model.freeze_changes()
		try:
			self.prepare_plot_labels(plot_model)
			self.prepare_plot_isocandels(plot_model)
		finally:
			plot_model.thaw_changes()

	def prepare_plot_labels(self, plot_model):
		plot_model.freeze_changes()
		try:
			plot_model.xlabel.value = r'$\omega [deg]$'
			plot_model.ylabel.value = r'$\sigma [deg]$'
		finally:
			plot_model.thaw_changes()

	def prepare_plot_isocandels(self, plot_model):
		"""Pripravi plot z vyhodnocene zkousky"""
		plot_model.freeze_changes()
		try:
			for idx, i in enumerate(self.def_isocandels):
				i.add_to_plot(plot_model, color = color_defined(idx), movegroup = 'def')

			for idx, i in enumerate(self.meas_isocandels):
				i.add_to_plot(plot_model, color = color_measured(idx), movegroup = 'meas')
		finally:
			plot_model.thaw_changes()

##############################################################################
# Smerove typove zkousky #####################################################
##############################################################################

# EvalDirectional ###########################################################

class EvalDirectional(EvalIso):
	"""Vyhodnocen typov zkouky pro smrov nvstidla"""

	meas_corrector_settings = {'apply_simfilter': True, 'apply_calibration_matrix': False}

	@classmethod
	def eval(cls, result, mesh, typetest_isocandels, max_uniformity, min_brightness):
		"""Vyhodnoti odmerena data.

		mesh: storage.Mesh, predpoklada ze jiz byly aplikovany prislusne korekce.
		typetest_isocandels: sekvence typetests.Isocandel*. Nultou izokandelu
		povazuji za hlavni svazek.
		
		pri chybe hazi EvalError
		"""
		result.min_brightness.value = min_brightness
		result.max_uniformity.value = max_uniformity
		# vyhodnotim izokandely
		if len(typetest_isocandels):
			cls.eval_isocandels(result, mesh, typetest_isocandels)
			cls.eval_main_isocandel(result.min, result.max, result.brightness, mesh, typetest_isocandels[0])
		cls.eval_unifomity(result)
		result.just_evaluated()

	@classmethod
	def eval_unifomity(cls, result):
		try:
			result.uniformity.value = result.max.intensity/result.min.intensity
		except ZeroDivisionError:
			log_error(u"Chyba pi potn rovnomrnosti, minimum m nulovou intenzitu.")
			raise EvalError

# EvalResultDirectional #####################################################

class EvalResultDirectional(EvalResultIso):
	"""Vysledek vyhodnoceni pro smerova navestidla"""
	_models = {
		'uniformity': FloatModel(0.0),
		'max_uniformity': FloatModel(0.0),
		'brightness': FloatModel(0.0),
		'min_brightness': FloatModel(0.0),
	}
	_models.update(EvalResultIso._models)
	default_eval_class = EvalDirectional

	#-- Testovani vysledku vyhodnoceni

	def check_uniformity(self):
		return self.uniformity.value <= self.max_uniformity.value
		                            
	def check_brightness(self):
		return self.brightness.value >= self.min_brightness.value

	def check_verdict(self):
		return (
			EvalResultIso.check_verdict(self) and
			self.check_uniformity() and
			self.check_brightness()
		)

	#-- html reporty

	@classmethod
	def html_report_main_extremes(self, min, max):
		return u"""
		<b>Extrmy hlavnho svazku</b><br>
		<b>Minimum: </b>%s<br>
		<b>Maximum: </b>%s<br>
		""" % (html_osi(min), html_osi(max))

	def html_report_uniformity(self):
		return u"""
		<b>Poadovan rovnomrnost: </b>%.2f<br>
		<b>Namen rovnomrnost: </b>%.2f<br>
		<b>Test rovnomrnosti: </b>%s<br>
		""" % (
			self.max_uniformity.value,
			self.uniformity.value,
			html_compliant(self.check_uniformity()),
			)

	def html_report_brightness(self):
		return u"""
		<b>Poadovan stedn svtivost: </b>%.2f<br>
		<b>Namen stedn svtivost: </b>%.2f<br>
		<b>Test svtivosti: </b>%s<br>
		""" % (
			self.min_brightness.value,
			self.brightness.value,
			html_compliant(self.check_brightness()),
			)

	def html_report(self):
		return u'<p>\n'.join((
			self.html_report_meas_corrections(),
			self.html_report_main_extremes(self.min, self.max),
			self.html_report_uniformity(),
			self.html_report_brightness(),
			self.html_report_isocandels(),
		))

	#-- pridavani do grafu

	@classmethod
	def prepare_plot_extremes(cls, min, max, plot_model):
		plot_model.freeze_changes()
		try:
			plot_measpt(plot_model, min, u'Min: %.0f' % min.intensity, Color('darkblue'))
			plot_measpt(plot_model, max, u'Max: %.0f' % max.intensity, Color('darkred'))
		finally:
			plot_model.thaw_changes()

	def prepare_plot(self, plot_model):
		plot_model.freeze_changes()
		try:
			EvalResultIso.prepare_plot(self, plot_model)
			self.prepare_plot_extremes(self.min, self.max, plot_model)
			# Vsechny izokandely def_lines udelam pohyblive kvuli zmene zamereni.
			for pl in plot_model.plottables:
				if pl.movegroup.value.find('def_lines') >= 0:
					pl.movable.value = True
					if type(pl) == plot.PlotLines:
						plot_model.movegroups.add_plottable_to_group(pl, 'def_lines')
		finally:
			plot_model.thaw_changes()

##############################################################################
# Sestupove typove zkousky ###################################################
##############################################################################

# BadSymmetryPoint ###########################################################
		
class BadSymmetryPoint(CompositeModel):
	_models = {
		'p0': storage.MeasuredPoint(),
		'p1': storage.MeasuredPoint(),
	}
	def html_report(self):
		try:
			symmetry = u'%.2f' % (self.p0.intensity/self.p1.intensity)
		except ZeroDivisionError:
			symmetry = u"Nedefinovan"
		return u'%s - %s, symetrie: %s' % (html_osi(self.p0), html_osi(self.p1), symmetry)

# EvalApproach ##############################################################

class EvalApproach(EvalIso):
	"""Vyhodnocen typov zkouky pro sestupova nvstidla"""
	meas_corrector_settings = {'apply_simfilter': False, 'apply_calibration_matrix': False}

	@classmethod
	def eval_symmetry(cls, result, mesh, sym_min, sym_max):
		result.sym_min.value = sym_min
		result.sym_max.value = sym_max
		for twins in mesh.find_symmetric_points():
			p0, p1 = mesh.indices2meas_points(twins)
			try:
				s = p0.intensity/p1.intensity
			except ZeroDivisionError:
				log_error(u"Nulov intenzita v erven sti, nelze spotat symetrii")
				raise EvalError
			if sym_min <= s <= sym_max:
				continue
			bsp = BadSymmetryPoint()
			bsp.p0.copy_from(p0)
			bsp.p1.copy_from(p1)
			result.bad_symmetry_points.append(bsp)

	@classmethod
	def eval(cls, result, mesh, typetest_isocandels, sym_min, sym_max):
		"""Vyhodnoti odmerena data.

		mesh: storage.Mesh, predpoklada ze jiz byly aplikovany prislusne korekce.
		typetest_isocandels: sekvence typetests.Isocandel*. Nultou izokandelu
		povazuji za hlavni svazek.
		
		pri chybe hazi EvalError
		"""
		cls.eval_symmetry(result, mesh, sym_min, sym_max)
		if len(typetest_isocandels):
			cls.eval_isocandels(result, mesh, typetest_isocandels)
			cls.eval_main_isocandel(result.min, result.max, None, mesh, typetest_isocandels[0])
		result.just_evaluated()
		
# EvalResultApproach #####################################################

class EvalResultApproach(EvalResultIso):
	"""Vysledek vyhodnoceni pro sestupva navestidla"""
	_models = {
		'bad_symmetry_points': ArrayModel(BadSymmetryPoint()),
		'symmetry': FloatModel(0.0),
		'sym_min': FloatModel(0.0),
		'sym_max': FloatModel(0.0),
	}
	_models.update(EvalResultIso._models)
	default_eval_class = EvalApproach

	def clear(self):
		EvalResultIso.clear(self)
		self.bad_symmetry_points.empty()

	#-- Testovani vysledku

	def check_verdict(self):
		return (
			EvalResultIso.check_verdict(self) and
			self.check_bsp()
		)

	def check_bsp(self):
		return len(self.bad_symmetry_points) == 0

	#-- html reporty

	def html_report_symmetry(self):
		"""Urcene pro potomky, text do reportu popisujici dalsi mereni
		
		Ma vratit html text, ktery bude zarazen do reportu mezi extremy
		hlavniho svazku (posledni polozka pred definici izokandel) a
		definici izokandel. Urcene pro symetrii, rovnomernost atd...
		
		vraci: str
		"""

		report = u"""
		<p>
		<b>Poadovan symetrie:</b> %.2f &lt;= bl/erven &lt; %.2f<br>
		""" % (self.sym_min.value, self.sym_max.value)

		ok = self.check_bsp()
		if not ok:
			report = u''.join((report,
				u"<b>Nalezeny body nesplujc podmnku symetrie:</b><br>\n"))
			report = report + "<br>\n".join([p.html_report() for p in  self.bad_symmetry_points])
		return report + u"<br>\n<b>Test symetrie: </b>%s<br>" % html_compliant(ok)

	def html_report(self):
		return u'<p>\n'.join((
			self.html_report_meas_corrections(),
			self.html_report_symmetry(),
			self.html_report_isocandels(),
		))

	#-- pridavani do grafu

	def prepare_plot(self, plot_model):
		plot_model.freeze_changes()
		try:
			EvalResultIso.prepare_plot(self, plot_model)
			self.add_bsp_to_plot(plot_model)
		finally:
			plot_model.thaw_changes()

	def get_lower_lumi_range(self):
		if len(self.def_isocandels):
			return min(i.isocandel.red_intensity.value for i in self.def_isocandels)
		else:
			return 0.0

	def add_bsp_to_plot(self, plot_model):
		"""Body spatne symetrie"""
		for bsp in self.bad_symmetry_points:
			for pt in bsp.p0, bsp.p1:
				plot_pt = plot.PlotPoint()
				plot_pt.pos.set_xy(pt.omega, pt.sigma)
				plot_pt.markerwidth.value = 1.0
				plot_pt.markersize.value = 5.0
				plot_pt.color.set_name('red')
				plot_model.plottables.append(plot_pt)

###############################################################################
###############################################################################
# Polarni typove zkousky (vsesmeraky, svitidla) ###############################
###############################################################################
###############################################################################

# MinIntensityPoint ###########################################################

class MinIntensityPoint(FloatContainerModel):
	"""Vyhodnocovany bod, ma definovanou minimalni intenzitu."""
	_float_attrs = ('omega', 'sigma', 'intensity', 'min_intensity')
	def is_bad(self):
		return self.intensity < self.min_intensity

	def as_html(self):
		return html_osi(self) + (u', min. I = %.0f cd' % self.min_intensity)

# EvaluatedSpan #############################################################

class EvaluatedSpan(CompositeModel):
	"""Vyhodnoceny SigmaSpan/OmegaSpan"""
	_models = {
		'evaluated_points': ArrayModel(MinIntensityPoint()),
	}

	def find_max_radius(self):
		"""Nalezne nejvetsi polomer, ktery namerenych nebo definovanych car v grafu."""
		def_radius = self.def_span.intensity.value
		if len(self.evaluated_points) == 0:
			return def_radius
		meas_radius = max(pt.intensity for pt in self.evaluated_points)
		return max(meas_radius, def_radius)

	def add_to_plot(self, plot_model, radius, index):
		"""Prida se do grafu. 
		
		plot_model: plot.PlotModel()
		radius: Max. polomer grafu
		index: Poradi cary (urcujici pro barvu cary)
		"""
		def p2c(t, r): 
			"""Polar to cartesian"""
			t = misc_num.deg2rad(t)
			return (r*cos(t), r*sin(t))

		def polar_line(t0, r0, t1, r1):
			x0, y0 = p2c(t0, r0/radius)
			x1, y1 = p2c(t1, r1/radius)
			return Line().set(x0, y0, x1, y1)

		const_unit = self.const_unit
		var_unit = self.var_unit
		plot_model.freeze_changes()
		try:
			line_def = plot.PlotLines()
			line_meas = plot.PlotLines()
			for p1, p2 in pairwise(self.evaluated_points):
				span_from, span_to = getattr(p1, var_unit), getattr(p2, var_unit)
				line_def.lines.append(polar_line(span_from, p1.min_intensity, span_to, p2.min_intensity))
				line_meas.lines.append(polar_line(span_from, p1.intensity, span_to, p2.intensity))

			# definovana cara
			const_val = getattr(self.def_span, const_unit).value

			line_def.width.value = 2
			set_color_defined(line_def.color, index)
			line_def.description.value = u"Definovan pro %s = %.0f" % (const_unit, const_val)
			plot_model.plottables.append(line_def)

			pt = plot.PlotText()
			pt.text.value = u'%.0f' % const_val
			pt.color.copy_from(line_def.color)
			pt.movable.value = True
			line_def.set_text_props(pt, position = 'right')
			plot_model.plottables.append(pt)

			# merena cara
			line_meas.width.value = 3
			set_color_measured(line_meas.color, index)
			line_meas.description.value = u"Men pro %s = %.0f" % (const_unit, const_val)
			plot_model.plottables.append(line_meas)

			pt = plot.PlotText()
			pt.text.value = u'%.0f' % const_val
			pt.color.copy_from(line_meas.color)
			pt.movable.value = True
			line_meas.set_text_props(pt, position = 'right')
			plot_model.plottables.append(pt)
		finally:
			plot_model.thaw_changes()

# EvaluatedOmegaSpan ########################################################

class EvaluatedOmegaSpan(EvaluatedSpan):
	"""Vyhodnoceny OmegaSpan, tedy rozsah pro konstantni sigma."""
	_models = {
		'def_span': typetests.OmegaSpan(),
	}
	_models.update(EvaluatedSpan._models)
	const_unit = 'sigma'
	var_unit = 'omega'

# EvaluatedSigmaSpan ########################################################

class EvaluatedSigmaSpan(EvaluatedSpan):
	"""Vyhodnoceny SigmaSpan, tedy rozsah pro konstantni omega."""
	_models = {
		'def_span': typetests.SigmaSpan(),
	}
	_models.update(EvaluatedSpan._models)
	const_unit = 'omega'
	var_unit = 'sigma'

# EvalPolar #################################################################

class EvalPolar(Eval):
	"""Vyhodnocovac polarniho mereni"""

	@staticmethod
	def eval_span(eval_span, def_span, meas_data, correct_point_func):
		"""Vyhodnoti rozsah na datech meas_data podle definovaneho rozsahu def_span.
		
		Vyhodnocuje nezavisle na min/max parametru u SigmaSpan. 
		Kazdy bod mereni jeste upravi funkci correct_point_func. 
		Updatne eval_span.def_span podle def_span.
		"""
		def find_point(omega, sigma):
			mp = meas_data.find_point(omega, sigma)
			if mp == None:
				raise_point_not_found(omega, sigma)
			p = MinIntensityPoint()
			p.omega, p.sigma, p.intensity = mp.omega_def, mp.sigma_def, mp.intensity
			try:
				correct_point_func(p)
			except MeasCorrectorError:
				log_error(u"Chyba pi korekci mench dat")
				raise EvalError
			return p

		eval_span.evaluated_points.empty()
		for omega, sigma, min_intensity in def_span.points_with_intensity():
			pt = find_point(omega, sigma)
			pt.min_intensity = min_intensity
			eval_span.evaluated_points.append(pt)

		eval_span.def_span.copy_from(def_span)

	@classmethod
	def eval_spans(cls, result, meas_data, def_omega_spans, def_sigma_spans):
		"""Vyhodnoti rozsahy na merenych datech.

		result: EvalResultPolar()
		meas_data: storage.MeasData()
		def_omega_spans: ArrayModel(typetests.OmegaSpan())
		def_sigma_spans: ArrayModel(typetests.SigmaSpan())

		Nejprve se podiva, zda def_omega_spans zahrnuje alespon jeden rozsah. Pokud
		ano, vezme prvni z nich a nalezne v nem bod s nejnizsi/nejvyssi
		intenzitou. Dale vyhodnocuje def_omega_spans, pro kazdy rozsah vytvori
		EvaluatedOmegaSpan, vyhodnoti jej a prida(append) do result_omega_spans.
		Pote vyhodnocuje def_sigma_spans. Mohou nastat tri moznosti:
		 - def_sigma.parameter == 'value': Vyhodnoti primo podle hodnoty
		   def_sigma.omega
		 - def_sigma.parameter != 'value' a zaroven jsem nalezl v def_omega_spans
		   min/max. V tom pripade presune do def_sigma.omega prislusny min/max
		   bod a vyhodnoti.
		 - def_sigma.parameter != 'value' a zaroven jsem nenalezl v def_omega_spans
		   min/max. V tom pripade hodi EvalError.
	    """

		def raise_no_minmax_error(what):
			log_error(u"Seznam rozsah pro konstantn sigma je przdn, nelze najt %s." % what)
			raise EvalError

		correct_point_func = result.meas_corrector.correct_point

		# najdu min/max
		# minimum/maximum, uz prosle korekcni funkci.
		min_point = max_point = None
		if len(def_omega_spans) > 0:
			for o, s in def_omega_spans[0].points():
				pt = meas_data.find_point(o, s)
				if pt == None:
					raise_point_not_found(o, s)
				try:
					correct_point_func(storage.MeasuredPoint().copy_from(pt))
				except MeasCorrectorError:
					log_error(u"Chyba pi korekci intenzity bod.")
					raise EvalError
				if min_point == None or pt.intensity < min_point.intensity:
					min_point =  pt
				if max_point == None or pt.intensity > max_point.intensity:
					max_point =  pt

		# vyhodnotim OmegaSpans
		for def_span in def_omega_spans:
			evaluated_span = EvaluatedOmegaSpan()
			cls.eval_span(evaluated_span, def_span, meas_data, correct_point_func)
			result.omega_spans.append(evaluated_span)

		# vyhodnotim SigmaSpans
		for def_span in def_sigma_spans:
			evaluated_span = EvaluatedSigmaSpan()
			if def_span.parameter.value == 'min':
				if min_point == None:
					raise_no_minmax_error(u"minimum")
				def_span.omega.value = min_point.omega
			elif def_span.parameter.value == 'max':
				if max_point == None:
					raise_no_minmax_error(u"maximum")
				def_span.omega.value = max_point.omega
			cls.eval_span(evaluated_span, def_span, meas_data, correct_point_func)
			result.sigma_spans.append(evaluated_span)

		result.just_evaluated()

# EvalResultPolar ###########################################################

class EvalResultPolar(EvalResult):
	_models = {
		'omega_spans': ArrayModel(EvaluatedOmegaSpan()), # seznam pro konst. sigma
		'sigma_spans': ArrayModel(EvaluatedSigmaSpan()), # seznam pro konst. omega
	}
	_models.update(EvalResult._models)
	default_eval_class = EvalPolar

	def clear(self):
		self.omega_spans.empty()
		self.sigma_spans.empty()

	def check_points(self):
		verdict = True
		for pt in self.iter_all_meas_points():
			if pt.is_bad():
				verdict = False
		return verdict

	def check_verdict(self):
		return self.check_points()

	def iter_all_omega_spans_points(self):
		for span in self.omega_spans:
			for pt in span.evaluated_points:
				yield pt

	def iter_all_sigma_spans_points(self):
		for span in self.sigma_spans:
			for pt in span.evaluated_points:
				yield pt

	def iter_all_meas_points(self):
		for pt in self.iter_all_omega_spans_points():
			yield pt
		for pt in self.iter_all_sigma_spans_points():
			yield pt

	def html_report_points(self):
		points_verdict = self.check_points()
		verdict = (u"<b>Test minimln svtivosti mench bod: </b>%s<br>" %
		            html_compliant(points_verdict))

		if points_verdict:
			return u'<br>' + verdict
		else:
			return u'<br>\n'.join((
			u'',
			u"<b>Nalezeny body nesplujc podmnku minimln svtivosti:</b>",
			u'<br>\n'.join(pt.as_html() for pt in self.iter_all_meas_points() if pt.is_bad()),
			verdict,
			))

	def html_report(self):
		return u'<p>\n'.join((
			self.html_report_meas_corrections(),
			self.html_report_points(),
			))

	def prepare_plot_omega_spans(self, plot_model):
		"""Pripravi graf pro konstantni sigma. Nevyplnuje titulek"""
		return self.prepare_plot(plot_model, self.omega_spans)

	def prepare_plot_sigma_spans(self, plot_model):
		"""Pripravi graf pro konstantni omega. Nevyplnuje titulek"""
		return self.prepare_plot(plot_model, self.sigma_spans)

	def prepare_plot(self, plot_model, spans):
		plot_model.freeze_changes()
		try:
			plot_model.clear()
			plot_model.polar.value = True

			# nejdriv najdu nejvyssi intenzitu a polomerovy rozsah grafu.
			if len(spans) == 0:
				max_i = 1
			else:
				max_i = max(span.find_max_radius() for span in spans)
			radius = misc_num.round_up_1(max_i)
			if radius <= 0:
				log_error(u"Nenalezena dn kladn souadnice")
				raise EvalError
			plot_model.extent.set_lbwh(0, 0, 0, radius) # u polarniho grafu h rika polomer
			for index, span in enumerate(spans):
				span.add_to_plot(plot_model, radius, index)
		finally:
			plot_model.thaw_changes()

# EvalOmni ##################################################################

class EvalOmni(EvalPolar):
	pass

# EvalResultOmni ############################################################

class EvalResultOmni(EvalResultPolar):
	pass

# EvalLight #################################################################

class EvalLight(EvalPolar):
	pass

# EvalResultLight ###########################################################

class EvalResultLight(EvalResultPolar):
	pass

###############################################################################
###############################################################################
# Tabulkove vyhodnoceni #######################################################
###############################################################################
###############################################################################

# EvalTable ################################################################

class EvalTable(Eval):

	@staticmethod
	def find_point(points, omega, sigma):
		"""V ArrayModel(TablePoint) points nalezne bod se souradnici omega, sigma.

		vraci index do pole nebo hodi vyjimku bod nenalezen.
		"""
		for idx, pt in enumerate(points):
			if (omega - epsilon <= pt.omega.value <= omega + epsilon and
				sigma - epsilon <= pt.sigma.value <= sigma + epsilon):
					return idx
		raise_point_not_found(omega, sigma)

	@classmethod
	def eval_points(cls, result, meas_data, def_points):
		"""Vyhodnoti body merenych dat podle definovanych bodu.

		result: instance EvalResult
		meas_data: instance MeasData
		def_points: instance ArrayModel(TablePoint) nebo jinych bodu s
			protokolem TablePoint.  

		Pokud jiz result obsahoval nejake namerene body, nejsou vyrazeny a nove
		vyhodnocene jsou do result pridany.
		"""
		wanted_points = list(def_points)
		for measpt in meas_data.points:
			idx = cls.find_point(wanted_points, measpt.omega_def, measpt.sigma_def)
			wanted_pt = wanted_points.pop(idx)

			pt = MinIntensityPoint()
			pt.omega, pt.sigma, pt.intensity = measpt.omega_def, measpt.sigma_def, measpt.intensity
			pt.min_intensity = wanted_pt.intensity.value
			try:
				result.meas_corrector.correct_point(pt)
			except MeasCorrectorError:
				log_error(u"Chyba pi korekci bod.")
				raise EvalError
			result.evaluated_points.append(pt)

		if len(wanted_points):
			log_error(u"Seznam neodmench bod")
			for p in wanted_points:
				log_error(str_omega_sigma(p.omega, p.sigma))
			log_error(u"Nkter body definovan v typov zkouce nebyly odmeny.")
			raise EvalError

		result.just_evaluated()

# EvalResultTable #############################################################

class EvalResultTable(EvalResult):
	_models = {
		'evaluated_points': ArrayModel(MinIntensityPoint()),
		}
	_models.update(EvalResult._models)
	default_eval_class = EvalTable

	def clear(self):
		self.evaluated_points.empty()

	def check_verdict(self):
		return self.check_points()

	def check_points(self):
		for p in self.evaluated_points:
			if p.is_bad():
				return False
		return True

	def html_report_points(self):
		verdict = self.check_points()
		test_text = u"<b>Test minimln svtivosti mench bod: </b>%s<br>" % html_compliant(verdict)
		if verdict:
			return u'<br>' + test_text

		return u'<br>\n'.join((
			u'',
			u"<b>Nalezeny body nesplujc podmnku minimln svtivosti:</b>",
			u'<br>\n'.join([pt.as_html() for pt in self.evaluated_points if pt.is_bad()]),
			test_text
			))

	def html_report(self):
		return u'<p>\n'.join((
			self.html_report_meas_corrections(),
			self.html_report_points(),
			))
