# -*- encoding: cp1250 -*-

"""Modul pro vyhodnocovani typovych zkousek"""

import wx

from mvc import *
import plot
import storage
import typetests
from misc import SubclassResponsibilityError, native2html
from meas_corrections_gui import gui_meas_corrector
import eval
from error_logging import *

import debug

#############################################################################
# Predci mereni a vysledku ##################################################
#############################################################################

# EvalTypetestError #################################################################

class EvalTypetestError(eval.EvalError):
	"""Vyjimka hazena pri chybe vyhodnocovani typove zkousky"""
	pass

# EvalTypetest ################################################################

class EvalTypetest(object):
	"""Vyhodnocovac typovych zkousek.
	
	Predpoklada se, ze konkretni vyhodnocovace budou vzdy potomky EvalTypetest
	a eval.Eval*
	"""

	@classmethod
	def create_eval_result(cls):
		raise SubclassResponsibilityError

	@classmethod
	def validate_typetest(cls, typetest):
		"""Validuje typovou zkousku.

		Pokud doslo pri validaci k chybe, hodi EvalTypetestError
		"""
		typetest.validate()
		if not typetest.validation.valid:
			log_model_errors(typetest)
			log_error("Chyba pi kontrole typov zkouky")
			raise EvalTypetestError
		return typetest

	@classmethod
	def eval_typetest(cls, meas_data, typetest):
		"""Vyhodnoti typovou zkousku.

		meas_data: instance storage.MeasData
		typetest: instance typetests.Typetest
		vraci: instanci cls.eval_result_class
		"""
		cls.validate_typetest(typetest)
		result = cls.create_eval_result()
		result.eval_result.meas_corrector.set_corrections(
				cls.meas_corrector_settings['apply_simfilter'], 
				cls.meas_corrector_settings['apply_calibration_matrix']
			)
		gui_meas_corrector(result.eval_result.meas_corrector, typetest.get_extent())
		result.meas_data.copy_from(meas_data)
		result.typetest.copy_from(typetest)
		try:
			cls.do_eval_typetest(result, meas_data, typetest)
		except (eval.EvalError, EvalTypetestError):
			log_error(u"Chyba pi vyhodnocovn typov zkouky")
			raise EvalTypetestError
		result.report.changed()
		return result

	@classmethod
	def reeval_typetest(cls, result):
		"""Prevyhodnoti typovou zkousku."""
		cls.validate_typetest(result.typetest)
		result.clear()
		try:
			cls.do_eval_typetest(result, result.meas_data, result.typetest)
		except (eval.EvalError, EvalTypetestError):
			log_error(u"Chyba pi vyhodnocovn typov zkouky")
			raise EvalTypetestError
		result.report.changed()
		return result

	@classmethod
	def do_eval_typetest(cls, result, meas_data, typetest):
		"""Ma za ukol provest samotne vyhodnoceni typove zkousky."""
		raise SubclassResponsibilityError

# EvalTypetestResult #########################################################

class EvalTypetestResult(CompositeModel):
	"""Vysledek vyhodnoceni typove zkousky."""
	_models = {
		'meas_data': storage.MeasData(),
		#'eval_result': eval.EvalResult*
	}
	default_eval_class = EvalTypetest

	def __init__(self):
		CompositeModel.__init__(self)
		self.report = FuncAdaptor(self.html_report)

	def clear(self):
		self.eval_result.clear()

	def reeval_typetest(self):
		return self.default_eval_class.reeval_typetest(self)

	def xml_load(self, *args, **kwargs):
		CompositeModel.xml_load(self, *args, **kwargs)
		self.report.changed()

	def html_report_header(self):
		"""Hlavicka html reportu"""
		return u"""
			<meta HTTP-EQUIV="Content-Type" CONTENT="text/html; charset=cp1250">
			<p><font size="+1"><b>Typov zkouka svtivosti: </b>%s</font><br>
			<p>
			<b>Datum vyhodnocen: </b>%s<br>
			""" % (
				native2html(self.typetest.name.value),
				native2html(str(self.eval_result.eval_timestamp)),
			) + self.meas_data.meas_info.html_report() 

	def check_verdict(self):
		return self.eval_result.verdict.value

	def html_report_footer(self):
		"""Paticka html reportu"""
		verdict_html = eval.html_compliant(self.check_verdict())
		return u"<font size=\"+2\"><b>Vsledek zkouky: </b>%s</font>" % verdict_html

	def ascii_report(self, include_data = True, include_translated_data = True):
		return self.meas_data.ascii_report(include_data, include_translated_data)

	def html_report(self):
		return u'<p>\n'.join((
			self.html_report_header(),
			self.eval_result.html_report(),
			self.html_report_footer(),
			))

	def get_protocol_path(self):
		return self.typetest.dir_proto

	def get_protocol_title(self):
		return (u"Typov zkouka: %s  Odmeno %s  Vyhodnoceno: %s" % (
			self.typetest.name.value, 
			self.meas_data.meas_info.timestamp.str_datetime(), 
			self.eval_result.eval_timestamp.str_datetime()
		))

#############################################################################
# Predci smerovych a sestupovych typovych zkouskek ##########################
#############################################################################

# EvalTypetestIso ############################################################

class EvalTypetestIso(EvalTypetest):
	"""Vyhodnocovac izokandelovych typovych zkousek"""

	@classmethod
	def mesh_read(cls, meas_data):
		"""Nacte mrizku z meas_data nebo hodi chybu."""
		mesh = meas_data.as_mesh()
		if not mesh:
			log_error(u"Chyba pi vytven mky z namench dat")
			raise EvalTypetestError
		return mesh

	@classmethod
	def mesh_scale(cls, mesh, cols, rows):
		try:
			mesh.scale_to_cols_rows(cols, rows)
		except storage.MeshScaleError:
			log_error(u"Chyba pi pizpsobovn dat typov zkouce.")
			raise EvalTypetestError

	@classmethod
	def call_evaluator(cls, result, mesh, typetest):
		raise SubclassResponsibilityError

	@classmethod
	def correct_mesh(cls, mesh, meas_corrector):
		meas_corrector.correct_mesh(mesh)

	@classmethod
	def typetest_fits_in_mesh(cls, typetest, mesh):
		# otestuji, zda se typ. zkouska vejde do namerenych dat
		overlaps = mesh.contains_rect(typetest.get_bbox(), eval.epsilon)
		if overlaps:
			for o in overlaps:
				log_error(o)
			log_error(u"Typov zkouka se nevejde do namench dat")
			raise EvalTypetestError

	@classmethod
	def do_eval_typetest(cls, result, meas_data, typetest):
		eval_result = result.eval_result

		mesh = cls.mesh_read(meas_data)
		cls.mesh_scale(mesh, typetest.get_cols(), typetest.get_rows())
		cls.correct_mesh(mesh, result.eval_result.meas_corrector)
		cls.typetest_fits_in_mesh(typetest, mesh)

		cls.call_evaluator(result, mesh, typetest)

		result.plot.mesh.copy_from(mesh)
		result.prepare_plot()

# EvalTypetestResultIso ######################################################

class EvalTypetestResultIso(EvalTypetestResult):
	"""Vysledek vyhodnoceni izokandelovych typovych zkousek"""
	_models = {
		# 'typetest': typetests.Typetest...(),
		# 'eval_result': eval.EvalResult...(),
		'plot': plot.PlotModel(),
	}
	_models.update(EvalTypetestResult._models)

	def _get_mesh(self):
		return self.plot.mesh
	mesh = property(_get_mesh)

	def clear(self):
		EvalTypetestResult.clear(self)
		self.plot.clear()

	def prepare_plot_mesh(self):
		self.plot.extent.copy_from(self.typetest.get_extent())
		self.plot.grid.set_grid_offset(True,
				self.typetest.omega_min.value, self.typetest.omega_step.value,
				self.typetest.sigma_min.value, self.typetest.sigma_step.value)

	def prepare_plot_lumi_ranges(self):
		min_val = self.eval_result.get_lower_lumi_range()
		_, max_val = self.mesh.find_minmax()
		if not max_val > min_val:
			return
		self.plot.lumi_min.value, self.plot.lumi_max.value = min_val, max_val

	def get_plot_title(self):
		raise SubclassResponsibilityError

	def prepare_plot(self):
		"""Pripravi graf.

		predpoklada, ze do plot.mesh uz je nakopirovana spravna mrizka.
		"""
		self.prepare_plot_lumi_ranges()
		self.prepare_plot_mesh()
		self.eval_result.prepare_plot(self.plot) # pred timto musi byt prepare_plot_mesh
		self.plot.title.value = self.get_plot_title()

#############################################################################
# Smerove typove zkousky ####################################################
#############################################################################

# EvalTypetestDirectional ####################################################

class EvalTypetestDirectional(EvalTypetestIso):
	"""Vyhodnocovac smerovych typovych zkousek"""
	meas_corrector_settings = {'apply_simfilter': True, 'apply_calibration_matrix': True}

	@classmethod
	def create_eval_result(cls):
		return EvalTypetestResultDirectional()

	@classmethod
	def call_evaluator(cls, result, mesh, typetest):
		new_typetest = typetest.copy()
		new_typetest.move_isocandels(result.defined_offset)
		cls.typetest_fits_in_mesh(new_typetest, mesh)
		eval.EvalDirectional.eval(result.eval_result, mesh, new_typetest.isocandels, 
			new_typetest.uniformity.value, new_typetest.brightness.value)


# EvalTypetestResultDirectional ##############################################

class EvalTypetestResultDirectional(EvalTypetestResultIso):
	"""Vysledek vyhodnoceni smerovych typovych zkousek"""
	_models = {
		'typetest': typetests.TypetestDirectional(),
		'eval_result': eval.EvalResultDirectional(),
		'defined_offset': Point(),
		}
	_models.update(EvalTypetestResultIso._models)
	default_eval_class = EvalTypetestDirectional

	max_offset = Point().set_xy(1.0, 1.0)

	def prepare_plot(self):
		EvalTypetestResultIso.prepare_plot(self)
		if not self.check_offset():
			self.plot.notes.append(StrModel(u"Pekroena maximln zmna zamen"))

	def check_verdict(self):
		return (EvalTypetestResultIso.check_verdict(self) and self.check_offset())

	def check_offset(self):
		mo = self.max_offset
		bounds = Rect().set_p1p2(mo.negated(), mo).expand(eval.epsilon, eval.epsilon)
		return bounds.contains_point(self.defined_offset)

	def html_report_offset(self):
		"""HTML report zmeny zamereni"""
		return u"""
		<b>Zmna zamen: </b>%s<br>
		<b>Maximln povolen zmna zamen: </b>%s<br>
		<b>Test zmny zamen: </b>%s<br>
		""" % (
			eval.html_os(self.defined_offset.x, self.defined_offset.y),
			eval.html_os(self.max_offset.x, self.max_offset.y),
			eval.html_compliant(self.check_offset()),
			)

	def html_report(self):
		return u'<p>\n'.join((
			self.html_report_header(),
			self.html_report_offset(),
			self.eval_result.html_report(),
			self.html_report_footer(),
		))

	def get_plot_title(self):
		return u"Typov zk. smrovch nvstidel: %s" % self.typetest.name.value

#############################################################################
# Sestupove typove zkousky ##################################################
#############################################################################

# EvalTypetestApproach ######################################################

class EvalTypetestApproach(EvalTypetestIso):
	"""Vyhodnocovac sestupovych typovych zkousek"""
	meas_corrector_settings = {'apply_simfilter': False, 'apply_calibration_matrix': False}

	@classmethod
	def create_eval_result(cls):
		return EvalTypetestResultApproach()

	@classmethod
	def call_evaluator(cls, result, mesh, typetest):
		e = eval.EvalApproach
		e.eval(result.eval_result, mesh, typetest.isocandels, typetest.sym_min.value, typetest.sym_max.value)

# EvalTypetestResultApproach ################################################

class EvalTypetestResultApproach(EvalTypetestResultIso):
	"""Vysledek vyhodnoceni smerovych typovych zkousek"""
	_models = {
		'typetest': typetests.TypetestApproach(),
		'eval_result': eval.EvalResultApproach(),
		'defined_offset': Point(),
		}
	_models.update(EvalTypetestResultIso._models)
	default_eval_class = EvalTypetestApproach

	def get_plot_title(self):
		return u"Typov zk. sestupovch nvstidel: %s" % self.typetest.name.value

###############################################################################
# Polarni typove zkousky (vsesmeraky, svitidla) ###############################
###############################################################################

# EvalPolar ###################################################################

class EvalTypetestPolar(EvalTypetest):
	"""Vyhodnocovac polarniho mereni"""

	@classmethod
	def get_eval_class(cls):
		raise SubclassResponsibilityError

	@classmethod
	def do_eval_typetest(cls, result, meas_data, typetest):
		e = cls.get_eval_class()
		eval_result = result.eval_result
		plot_title = cls.get_plot_title()

		e.eval_spans(eval_result, meas_data, typetest.omega_spans, typetest.sigma_spans)
		eval_result.prepare_plot_omega_spans(result.plot_omega_spans)
		result.plot_omega_spans.title.value = u"%s, konst. sigma" % plot_title
		eval_result.prepare_plot_sigma_spans(result.plot_sigma_spans)
		result.plot_sigma_spans.title.value = u"%s, konst. omega" % plot_title

# EvalResultPolar ###########################################################

class EvalTypetestResultPolar(EvalTypetestResult):
	_models = {
		'meas_data': storage.MeasData(),
		'plot_omega_spans': plot.PlotModel(), # graf pro konst. sigma
		'plot_sigma_spans': plot.PlotModel(), # graf pro konst. omega
	}
	_models.update(EvalTypetestResult._models)
	default_eval_class = EvalTypetestPolar

	def clear(self):
		EvalTypetestResult.clear(self)
		self.plot_omega_spans.clear()
		self.plot_sigma_spans.clear()

# EvalOmni ##################################################################

class EvalTypetestOmni(EvalTypetestPolar):
	meas_corrector_settings = {'apply_simfilter': True, 'apply_calibration_matrix': True}

	@classmethod
	def get_plot_title(cls):
		return u"Typov zk. vesmr. nvstidel"

	@classmethod
	def get_eval_class(cls):
		return eval.EvalOmni

	@classmethod
	def create_eval_result(cls):
		return EvalTypetestResultOmni()

# EvalResultOmni ############################################################

class EvalTypetestResultOmni(EvalTypetestResultPolar):
	_models = {
			'typetest': typetests.TypetestOmni(), 
			'eval_result': eval.EvalResultOmni(),
		}
	_models.update(EvalTypetestResultPolar._models)
	default_eval_class = EvalTypetestOmni

# EvalLight #################################################################

class EvalTypetestLight(EvalTypetestPolar):
	meas_corrector_settings = {'apply_simfilter': True, 'apply_calibration_matrix': True}

	@classmethod
	def get_plot_title(cls):
		return u"Typov zk. svtidel"

	@classmethod
	def get_eval_class(cls):
		return eval.EvalLight

	@classmethod
	def create_eval_result(cls):
		return EvalTypetestResultLight()


# EvalResultLight ###########################################################

class EvalTypetestResultLight(EvalTypetestResultPolar):
	_models = { 
			'typetest': typetests.TypetestLight(),
			'eval_result': eval.EvalResultLight(),
			}
	_models.update(EvalTypetestResultPolar._models)
	default_eval_class = EvalTypetestLight

###############################################################################
# Tabulkove typove zkousky ####################################################
###############################################################################

# EvalTypetestTable ###########################################################

class EvalTypetestTable(EvalTypetest):
	"""Vyhodnocovac tabulkovych typovych zkousek"""
	meas_corrector_settings = {'apply_simfilter': True, 'apply_calibration_matrix': True}

	@classmethod
	def create_eval_result(cls):
		return EvalTypetestResultTable()

	@classmethod
	def do_eval_typetest(cls, result, meas_data, typetest):
		eval.EvalTable.eval_points(result.eval_result, meas_data, typetest.points)

# EvalTypetestResultTable ####################################################

class EvalTypetestResultTable(EvalTypetestResult):
	_models = {
		'typetest': typetests.TypetestTable(),
		'eval_result': eval.EvalResultTable(),
		}
	_models.update(EvalTypetestResult._models)
	default_eval_class = EvalTypetestTable
