# -*- encoding: cp1250 -*-

"""Korekce mereni. (Simulovany filtr, kalibracni matice, pripadne dalsi... """

from mvc import *
import registry
from storage import MeasData, Mesh, MeshError, MeshScaleError
from bicubic import BicubicMesh

class MeasCorrectorError(Exception):
	pass

# SimFilter ####################################################################

class SimFilter(CompositeModel):
	_models = {
		'enabled': BoolModel(False),
		'value': FloatModel(100.0),
		'color': StrModel("Barva"),
	}
	def validator(self):
		if not 0.0 <= self.value.value <= 100.0:
			msg = u"Hodnota mus leet v rozsahu 0 .. 100 %"
			self.validator_failure(msg, self.value)
			return False
		return True

	def html_report(self):
		simfil_enabled = u"<b>Simulovan filtr: </b>%s<br>" % \
		                 {True: self.color.value, False: u"Nen"}[self.enabled.value]
		if not self.enabled.value:
			return simfil_enabled
		return '\n'.join((simfil_enabled,
						 u"<b>Hodnota simulovanho filtru: </b>%.1f %%<br>" % self.value.value
		                ))

	def get_ratio(self):
		if not self.enabled.value:
			return 1.0
		return self.value.value/100.0

	def correct_point(self, point):
		if self.enabled.value:
			point.intensity *= self.get_ratio()

	def correct_mesh(self, mesh):
		if self.enabled.value:
			mesh.multiply(self.get_ratio())

# CalibrationMatrixError ####################################################

class CalibrationMatrixError(MeasCorrectorError):
	pass

# CalibrationMatrix #########################################################

class CalibrationMatrix(CompositeModel):
	_models = {
		'enabled': BoolModel(False),
		'meas1': MeasData(),
		'meas2': MeasData(),
	}

	def __init__(self):
		CompositeModel.__init__(self)
		self.add_change_observer(self.invalidate_matrix)
		self.calibration_mesh = None
		self.bicubic_mesh = None
		self.meas1.add_change_observer(self.invalidate_matrix)
		self.meas2.add_change_observer(self.invalidate_matrix)

		self.description = StrModel()
		self.add_change_observer(self.on_changed)

	def on_changed(self):
		self.description.value = u'\n'.join((
			u"",
			u"Men z kroku 1",
			u"----------------",
			u"",
			self.meas1.meas_info.ascii_report(),
			u"",
			u"Men z kroku 2",
			u"----------------",
			u"",
			self.meas2.meas_info.ascii_report(),
		))

	def html_report(self):
		if self.enabled.value:
			return u"\n <b>Kalibran matice: </b>Ano<br> \n"
		else:
			return u"\n <b>Kalibran matice: </b>Ne<br> \n"

	def get_calibration_mesh(self):
		""" Spocte kalibracni matici z namerenych dat. 
		
		Kalibracni matice je matice merena jednim zpusobem / matice merena
		druhym zpusobem.
		
		Chyby hlasim pomoci log_error a hodim CalibrationMatrixError

		vraci: vyslednou kalibracni mrizku (instanci Mesh)
		"""
		# Proted predpokladam, ze mam meas1 i meas2 odmerene ve
		# stejnych bodech. Sice to kontroluji, ale kdyz by se nahodou ukazalo,
		# ze nemam, tak jen hodim chybu.

		def raise_err(txt):
			log_error(txt)
			log_error(u"Chyba pi potn kalibran mky z namench dat.")
			raise CalibrationMatrixError

		if self.calibration_mesh != None:
			return self.calibration_mesh

		if self.is_empty():
			self.calibration_mesh = Mesh()
			return self.calibration_mesh

		mesh1 = self.meas1.as_mesh()
		if mesh1 == None:
			raise_err(u"Chyba pi pevodu dat z kroku 1 do mky.")

		mesh2 = self.meas2.as_mesh()
		if mesh2 == None:
			raise_err(u"Chyba pi pevodu dat z kroku 2 do mky.")

		if not mesh2.has_same_points_as(mesh1, 0.005):
			raise_err(u"Mky z kroku 1 a z kroku 2 jsou odmeny v jinch bodech!")

		try:
			mesh2.divide_by_mesh(mesh1)
		except MeshError:
			log_error(u"Chyba pi potn kalibran matice")

		self.calibration_mesh = mesh2
		return self.calibration_mesh

	def validate(self):
		CompositeModel.validate(self)
		if not self.validation.valid:
			return
		mesh = self.get_calibration_mesh()
		mesh.validate()
		if not mesh.validation.valid:
			self.validation.set(False, u"Chyba pi kontrole kalibran matice", mesh)

	def get_bicubic_mesh(self):
		if not self.bicubic_mesh:
			cm = self.get_calibration_mesh()
			self.bicubic_mesh = BicubicMesh(cm.mesh.value, cm.cols.value, cm.rows.value)
		return self.bicubic_mesh

	def get_extent(self):
		return self.get_calibration_mesh().get_extent()

	def is_empty(self):
		return self.meas1.is_empty() or self.meas2.is_empty()

	def invalidate_matrix(self):
		self.calibration_mesh = self.bicubic_mesh = None

	def correct_point(self, point):
		try:
			point.intensity *= self.interpolate_point(point.omega, point.sigma)
		except CalibrationMatrixError:
			raise CalibrationMatrixError(u"Chyba pi korekci pomoc kalibran matice")

	def interpolate_point(self, omega, sigma):
		"""Spocte interpolovanou hodnotu v bode omega, sigma. 
		
		vraci: float

		Pokud se bod nachazi prilis daleko od interpolovane mrizky, zahlasim to
		pomoci log_error a hodim CalibrationMatrixError
		"""
		if not self.enabled.value:
			return 1.0

		if self.is_empty():
			log_error(u"Kalibran mka przdn")
			raise CalibrationMatrixError

		mesh = self.get_calibration_mesh()
		bounds = mesh.get_extent().expand(mesh.get_omega_step(), mesh.get_sigma_step())
		if not bounds.contains_point(Point().set_xy(omega, sigma)):
			log_error(u"Interpolovan bod le pli daleko od kalibran mky")
			raise CalibrationMatrixError

		return self.get_bicubic_mesh().interpolate_point(omega, sigma)

	def correct_mesh(self, mesh):
		if not self.enabled.value:
			return
		new_calibration_mesh = self.get_calibration_mesh().copy()
		try:
			new_calibration_mesh.scale_to_mesh(mesh)
		except MeshScaleError:
			log_error(u"Chyba pi aplikaci kalibran matice")
			raise CalibrationMatrixError
		mesh.multiply_by_mesh(new_calibration_mesh)

# MeasCorrector ##########################################################

class MeasCorrector(CompositeModel):
	_models = {
		'simfilter': SimFilter(),
		'calibration_matrix': CalibrationMatrix(),
	}

	def __init__(self, apply_simfilter = False, apply_calibration_matrix = False):
		CompositeModel.__init__(self)
		self.set_corrections(apply_simfilter, apply_calibration_matrix)

	def set_corrections(self, apply_simfilter = False, apply_calibration_matrix = False):
		self.apply_simfilter = apply_simfilter
		self.apply_calibration_matrix = apply_calibration_matrix
		return self

	def correct_point(self, point):
		"""Aplikuje merici korekce na jeden bod 
		
		point: Musi mit tri floatove atributy: omega, sigma, intensity.
		point.intensity bude zmenena podle korekci a souradnic bodu.
		"""
		if self.apply_simfilter:
			self.simfilter.correct_point(point)
		if self.apply_calibration_matrix:
			self.calibration_matrix.correct_point(point)

	def correct_mesh(self, mesh):
		"""Aplikuje merici korekce na celou mrizku (storage.Mesh)"""
		if self.apply_simfilter:
			self.simfilter.correct_mesh(mesh)
		if self.apply_calibration_matrix:
			self.calibration_matrix.correct_mesh(mesh)

	def html_report(self):
		reports = []
		if self.apply_simfilter:
			reports.append(self.simfilter.html_report())
		if self.apply_calibration_matrix:
			reports.append(self.calibration_matrix.html_report())
		return u'<br>'.join(reports)
