# -*- encoding: cp1250 -*-

""" korekce mereni - GUI """

__all__ = ('SimFilterDlg', 'ApplyCalibrationDlg', 'gui_measure_matrix', 'gui_meas_corrector')

from mvc import *
import registry

from meas_corrections import SimFilter, CalibrationMatrix, CalibrationMatrixError

from measure import MeasureIso
import measure_gui
from gfm_common_gui import *

from storage import MeasInfo, MeshError
from ckdm_interface import version as ckdm_interface_version
from plot import PlotModel, PlotView, PlotController

resources = registry.xresources.meas_corrections

# SimFilterDlg ##################################################################

class SimFilterDlg(Dialog):
	def __init__(self, model, parent = None):
		Dialog.__init__(self, model, resources, 'Dlg_Simfil', parent)
		m, f = self.model, self.frame
		self.add_view(FloatEditor(m.value, f, 'Text_value', '%.1f'))
		self.add_view(StrEditor(m.color, f, 'Text_color'))

	def post_ok(self):
		self.model.enabled.value = True

# CalibrationMatrixExecutionChain ###########################################

class CalibrationMatrixExecutionChain(object):
	"""Popisuje jednotlive kroky mereni kalibracni matice."""
	def __init__(self, model, parent = None):
		"""model: CalibrationMatrix"""
		self.model, self.parent = model, parent

	def run_mpd(self, mpd):
		if ckdm_interface_version == 'simul':
			mpd.run_modeless()
		else:
			mpd.run_modal()

	def run(self, parent = None):
		try:
			self.measure_step_1 = MeasureIso()
			self.measure_step_1.title.value = u"Odmr kroku 1"
			self.measure_step_1.meas_data.meas_info.meas_type.value = MeasInfo.NOWEDGE
			mpd = MeasurePrepareStep1Dlg(self.measure_step_1, self.parent.frame)
			mpd.set_ok_callback(self.on_measure_prepare_1_ok)
		except measure_gui.MeasurePrepareError:
			return
		self.run_mpd(mpd)

	def on_measure_prepare_1_ok(self):
		mf = CalibrationMeasureFrame(self.measure_step_1, self.parent.frame)
		mf.continue_action = self.on_measure_1_ok
		mf.run()

	def on_measure_1_ok(self):
		self.model.meas1.copy_from(self.measure_step_1.meas_data)

		self.measure_step_2 = self.measure_step_1.copy()
		self.measure_step_2.meas_data.empty()
		self.measure_step_2.meas_data.meas_info.meas_type.value = MeasInfo.NOWEDGE
		self.measure_step_2.title.value = u"Odmr kroku 2"
		try:
			mpd = MeasurePrepareStep2Dlg(self.measure_step_2, self.parent.frame)
			mpd.set_ok_callback(self.on_measure_prepare_2_ok)
		except measure_gui.MeasurePrepareError:
			return
		self.run_mpd(mpd)

	def on_measure_prepare_2_ok(self):
		mf = CalibrationMeasureFrame(self.measure_step_2, self.parent.frame)
		mf.continue_action = self.on_measure_2_ok
		mf.run()

	def on_measure_2_ok(self):
		self.model.meas2.copy_from(self.measure_step_2.meas_data)
		CalibrationMatrixFrame.run_frame(self.model)

# MeasurePrepareStep1Dlg ####################################################

class MeasurePrepareStep1Dlg(measure_gui.MeasurePrepareBasic):
	def __init__(self, model, parent = None):
		measure_gui.MeasurePrepareBasic.__init__(self, model, parent)
		self.attach_panel(measure_gui.resources, 'UNKN_custom_panel', resources, 'Panel_MeasureCalibrationMatrix1')

		for coord in 'omega_min', 'omega_max', 'omega_step', 'sigma_min', 'sigma_max', 'sigma_step':
			self.add_view(OmegaSigmaEditor(getattr(self.model, coord), self.frame, 'Text_%s' % coord))

	def get_meas_info(self):
		return self.model.meas_data.meas_info

# MeasurePrepareStep2Dlg ####################################################

class MeasurePrepareStep2Dlg(measure_gui.MeasurePrepareBasic):
	def __init__(self, model, parent = None):
		measure_gui.MeasurePrepareBasic.__init__(self, model, parent)
		self.attach_panel(measure_gui.resources, 'UNKN_custom_panel', resources, 'Panel_MeasureCalibrationMatrix2')

	def get_meas_info(self):
		return self.model.meas_data.meas_info

# CalibrationMeasureFrame ####################################################

class CalibrationMeasureFrame(measure_gui.UnSaveableMeasuresMixin, measure_gui.MeasureIsoFrame):
	resources = resources
	frame_id = 'Frame_MeasureCalibrationMatrix'

	def __init__(self, model, parent): 
		measure_gui.MeasureIsoFrame.__init__(self, model, parent)
		measure_gui.UnSaveableMeasuresMixin.init_unsaveable_measures(self)

# CalibrationMatrixView #####################################################

class CalibrationMatrixView(View):
	"""Pohled na kalibracni matici. """
	def __init__(self, model, frame, res, unkn_id):
		View.__init__(self, model, frame)
		m, f = self.model, self.frame

		self.views = ViewGroup()
		self.attach_panel(res, unkn_id, resources, 'Panel_CalibrationMatrix')

		self.plot1 = PlotModel()
		plot1_view = PlotView(self.plot1, f, resources, 'Plot1')
		self.views.append(plot1_view)
		self.plot1_controller = PlotController(self.plot1, plot1_view, 'omega', 'sigma')

		self.plot2 = PlotModel()
		plot2_view = PlotView(self.plot2, f, resources, 'Plot2')
		self.views.append(plot2_view)
		self.plot2_controller = PlotController(self.plot2, plot2_view, 'omega', 'sigma')

		self.plot_matrix = PlotModel()
		plot_matrix_view = PlotView(self.plot_matrix, f, resources, 'Plot_matrix')
		self.views.append(plot_matrix_view)
		self.plotmatrix_controller = PlotController(self.plot_matrix, plot_matrix_view, 'omega', 'sigma')

		self.views.append(StrView(m.description, f, 'Text_description'))

	def update(self):
		m = self.model
		prepare_isoplot_from_meas_data(self.plot1, m.meas1, '')
		prepare_isoplot_from_meas_data(self.plot2, m.meas2, '')
		if m.meas1.is_empty() or m.meas2.is_empty():
			return 
		try:
			mesh = m.get_calibration_mesh()
		except CalibrationMatrixError:
			log_error(u"Chyba pi potn kalibran mky z namench dat.")
			return
		prepare_isoplot_from_mesh(self.plot_matrix, mesh, '')
		self.plot_matrix.colorbar_title.value = u"Hodnota"
		self.views.update_all_views()

# CalibrationMatrixFrame ####################################################

class CalibrationMatrixFrame(DocumentFrame):
	def __init__(self, model, parent = None):
		DocumentFrame.__init__(self, model, resources, 'Frame_CalibrationMatrix', parent)
		m, f = self.model, self.frame

		self.add_view(CalibrationMatrixView(m, f, resources, 'UNKN_CalibrationMatrix'))

		self.set_menu('MB_CalibrationMatrix')
		self.bind_file_events(None, 'MI_Open', 'MI_Save', 'MI_SaveAs', 'MI_Quit')
		self.setup_file_dialogs(registry.paths.calibration_matrix, '', 
			u"Otevt soubor s kalibran matic", "Uloit soubor s kalibran matic", 
			registry.wildcards.xml)

# ApplyCalibrationDlg #######################################################

class ApplyCalibrationDlg(Dialog):
	def __init__(self, model, parent = None, minsize = None):
		Dialog.__init__(self, model, resources, 'Dlg_ApplyCalibrationMatrix', parent)

		self.matrix_fname = FilenameModel()
		self.minsize = minsize

		self.filename_browse_editor = FilenameBrowseEditor(
			self.matrix_fname, self.frame, 'BrowseText_matrix', 'BrowseBut_matrix', 
			dir = registry.paths.calibration_matrix, wildcard = registry.wildcards.xml, 
			callback = self.read_matrix)
		self.add_view(self.filename_browse_editor)

		self.add_view(CalibrationMatrixView(self.model, self.frame, resources, 'UNKN_CalibrationMatrix'))

	def read_matrix(self, evt = None):
		try:
			self.model.xml_load(self.matrix_fname.value)
		except XMLModelLoadError:
			self.filename_browse_editor.notify_validation(False)
			log_error(u"Chyba pi natn kalibran matice")
			return False
		self.model.validate()
		if not self.model.validation.valid:
			log_model_errors(self.model.typetest)
			self.filename_browse_editor.notify_validation(False)
			log_error(u"Chyba pi kontrole kalibran matice")
			return False
		self.filename_browse_editor.notify_validation(True)
		return True

	def valid_or_report_error(self):
		if not self.read_matrix():
			return False
		if self.minsize != None and True != self.model.get_calibration_mesh().contains_rect(self.minsize):
			log_error(u"Tato kalibran matice nepokryje cel namen data.")
			return False
		return Dialog.valid_or_report_error(self)

	def post_ok(self):
		self.model.enabled.value = True

# gui_measure_matrix ########################################################

def gui_measure_matrix(parent):
	matrix = CalibrationMatrix()
	execution_chain = CalibrationMatrixExecutionChain(matrix, parent)
	execution_chain.run()

# gui_meas_corrector #####################################################

def gui_meas_corrector(meas_corrector, minsize = None):
		"""Interaktivne si vyzada parametry korekcnich funkci. 

		To, ktere funkce budou dotazovany, se nastavuje primo v
		meas_corrector pomoci metody set_corrections.
		"""
		if meas_corrector.apply_simfilter:
			sm = SimFilter()
			if wx.ID_OK == SimFilterDlg.run_dialog(sm):
				meas_corrector.simfilter.copy_from(sm)

		if meas_corrector.apply_calibration_matrix:
			cm = CalibrationMatrix()
			dlg = ApplyCalibrationDlg(cm, minsize = minsize)
			if wx.ID_OK == dlg.run_modal():
				meas_corrector.calibration_matrix.copy_from(cm)

# gui_view_calibration_matrix ################################################

def gui_view_calibration_matrix(parent):
	CalibrationMatrixFrame.run_frame(CalibrationMatrix())
