# -*- encoding: cp1250 -*-

from mvc import *
from storage import MeasData
import typetests
from typetests import validator_omega, validator_sigma, validator_at_least_005, \
                      validate_omega_step, validate_sigma_step
import debug

# tuhle znacku vraci next_point, kdyz jeste nedokaze vratit dalsi bod. Muze se
# stat hlavne kdyz souradnice dalsiho bodu zavisi na namerenych hodnotach
# (priklad treba MeasurePolar)
TRY_LATER='try later'

# Hazena kdyz se nepodari spocitat dalsi bod mereni.
class NextPointError(Exception):
	pass

# LenAdaptor ################################################################

class LenAdaptor(object):
	def __init__(self, adapted_model, target):
		self.adapted_model, self.target = adapted_model, target
		self.adapted_model.add_change_observer(self.on_model_changed)
		self.on_model_changed()

	def on_model_changed(self):
		self.target.value = len(self.adapted_model)

# Measure ###################################################################

class Measure(CompositeModel):
	"""Zakladni trida popisujici mereni

	Tato trida a tridy z ni odvozene se pouzivaji pro mereni novych dat.  Maji
	atribut meas_data, do ktereho se ukladaji nove odmerene body, a metody
	ktere rikaji, ktere dalsi body se maji odmerit.
	"""
	_models = {
		'meas_data': MeasData(),
	}

	def __init__(self):
		CompositeModel.__init__(self)
		self.n_points = IntModel()
		self.title = StrModel()

	def next_point(self):
		"""Vrati dalsi bod
		
		vraci: 
			1. tuple (new_omega, new_sigma) - souradnice dalsiho bodu k mereni
			2. TRY_LATER - potrebuji mit odmerene dalsi body k tomu, abych
				zjistil jaky dalsi bod se ma merit. Zavolej next_point znovu az
				bude k dispozici vice dat.
			3. None - Zadny dalsi bod uz odmerit nechci.
		"""
		raise NotImplementedError

	def reset_points(self):
		"""Po zavolani teto fce next_point vrati prvni bod k mereni."""
		raise NotImplementedError

# MeasureIso #################################################################

class MeasureIso(Measure):
	"""Mereni izokandelovych (mrizkovych) dat"""
	_models = {
		'omega_min': FloatModel(0.0, validator_omega),
		'omega_max': FloatModel(0.0, validator_omega),
		'omega_step': FloatModel(1.0, validator_at_least_005),
		'sigma_min': FloatModel(0.0, validator_sigma),
		'sigma_max': FloatModel(0.0, validator_sigma),
		'sigma_step': FloatModel(1.0, validator_at_least_005),
	}
	_models.update(Measure._models)

	class IsoPointsCounter(object):
		def __init__(self, adapted_model, target):
			self.adapted_model, self.target = adapted_model, target
			self.adapted_model.add_change_observer(self.on_model_changed)
			self.on_model_changed()

		def on_model_changed(self):
			a = self.adapted_model
			try:
				new_cnt = int((a.omega_max.value-a.omega_min.value)/a.omega_step.value + 1) * \
				          int((a.sigma_max.value-a.sigma_min.value)/a.sigma_step.value + 1)
			except ZeroDivisionError:
				new_cnt = 0
			self.target.value = new_cnt

	def __init__(self):
		Measure.__init__(self)
		self.n_points_counter = self.IsoPointsCounter(self, self.n_points)

	def init_on_iso_range(self, iso_range, title = ''):
		"""Inicializuje mereni na definici rozsahu.

		iso_range musi obsahovat atributy omega_min/max/step,
			sigma_min/max/step (FloatModely).
		"""
		self.freeze_changes()
		try:
			for a in 'omega_min', 'omega_max', 'omega_step', 'sigma_min', 'sigma_max', 'sigma_step':
				getattr(self, a).value = getattr(iso_range, a).value
			self.reset_points()
			self.title.value = title
		finally:
			self.thaw_changes()

	def init_on_typetest(self, typetest_iso):
		"""Inicializuje mereni na definici izokandelove typove zkousky"""
		self.init_on_iso_range(typetest_iso, typetest_iso.name.value)

	def reset_points(self):
		self.last_coords = (None, None)

	def _next(self, omega, sigma):
		new_omega = omega + self.omega_step.value
		new_sigma = sigma
		if new_omega > (self.omega_step.value/2) + self.omega_max.value:
			new_omega = self.omega_min.value
			new_sigma += self.sigma_step.value
			if new_sigma > (self.sigma_step.value/2) + self.sigma_max.value:
				return None
		return (new_omega, new_sigma)

	def next_point(self):
		"""Vrati dalsi bod
		
		vraci: tuple (new_omega, new_sigma) nebo None pokud dalsi bod jiz neni (konec mereni)
		"""
		if self.last_coords == None:
			return None
		if self.last_coords == (None, None):
			self.last_coords = (self.omega_min.value, self.sigma_min.value)
		else:
			self.last_coords = self._next(*self.last_coords)	
		return self.last_coords

	def replot_hint(self, omega, sigma):
		np = self._next(omega, sigma)
		if np == None:
			return True
		if self.omega_min.value != np[0] or self.sigma_min.value + self.sigma_step.value >= np[1]:
			return False
		return True

	def get_extent(self):
		return Rect().set_lbrt(self.omega_min.value, self.sigma_min.value,
		                       self.omega_max.value, self.sigma_max.value)

	def validator(self):
		return validate_omega_step(self) and validate_sigma_step(self)

# MeasureTable ##############################################################

class MeasureTable(Measure):
	"""Bodova mereni - nahodne rozmistene body"""
	def __init__(self):
		Measure.__init__(self)
		self.points = ArrayModel(Point())
		self.n_points_counter = LenAdaptor(self.points, self.n_points)

	def init_on_typetest(self, typetest):
		return self.init_on_table_points(typetest.points, typetest.name.value)

	def init_on_points(self, points, title = ''):
		"""points: sekvence tuplu (x, y)"""
		self.n_points.freeze_changes()
		try:
			self.points.empty()
			for p in points:
				self.points.append(Point().set_xy(*p))
			self.reset_points()
		finally:
			self.n_points.thaw_changes()
		self.title.value = title
		return self

	def init_on_table_points(self, table_points, title = ''):
		"""table_points: instance typetests.TablePoint"""
		self.n_points.freeze_changes()
		try:
			self.points.empty()
			for p in table_points:
				self.points.append(Point().set_xy(p.omega.value, p.sigma.value))
			self.reset_points()
		finally:
			self.n_points.thaw_changes()
		self.title.value = title
		return self

	def reset_points(self):
		self._next_point = 0

	def next_point(self):
		if self._next_point >= self.n_points.value:
			return None
		p = self.points[self._next_point]
		self._next_point += 1
		return (p.x, p.y)

# MeasurePolar #######################################################

class MeasurePolar(Measure):
	"""Polarni mereni - rozsahy.

	Zde je situace mirne komplikovana kvuli moznemu parametru u SigmaSpans.
	Vyuzivam zde proto TRY_LATER, pokud je v nekterem ze sigma_spans parametr
	jiny nez 'value'.
	"""

	class PolarPointsCounter(object):
		def __init__(self, adapted_model, target):
			self.omega_spans, self.sigma_spans = adapted_model.omega_spans, adapted_model.sigma_spans
			self.target = target
			self.omega_spans.add_change_observer(self.on_model_changed)
			self.sigma_spans.add_change_observer(self.on_model_changed)

		def on_model_changed(self):
			new_cnt = reduce(lambda cnt, os: cnt + len(os), self.omega_spans, 0)
			new_cnt = reduce(lambda cnt, ss: cnt + len(ss), self.sigma_spans, new_cnt)
			self.target.value = new_cnt

	def __init__(self):
		Measure.__init__(self)
		dummy_typetest = typetests.TypetestPolar()
		self.omega_spans = dummy_typetest.omega_spans
		self.sigma_spans = dummy_typetest.sigma_spans
		self.n_points_counter = self.PolarPointsCounter(self, self.n_points)

	def init_on_spans(self, omega_spans, sigma_spans, title = ''):
		self.freeze_changes()
		try:
			self.omega_spans.copy_from(omega_spans)
			self.sigma_spans.copy_from(sigma_spans)
			self.title.value = title
			self.reset_points()
		finally:
			self.thaw_changes()
		return self

	def init_on_typetest(self, typetest):
		"""typetest: instance TypetestPolar"""
		return self.init_on_spans(typetest.omega_spans, typetest.sigma_spans, typetest.name.value)

	def reset_points(self):
		# definuji bod, ktery priste vratim.

		# Bod je urcen nasledujicimi promennymi:
		# self.current_spans: 0 nebo 1, omega_spans nebo sigma_spans 
		# self.current_span: Index rozsahu v omega_spans nebo sigma_spans
		# self.current_point: Index bodu v aktualnim rozsahu
		self.current_spans = 0
		self.current_span = 0
		self.current_point = 0

		# pro mereni sigma_spans podle parametru
		self.min_point = self.max_point = None 

	def contains_parametrized_span(self):
		"""Zjistuje, zda nektery rozsah potrebuje spoctene minimum/maximum

		Vraci True, pokud ano, False, pokud ne.
		"""
		for s in self.sigma_spans:
			if s.parameter.value != 'value':
				return True
		return False

	def get_spans(self):
		if self.current_spans == 0:
			return self.omega_spans
		elif self.current_spans == 1:
			# pokud jeste nemam spocitane minimum a maximum z namerenych hodnot
			# prvniho rozsahu, ale budu je potrebovat, spocitam si minimum a
			# maximum z namerenych hodnot prvniho radku definic pro konst.
			# sigma
			if self.min_point == None and self.contains_parametrized_span():
				spans = self.omega_spans
				if len(spans) == 0:
					log_error("Pro konstantn sigma nen definovn dn rozsah.")
					log_error("Nelze spotat maximum/minimum.")
					raise NextPointError
				n_minmax_points = len(spans[0])
				if n_minmax_points == 0:
					log_error("Pro prvn rozsah konstantnho sigma nen definovn dn bod.")
					log_error("Nelze spotat maximum/minimum.")
					raise NextPointError
				if len(self.meas_data.points) < n_minmax_points:
					return TRY_LATER
				self.min_point, self.max_point = self.meas_data.find_minmax(n_minmax_points)
			return self.sigma_spans
		else:
			return None

	def get_span(self):
		while True:
			spans = self.get_spans()
			if spans == None or spans == TRY_LATER:
				return spans

			if self.current_span < len(spans):
				return spans[self.current_span]
			else:
				self.current_spans += 1
				self.current_span = 0
				self.current_point = 0

	def get_point(self):
		while True:
			span = self.get_span()
			if span == None or span == TRY_LATER:
				return span

			if self.current_point < len(span):
				pt = span[self.current_point]
				if isinstance(span, typetests.SigmaSpan):
					if span.parameter.value == 'min':
						pt = (self.min_point.omega, pt[1])
					if span.parameter.value == 'max':
						pt = (self.max_point.omega, pt[1])
				return pt
			else:
				self.current_span += 1
				self.current_point = 0

	def next_point(self):
		"""Vrati dalsi bod
		
		vraci: tuple (new_omega, new_sigma) nebo None pokud dalsi bod jiz neni
			(konec mereni)

		"""
		pt = self.get_point()
		if pt != None and pt != TRY_LATER:
			self.current_point += 1
		return pt

# MeasureDirectional ########################################################

class MeasureDirectional(MeasureIso):
	pass

# MeasureApproach ###########################################################

class MeasureApproach(MeasureIso):
	pass

# MeasureOmni ###############################################################

class MeasureOmni(MeasurePolar):
	pass

# MeasureLight ##############################################################

class MeasureLight(MeasurePolar):
	pass

