# -*- encoding: cp1250 -*-

"""MVC nadstavba nad matplotlibovskym grafem"""

import os

import matplotlib
matplotlib.use('WX')
from matplotlib.backends.backend_wxagg import FigureCanvasWxAgg as FigureCanvasWx
from matplotlib.backends.backend_agg import FigureCanvasAgg as FigureCanvas
from matplotlib.figure import Figure
from matplotlib import ticker
from matplotlib.collections import LineCollection

import wx

from mvc import *

import settings
from misc import uni, any, format_float
import misc_num
import storage
import debug
import registry
from numpy import sin, cos
from error_logging import *

# Jedna instance FontProperties nestaci, pri kazdem text() si do toho
# matplotlib neco cpe.
def _fp(basename):
	fname = os.path.join(registry.paths.fonts, '%s.ttf' % basename)
	return matplotlib.font_manager.FontProperties(fname = fname)

view_fontprops = {
	'title_fontprops': {'fontsize': 14, 'fontproperties': _fp('verdana')},
	'label_fontprops': {'fontsize': 14, 'fontproperties': _fp('verdana')},
	'coords_fontprops': {'fontsize': 11, 'fontproperties': _fp('verdana')},
	'texts_fontprops': {'fontsize': 11, 'fontproperties': _fp('verdana')},
	'notes_fontprops': {'fontsize': 12, 'fontproperties': _fp('verdana')},
}

print_fontprops = {
	'title_fontprops': {'fontsize': 12, 'fontproperties': _fp('verdana')},
	'label_fontprops': {'fontsize': 12, 'fontproperties': _fp('verdana')},
	'coords_fontprops': {'fontsize': 10, 'fontproperties': _fp('verdana')},
	'texts_fontprops': {'fontsize': 9,  'fontproperties': _fp('verdana')},
	'notes_fontprops': {'fontsize': 10, 'fontproperties': _fp('verdana')},
}

def _str_bbox(bbox):
	return u"lbwh: " + str(bbox.get_bounds())

def _dist2(a, b):
	return (a.x - b.x)**2 + (a.y - b.y)**2

# Plottable #################################################################

class Plottable(CompositeModel):
	"""Predek objektu ktere se daji pridat do grafu."""
	_models = {
		'offset': Point(),
		'id': IntModel(), # ID jedinecne v ramci grafu
		'movable': BoolModel(False), # da se s nim hybat?
		'movegroup': StrModel(), # prislusnost do MoveGroup ('' - nepatri nikam)
	}

	_next_id = 0
	def __init__(self):
		CompositeModel.__init__(self)
		self.id.value = Plottable._next_id
		Plottable._next_id += 1

	def add_offset(self, offset):
		self.offset.add(offset)

	def add_to_plot(self, plot):
		raise NotImplementedError

	def pick(self, p, plot):
		"""Pokud pick vraci None, neni objekt picknutelny. 
		
		V opacnem pripade necht vrati druhou mocninu vzdalenosti (v
		souradnicich dat) od bodu p
		"""
		return None

	def on_right(self):
		"""Volano pri stisku praveho tlacitka na objektu"""
		pass

	def sketch(self, offset, plot):
		"""Melo by naskicovat objekt do platna.

		Ma byt rychle, takze se doporucuje pouzit primitiv wxDC. O ulozeni
		stavu DC pred skicou a obnoveni po skice uz je postarano.

		offset: offset od puvodni pozice (tj. te vcetne offsetu), kam
			se ma objekt naskicovat
		plot: instance Plot
		"""
		self._wx_sketch_lines(offset, plot, self._get_sketch_lines(plot))

	def get_description(self):
		"""Ma vratit kratkou popisku objektu (mimo jine pro statusbar)"""
		raise NotImplementedError()

	def _wx_sketch_lines(self, offset, plot, lines):
		"""

		offset: instance Point
		plot: instance Plot
		lines: [Line, ...]
		"""
		dc = wx.ClientDC(plot.get_canvas())
		dc.SetLogicalFunction(wx.XOR)            
		dc.ResetBoundingBox()
		dc.SetClippingRegion(*plot.wx_bounds().get_lbwh())
		dc.BeginDrawing()

		dc.SetLogicalFunction(wx.XOR)            
		dc.SetBrush(wx.Brush(wx.Colour(255,255,255), wx.TRANSPARENT))
		dc.SetPen(wx.Pen(wx.Colour(200, 200, 200), 1, wx.SOLID))

		o = self.offset + offset
		for l in lines:
			p1, p2 = l.get_points()
			p1.add(o); p2.add(o)
			wx_p1, wx_p2 = plot.wx_coords(p1), plot.wx_coords(p2)
			dc.DrawLine(wx_p1.x, wx_p1.y, wx_p2.x, wx_p2.y)

		dc.EndDrawing()
		dc.DestroyClippingRegion()

	def _get_sketch_lines(self, plot):
		"""Necht vrati seznam car, ktere se maji naskicovat.

		Cary necht jsou vraceny tak, jako by objekt byl v zakladni
		pozici (tj. vubec neposunut)
		"""
		raise NotImplementedError
	
# PlotPoint ####################################################################

class PlotPoint(Plottable):
	_models = {
		'pos':Point(),
		'style': StrModel('x'),
		'color': Color(),
		'markersize': FloatModel(10.0),
		'markerwidth': FloatModel(1.5),
	}
	_models.update(Plottable._models)

	def set_text_props(self, plot_text):
		"""Nastavi textu parametry tak, aby text byl u bodu."""
		plot_text.pos.copy_from(self.pos)
		plot_text.horizontalalignment.value = 'center'
		plot_text.verticalalignment.value = 'bottom'

	def add_to_plot(self, plot):
		pos = self.pos
		# viz http://matplotlib.sourceforge.net/matplotlib.lines.html
		plot.get_plot().plot((pos.x,), (pos.y,), self.style.value, 
							 markersize = self.markersize.value,
							 markerfacecolor = 'black',
							 markeredgecolor = self.color.get(),
							 markeredgewidth = self.markerwidth.value)
	
# PlotText ###############################################################

class PlotText(Plottable):
	"""Text ktery se da vlozit do grafu.  """
	_models = {
		'pos': Point(),
		'text': StrModel(),
		'color': Color(),
		'size': FloatModel(view_fontprops['texts_fontprops']['fontsize']),
		'horizontalalignment': StrModel('center'),
		'verticalalignment': StrModel('center'),
	}
	_models.update(Plottable._models)

	def __init__(self, text = '', x = 0.0, y = 0.0):
		Plottable.__init__(self)
		self.text.value = text
		self.pos.set_xy(x, y)
		self.text_artist = None # reference na naposledy kresleny matplotlib.text

	def _text_props(self):
		return {
			'color': self.color.get(),
			'size': self.size.value,
			'horizontalalignment': self.horizontalalignment.value,
			'verticalalignment': self.verticalalignment.value,
			'fontproperties': view_fontprops['texts_fontprops']['fontproperties'],
		}

	def add_to_plot(self, plot):
		pt = self.pos + self.offset
		p = plot.get_plot()
		self.text_artist = p.text(pt.x, pt.y, uni(self.text.value), self._text_props())
		return self.text_artist

	def bound(self, bounds):
		"""Pokud je text mimo bounds, je omezen do bounds."""
		self.pos.bound(bounds)

	# Veci pro D'n'D

	def pick(self, p, plot):
		renderer = plot.get_renderer()
		if not renderer:
			return None # Pred inicializaci se muze stat ze jeste nemame renderer.

		if self.text_artist == None:
			return None
		wx_bbox = self.text_artist.get_window_extent(renderer)
		bbox = Rect().set_lbwh(*(wx_bbox.get_bounds()))
		bbox.expand(2,2)
		data_bbox = plot.rect_screen2data(bbox)
		if data_bbox.contains_point(p):
			return 0
		return None
	
	def _get_sketch_lines(self, plot):
		t = self.text_artist
		if not t:
			raise RuntimeWarning(u"No text artist yet! probably a bug!")
		
		lbwh = t.get_window_extent(plot.get_renderer()).get_bounds()
		r = plot.rect_screen2data(Rect().set_lbwh(*lbwh))
		r.move(self.offset.negated())
		l, b, w, h = r.get_lbwh()
		return (Line().set(l,   b,   l+w, b  ),
		        Line().set(l+w, b,   l+w, b+h),
		        Line().set(l+w, b+h, l,   b+h),
		        Line().set(l,   b+h, l,   b  ))

	def get_description(self):
		return u"Popisek(%s)" % self.text

# PlotLines ###################################################################

class PlotLines(Plottable):
	_models = {
		'lines': ArrayModel(Line()),
		'color': Color(),
		'width': FloatModel(1.0),
		'description': StrModel(u""),
	}
	_models.update(Plottable._models)

	def __init__(self):
		Plottable.__init__(self)
		self.line_collection = None

	def is_empty(self):
		return len(self.lines) == 0

	def recompute_collection(self):
		self.line_collection = LineCollection(self.get_translated_lines())
		self.line_collection.set_color(self.color.get())
		self.line_collection.set_linewidth(self.width.value)

	def set_offset(self, o):
		assert(self.line_collection)
		if self.offset == o:
			return
		self.offset.copy_from(o)
		self.recompute_collection()

	def get_line_collection(self):
		"""Vrati instanci LineCollection podle aktualni cary.
		
		Cary jsou jiz posunuty o hodnotu nastavenou pomoci set_offset()
		"""
		if not self.line_collection:
			self.recompute_collection()
		return self.line_collection

	def get_translated_lines(self):
		""" [((x1, y1), (x2, y2)), ...]. Posunuto offsetem."""
		ox, oy = self.offset.get_xy()
		return [((l.x0 + ox, l.y0 + oy), (l.x1 + ox, l.y1 + oy)) for l in self.lines]

	def set_text_props(self, txt, position = 'top'):

		if position == 'right':
			max_x = reduce(lambda r, l: max(r, l.x0, l.x1), self.lines, None)
			txt.pos.set_xy(max_x, 0)
			txt.horizontalalignment.value = 'left'
			txt.verticalalignment.value = 'bottom'
			return

		if position == 'center':
			sums = reduce(lambda r, l: (r[0]+l.x0, r[1]+l.y0, r[2]+l.x1, r[2]+l.y1), 
			              self.lines, (0,0,0,0))
			cnt = len(self.lines)*2
			txt.pos.set_xy((sums[0]+sums[2]) / cnt, (sums[1]+sums[3]) / cnt)
			txt.horizontalalignment.value = 'center'
			txt.verticalalignment.value = 'center'
			return

		import operator
		try:
			cmp_op, valign = {
				'top': (operator.gt, 'bottom'),
				'bottom': (operator.lt, 'top')
			}[position]
		except KeyError:
			raise RuntimeWarning('set_text_props: %s not implemented'%position)

		x = 0
		y = None
		for l in self.lines:
			x += l.x0
			x += l.x1
			if y == None or cmp_op(l.y0, y):
				y = l.y0
			if y == None or cmp_op(l.y1, y):
				y = l.y1
		if self.is_empty():
			raise RuntimeWarning('len(self.lines) == 0')
			return

		x = x / (len(self.lines) * 2)

		txt.horizontalalignment.value = 'center'
		txt.verticalalignment.value = valign
		txt.pos.set_xy(x, y)
	
	def add_to_plot(self, plot):
		self.recompute_collection() 
		return plot.get_plot().add_collection(self.get_line_collection())

	# Veci pro D'n'D

	def pick(self, p, plot):
		col = self.get_line_collection()
		mindist = None
		for point in col.get_verts():
			dist = _dist2(p, Point().set_tuple(point))
			if (not mindist) or (dist < mindist):
				mindist = dist
		return mindist
	
	def _get_sketch_lines(self, plot):
		return self.lines

	def get_description(self):
		return self.description.value

# OffsetMultipleLocator ######################################################

class OffsetMultipleLocator(ticker.MultipleLocator):
	def __init__(self, base = 1.0, offset = 0.0):
		ticker.MultipleLocator.__init__(self, base)
		_, self.offset = divmod(offset, base)

	def __call__(self):
		locs = ticker.MultipleLocator.__call__(self)
		return locs + self.offset

	def autoscale(self):
		"""
		Set the view limits to the nearest multiples of base that
		contain the data
		"""

		self.verify_intervals()
		dmin, dmax = self.dataInterval.get_bounds()
		offset = self.offset

		vmin = self._base.le(dmin - offset) + offset
		vmax = self._base.ge(dmax - offset) + offset
		if vmin==vmax:
			vmin -=1
			vmax +=1

		from matplotlib.transforms import nonsingular
		return nonsingular(vmin, vmax)

# Grid #####################################################################

class Grid(CompositeModel):
	_models = {
		'enabled': BoolModel(False),
		'type': StrModel('offset'), # 'offset' nebo 'fixed'
		# pro Offset
		'xoffset': FloatModel(0.0),
		'xstep': FloatModel(1.0),
		'yoffset': FloatModel(0.0),
		'ystep': FloatModel(1.0),
		# pro Fixed
		'xlocations': ArrayModel(FloatModel(0.0)),
		'ylocations': ArrayModel(FloatModel(0.0)),
	}

	def set_grid_offset(self, enabled, xoffset, xstep, yoffset, ystep):
		for a in 'enabled', 'xoffset', 'xstep', 'yoffset', 'ystep':
			getattr(self, a).value = locals()[a]
		self.type.value = 'offset'

	def set_grid_fixed(self, enabled, xlocations, ylocations):
		"""
		enabled: bool
		xlocations, ylocations: [float, ...]
		"""
		self.enabled.value = enabled
		self.xlocations.empty()
		for l in xlocations:
			self.xlocations.append(FloatModel(l))
		self.ylocations.empty()
		for l in ylocations:
			self.ylocations.append(FloatModel(l))
		self.type.value = 'fixed'

	def set_at_plot(self, plot):
		"""Mrizka se podle svych parametru projevi na plot

		plot: instance matplotlib.Axes
		"""
		{
			'offset': self._set_at_plot_offset,
			'fixed':  self._set_at_plot_fixed,
		}[self.type.value](plot)

	def _set_at_plot_offset(self, plot):
		xloc = OffsetMultipleLocator(self.xstep.value, self.xoffset.value)
		yloc = OffsetMultipleLocator(self.ystep.value, self.yoffset.value)
		self._set_at_plot(plot, xloc, yloc)

	def _set_at_plot_fixed(self, plot):
		xloc = ticker.FixedLocator([l.value for l in self.xlocations])
		yloc = ticker.FixedLocator([l.value for l in self.ylocations])
		self._set_at_plot(plot, xloc, yloc)

	def _set_at_plot(self, plot, xloc, yloc):
		plot.get_xaxis().set_major_locator(xloc)
		plot.get_yaxis().set_major_locator(yloc)
		plot.grid(self.enabled.value)

# PlottableGroup ############################################################

class PlottableGroup(CompositeModel):
	"""Skupina plottables, ktere jsou nejakym zpusobem spolu provazane."""
	_models = {
		'name': StrModel(),
		'members': ArrayModel(IntModel()), # seznam id
	}

	def add_ids(self, ids):
		ids = list(ids)
		for m in self.members:
			if m.value in ids:
				ids.remove(m.value)
		for id in ids:
			self.members.append(IntModel(id))

	def contains_plottable(self, plottable):
		pid = plottable.id.value
		return any(m.value == pid for m in self.members)

# PlottableGroups ##########################################################

class PlottableGroups(ArrayModel):
	"""Seskupuje Plottables do skupin pro drag'n'drop.
	
	Kazdy Plottable ma atribut movegroup, kterym se zarazuje do nejake skupiny.
	Pokud zacnu tahat nejaky objekt, zaroven zacnu tahat i vsechny dalsi objekty
	ktere jsou ve stejne movegroup. Krome toho se daji PlottableGroups pouzit i
	k seskupovani Plottable za jinymi ucely.
	"""
	def __init__(self):
		ArrayModel.__init__(self, PlottableGroup())

	def find_group_by_plottable(self, plottable):
		for mg in self:
			if mg.contains_plottable(plottable):
				return mg
		return None

	def find_group(self, name):
		for mg in self:
			if mg.name.value == name:
				return mg
		return None

	def add_ids_to_group(self, ids, groupname):
		mg = self.find_group(groupname)
		if mg == None:
			mg = PlottableGroup()
			mg.name.value = groupname
			self.append(mg)
		mg.add_ids(ids)

	def add_id_to_group(self, id, groupname):
		self.add_ids_to_group([id], groupname)

	def add_plottable_to_group(self, plottable, groupname):
		self.add_id_to_group(plottable.id.value, groupname)

	def add_plottables_to_group(self, plottables, groupname):
		self.add_ids_to_group([p.id.value for p in plottables], groupname)

	def find_ids_in_group(self, groupname):
		mg = self.find_group(groupname)
		if mg == None:
			return []
		return [m.value for m in mg.members]

# PlotModel #################################################################

class PlotModel(CompositeModel):
	_models = {
		'extent': Rect(),
		'notes': ArrayModel(StrModel()),
		'mesh': storage.Mesh(),
		'plottables': ListModel([PlotPoint(), PlotText(), PlotLines()]),
		# Kazdy Plottable ma atribut movegroup, kterym se zarazuje do nejake
		# skupiny.  Pokud zacnu tahat nejaky objekt, zaroven zacnu tahat i
		# vsechny dalsi objekty ktere jsou ve stejne movegroup.
		'movegroups': PlottableGroups(),
		'title': StrModel(),
		'grid': Grid(),
		'lumi_theme': settings.LumiTheme(),
		'lumi_on': BoolModel(False),
		'colorbar_on': BoolModel(),
		'colorbar_title': StrModel(),
		'lumi_min': FloatModel(0.0),
		'lumi_max': FloatModel(0.0),
		'use_specified_lumi_range': BoolModel(False),
		'xlabel': StrModel(),
		'ylabel': StrModel(),
		'polar': BoolModel(False),
		}

	def clear(self):
		self.notes.empty()
		self.plottables.empty()
		self.movegroups.empty()

	def find_plottable_by_id(self, id):
		for p in self.plottables:
			if p.id.value == id:
				return p
		return None

	def find_id_of_plottable(self, plottable):
		for p in self.plottables:
			if p == plottable:
				return p.id.value
		return None

	def find_plottables_by_ids(self, ids):
		return [p for p in self.plottables if (p.id.value in ids)]

	def find_plottables_in_group(self, groupname):
		ids = self.movegroups.find_ids_in_group(groupname)
		return self.find_plottables_by_ids(ids)

	def find_plottables_in_group_with(self, plottable):
		"""Vraci vsechny Plottable ze skupiny plottable.movegroup

		pokud je plottable.movegroup prazdna (''), nebo neexistuje movegroup do
		ktere je pridano, vraci pouze [plottable].
		"""
		groupname = plottable.movegroup.value
		if groupname == '':
			return [plottable]
		plottables = self.find_plottables_in_group(groupname)
		if not len(plottables):
			return [plottable]
		return plottables

	def delete_plottables_with_ids(self, ids):
		indices = []
		for idx, p in enumerate(self.plottables):
			if p.id.value in ids:
				indices.append(idx)
		self.plottables.pop_indices(indices)

# PlotView ##################################################################

class PlotView(View):

	#-- inicializace
	def __init__(self, model, frame, resource, plotctrl_id):
		View.__init__(self, model, frame)
		self.fig = Figure(figsize=(0.1,0.1))
		self.canvas = FigureCanvasWx(frame, -1, self.fig)
		self.attach_window(resource, plotctrl_id, self.canvas)
		self.subplot = None

	def update(self):
		self.plot()

	#-- pristup k matplotlib atributum

	def get_renderer(self):
		"""Vrati instanci matplotlib.Renderer"""
		return self.canvas.get_renderer()

	def get_plot(self):
		"""Vrati instanci matplotlib.Axes kam se kresli graf"""
		return self.subplot

	def get_canvas(self):
		"""Vrati instanci matplotlib.FigureCanvas kam se kresli graf"""
		return self.canvas

	def get_statusbar(self):
		if not hasattr(self.frame, 'GetStatusBar'):
			return # wx.Dialog nema fci GetStatusBar
		statusbar = self.frame.GetStatusBar()
		if not statusbar:
			return None
		return statusbar

	def set_statusbar_text(self, txt):
		sb = self.get_statusbar()
		if not sb:
			return
		sb.SetStatusText(txt)

	#-- Vykreslovani

	def print_figure(self, fname, **kwargs):
		"""Ulozi obrazek do souboru.

		kwargs se dale predavaji FigureCanvasWX.print_figure(), dulezite jsou:
		dpi (=150)
		Podle pripony fname se urci, v jakem formatu se vysledek ulozi. V
		pripade nepodporovane pripony hazi IOError. Pokud soubor se zadanym
		jmenem priponu nema, pouzije se .png
		"""
		fig = Figure()
		canvas = FigureCanvas(fig)
		self.plot(fig, print_fontprops)
		canvas.print_figure(fname, **kwargs)

	def plot(self, figure = None, fontprops = view_fontprops):
		"""Vykresli graf.
		
		Nejprve smaze vse co mohlo byt ve starem grafu, a pote poctive vse
		nakresli znovu. 
		figure: Pokud neni zadano, vykresluje do self.figure, jinak do figure.
		sticky_notes_fontdict: je mozne zadat parametry pisma sticky notes
		notes_fontdict: parametry pisma doplnujicich poznamek (notes).
			viz matplotlib.text
		"""
		if self.model.polar.value:
			self.plot_polar(figure, fontprops)
		else:
			self.plot_cartesian(figure, fontprops)

	def plot_polar(self, figure = None, fontprops = view_fontprops):
		"""Vykresli graf v polarnich souradnicich
		
		pouziva stale kartezsky graf, predpoklada ze data jsou jiz prevedena
		jakoby do polarnich souradnic, neumoznuje tedy napriklad smysluplne
		hybani carami. Pouzil bych matplotlib.PolarAxes, ale jsou pekne
		nedodelany, bohuzel se v tom v aktualnim stavu (0.87.3) neda udelat
		seriozni polarni graf.

		z model.extent se pouziva pouze souradnice h, ktera rika, jaka je
		souradnice r (polomer) na okraji grafu. Podle ni se pak vykresli
		kruhova osa. Jinak predpokladam, ze vsechna data padnou do kartezskeho
		rozsahu -1..1 pro osy x i y (komplikuje praci na strane naplnovani
		daty, ale co se da delat :-(
		"""

		subplot_placement = (0.01, 0.01, 0.98, 0.9)

		if figure == None:
			fig = self.fig
		else:
			fig = figure
		fig.clear()
		if figure == None:
			self.subplot = self.fig.add_axes(subplot_placement)
		else:
			self.subplot = fig.add_axes(subplot_placement)
		plot = self.subplot

		plot.set_aspect('normal')
		plot.set_xlim((-1, 1), emit = False)
		plot.set_ylim((-1, 1), emit = False)
		plot.set_title(uni(self.model.title.value), **fontprops['title_fontprops'])
		plot.set_xlabel(self.model.xlabel.value, **fontprops['label_fontprops'])
		plot.set_ylabel(self.model.ylabel.value, **fontprops['label_fontprops'])
		plot.get_xaxis().set_major_locator(ticker.FixedLocator([]))
		plot.get_yaxis().set_major_locator(ticker.FixedLocator([]))
		plot.grid(False)

		graph_r = 0.9 # polomer grafu v pomeru k stranam plochy pro graf

		# vykreslim hlavni osy

		main_axes = LineCollection([((0, -graph_r), (0, graph_r)), ((-graph_r, 0), (graph_r,0))])
		main_axes.set_color('black')
		main_axes.set_linewidth(1)
		plot.add_collection(main_axes)

		def p2c(t, r): # polar to cartesian
			t = misc_num.deg2rad(t)
			return (r*cos(t), r*sin(t))

		# vykreslim paprskove osy

		radials = [x * 22.5 for x in xrange(int(360/22.5))]
		radial_axes = LineCollection([((0, 0), p2c(t, graph_r)) for t in radials])
		radial_axes.set_color('black')
		radial_axes.set_linewidth(1)
		radial_axes.set_linestyle('dotted')
		plot.add_collection(radial_axes)

		for t in radials:
			x, y = p2c(t, graph_r * 1.05) 
			plot.text(x, y, u'%.1f' % t, ha='center', va='center', **fontprops['notes_fontprops'])

		# vykreslim kruhove osy

		radius = self.model.extent.h
		r = graph_r
		for i in xrange(5):
			axis = LineCollection([(p2c(t, r), p2c(t + 2, r)) for t in xrange(0, 360, 2)])
			axis.set_color('black')
			axis.set_linewidth(1)
			axis.set_linestyle('dotted')
			plot.add_collection(axis)

			x, y = p2c(90, r) 
			plot.text(x, y, u' %.0f'%(r*radius), ha='left', va='center', **fontprops['notes_fontprops'])

			r -= 0.2

		# dokreslim zbytek grafu

		for p in self.model.plottables:
			p.add_to_plot(self)
				
		plot.text(0.01, 0.99,
		          uni('\n'.join(n.value for n in self.model.notes)),
	              horizontalalignment = 'left', verticalalignment = 'top', 
	              transform = plot.transAxes,
				  **fontprops['notes_fontprops'])
		         
		self.refresh()

	def plot_cartesian(self, figure = None, fontprops = view_fontprops):
		"""Vykresli graf v kartezskych souradnicich"""

		m = self.model

		def set_extent():
			e = m.extent.copy()
			if e.w < 0.001:
				e.w = 1.0
			if e.h < 0.001:
				e.h = 1.0
			l, b, r, t = e.get_lbrt()
			plot.set_xlim((l, r), emit = False)
			plot.set_ylim((b, t), emit = False)
			plot.set_aspect('normal')

		#tyhle hodnoty jsem pracne hledal tak, aby v tisku bylo vsechno videt
		#a zaroven byl okraj okolo grafu co nejuzsi.
		subplot_without_colorbar_placement = ((0.1, 0.082, 0.85, 0.86))
		subplot_with_colorbar_placement = ((0.1, 0.082, 0.7, 0.86))
		colorbar_placement = ((0.82, 0.082, 0.03, 0.86))

		if m.colorbar_on.value and m.lumi_on.value:
			subplot_placement = subplot_with_colorbar_placement 
		else:
			subplot_placement = subplot_without_colorbar_placement

		if figure == None:
			fig = self.fig
		else:
			fig = figure
		fig.clear()
		if figure == None:
			self.subplot = self.fig.add_axes(subplot_placement)
		else:
			self.subplot = fig.add_axes(subplot_placement)
		plot = self.subplot

		mesh = m.mesh

		#Zde to musi byt kvuli sticky sticky_notes, ktere potrebuji znat 
		#hranice grafu.
		set_extent()

		graph_notes = []

		while m.lumi_on.value: # pouzito jako if s breakama.
			cmap = m.lumi_theme.get_colormap()
			if mesh.is_empty():
				graph_notes.append(u"Mka je przdn.")
				break
			if mesh.contains_invalid_point():
				graph_notes.append(u"Mka obsahuje body s nekonenou hodnotou.")
				break

			if (m.use_specified_lumi_range.value and m.lumi_max.value - m.lumi_min.value > 0.1):
				kwargs = {'vmin': m.lumi_min.value, 'vmax': m.lumi_max.value}
			else:
				kwargs = {}
			lumi_axes = plot.imshow(mesh.mesh.value, 
				origin = 'lower', extent = mesh.get_extent().get_lrbt(),
				interpolation = 'bicubic', cmap = cmap, **kwargs)

			if m.colorbar_on.value and lumi_axes != None:
				cbaxes = fig.add_axes(colorbar_placement)
				fig.colorbar(lumi_axes, format = '%.0f cd', cax = cbaxes)
				cbaxes.set_title(uni(m.colorbar_title.value), horizontalalignment = 'center', 
				                 **fontprops['notes_fontprops'])
			break

		m.grid.set_at_plot(plot)
		plot.set_axisbelow(True) # mrizka pod ostatnimi
		class FloatFormatter(ticker.Formatter):
			def __call__(self, x, pos = None):
				return format_float(x, 3)
		formatter = FloatFormatter()
		plot.get_xaxis().set_major_formatter(formatter)
		plot.get_yaxis().set_major_formatter(formatter)

		for p in m.plottables:
			p.add_to_plot(self)
				
		plot.text(0.01, 0.99,
		          uni('\n'.join(graph_notes + [n.value for n in m.notes])),
	              horizontalalignment = 'left', verticalalignment = 'top', 
	              transform = plot.transAxes,
		          **fontprops['notes_fontprops'])

		# No a zde to musi byt podruhe, graf si zrejme dynamicky prepocitava
		# autorange po plot.plot()  (?)
		set_extent()

		plot.set_title(uni(m.title.value), **fontprops['title_fontprops'])
		plot.set_xlabel(m.xlabel.value, **fontprops['label_fontprops'])
		plot.set_ylabel(m.ylabel.value, **fontprops['label_fontprops'])

		self.refresh()

	def refresh(self):
		"""Prekresli graf"""
		self.canvas.Refresh()

	#-- Prevody souradnych soustav

	def wx_coords(self, p):
		"""Prevede souradnice x,y v datech na souradnice x, y pro wx funkce

		p: Point
		vraci: Point"""
		data_p = self.data2screen(p)
		figheight = self.get_canvas().figure.bbox.height()
		data_p.y = figheight - data_p.y
		return data_p
	
	def wx_bounds(self):
		"""Vraci obdelnik (l, b, w, h) v souradne soustave pro wx funkce"""
		l, b, w, h = self.get_plot().bbox.get_bounds()
		figheight = self.get_canvas().figure.bbox.height()
		return Rect().set_lbwh(l, figheight - b - h, w, h)

	def data2screen(self, p):
		t = self.subplot.transData
		return Point().set_xy(*t.xy_tup(p.get_xy()))

	def screen2data(self, p):
		t = self.subplot.transData
		return Point().set_xy(*t.inverse_xy_tup(p.get_xy()))

	def rect_screen2data(self, rect):
		t = self.subplot.transData
		l, b, w, h = rect.get_lbwh()
		dl, db = t.inverse_xy_tup((l, b))
		dr, dt = t.inverse_xy_tup((l + w, b + h))
		return Rect().set_lbwh(dl, db, dr - dl, dt - db)

	def datadist2screendist(self, dist):
		p0 = self.data2screen(Point().set(0.0, 0.0))
		po = self.data2screen(Point().set(dist, dist))
		return po - p0

	def screendist2datadist(self, dist):
		p0 = self.screen2data(0, 0)
		po = self.screen2data(*dist)
		return (po[0] - p0[0], po[1] - p0[1])

# PlotController #############################################################

class PlotController(Controller):
	class Callbacks(object):
		def __init__(self):
			self.id_callbacks = {}
			self.movegroup_callbacks = {}
			self.default_callback = None

		def register_default(self, callback):
			self.default_callback = callback

		def register_id(self, id, callback):
			assert type(id) == int
			if self.id_callbacks.has_key(id):
				msgbox_warn('register_id: %i already registered' % id)
			self.id_callbacks[id] = callback

		def register_movegroup(self, name, callback):
			assert type(name) == str
			if self.movegroup_callbacks.has_key(name):
				msgbox_warn('register_movegroup: %i already registered' % name)
			self.movegroup_callbacks[name] = callback

		def invoke(self, plottable, **kwargs):
			if plottable == None:
				return self.invoke_callback(self.default_callback, **kwargs)
			id, mg = plottable.id.value, plottable.movegroup.value
			kwargs['plottable'] = plottable
			if self.id_callbacks.has_key(id):
				return self.invoke_callback(self.id_callbacks[id], **kwargs)
			elif self.movegroup_callbacks.has_key(mg):
				return self.invoke_callback(self.movegroup_callbacks[mg], **kwargs)
			else:
				return self.invoke_callback(self.default_callback, **kwargs)

		def invoke_callback(self, callback, **kwargs):
			if callback == None:
				return None
			return callback(**kwargs)

	def __init__(self, model, view, xlabel = 'x', ylabel = 'y'):
		"""

		xlabel: Nazev jednotky xove osy (nebo uhlu v polarnim grafu)
		ylabel: Nazev jednotky yove osy (nebo polomeru v polarnim grafu)
		labely jsou pro popisky ve *statusbaru*, nikoliv na osach grafu!
		"""
		Controller.__init__(self, model, view)
		self.dragged = None #Co prave tahnu (pro drag'n'drop)
		self.xlabel, self.ylabel = xlabel, ylabel
		canvas = self.view.get_canvas()
		canvas.mpl_connect('motion_notify_event', self.on_mouse_move)
		canvas.mpl_connect('button_press_event', self.on_button_press)

		#-- public promenne
		# vyvolan pri ziskavani popisku pri mouseover
		self.cb_get_plottable_description = self.Callbacks()
		self.cb_get_plottable_description.register_default(self.default_cb_get_plottable_description)

		# vyvolan pri ziskavani popisku pri mouseover tazeneho Plottable
		self.cb_get_dragged_plottable_description = self.Callbacks()
		self.cb_get_dragged_plottable_description.register_default(self.default_cb_get_dragged_plottable_description)

		# vyvolan pri pusteni objektu
		self.cb_plottable_dropped = self.Callbacks()
		self.cb_plottable_dropped.register_default(self.default_cb_plottable_dropped)

		# vyvolan pri kliknutim pravym tlacitkem
		self.cb_on_right = self.Callbacks()
		self.cb_on_right.register_default(self.default_cb_on_right)

	#-- defaultni callbacky

	def default_cb_get_plottable_description(self, plottable, coords):
		pd = self.plottable_description(plottable, offset = None)
		coords_str = self.get_coords_str(coords)
		txt = '; '.join((coords_str, pd))
		if plottable.movable.value:
			return txt + "; Lev: posun"
		return txt

	def default_cb_get_dragged_plottable_description(self, plottable, coords, offset):
		coords_str = self.get_coords_str(coords)
		pd = self.plottable_description(plottable, offset = offset)
		return '; '.join((coords_str, pd))

	def default_cb_plottable_dropped(self, plottable, offset):
		for d in self.model.find_plottables_in_group_with(plottable):
			d.add_offset(offset)
		return True

	def default_cb_on_right(self, coords = None, plottable = None):
		pass

	#-- Dragged 

	class Dragged(object):
		"""Reprezentuje prave tazeny objekt a pomocne promenne"""
		__slots__ = ('dragged', 'coords', 'backup_dc')
		def __init__(self, dragged, coords, backup_dc):
			""" Konstruktor.

			dragged: [Plottable, ...]
			coords: Point, jaka byla pozice mysi kdyz jsem zacal tahat
			backup_dc: wx.DC, zalozni kontext.
			"""
			self.dragged = dragged
			self.coords = coords.copy()
			self.backup_dc = backup_dc
	
	#-- mysi handlery

	def on_mouse_move(self, evt):
		"""Stara se o drag'n'drop a informace ve statusbaru"""
		plot = self.view.get_plot()
		if not plot:
			return
		in_plot = (evt.inaxes == plot)
		coords = Point().set_xy(evt.xdata, evt.ydata)
		if self.dragged == None:
			if not in_plot: # mimo graf
				self.view.set_statusbar_text('')
				return

			o = self.pick(coords)
			if not o: # v grafu, ale mimo vsechny objekty
				self.view.set_statusbar_text(self.get_coords_str(coords))
				return

			# v grafu a nad nejakym objektem
			txt = self.cb_get_plottable_description.invoke(o, coords = coords)
			self.view.set_statusbar_text(txt)
			return

		# Zrovna tahnu nejaky objekt
		if not in_plot:
			wx.SetCursor(wx.StockCursor(wx.CURSOR_NO_ENTRY))
			return 
		wx.SetCursor(wx.StockCursor(wx.CURSOR_HAND))

		offset = coords - self.dragged.coords
		self.sketch_dragged_objects(offset)

		pd = self.cb_get_dragged_plottable_description.invoke(
		          self.dragged.dragged, coords = coords, offset = offset)
		self.view.set_statusbar_text('; '.join((pd, u"Lev: pustit, Prav: zruit")))

	def on_button_press(self, evt):
		"""Stara se o D'n'D a stisky tlacitek vubec"""
		a = evt.inaxes
		if a != self.view.get_plot():
			return
		
		coords = Point().set_xy(evt.xdata, evt.ydata)
		dragged = self.dragged
		if dragged != None:
			if evt.button == 1: # Pusteno
				offset = coords - dragged.coords
				self.model.freeze_changes()
				if False != self.cb_plottable_dropped.invoke(dragged.dragged, offset = offset):
					self.dragged = None
				self.model.thaw_changes()
				wx.SetCursor(wx.StockCursor(wx.CURSOR_ARROW))
				return

			elif evt.button == 3: # Pusteno na puvodni pozici
				self.dragged = None
				self.model.changed() # Prekresli graf 
				wx.SetCursor(wx.StockCursor(wx.CURSOR_ARROW))
				return

		else: # Nic zrovna netahnu
			o = self.pick(coords)
			if evt.button == 1 and o and o.movable.value:
				backup_dc = self.create_backup_dc()
				self.dragged = self.Dragged(o, coords, backup_dc)
				wx.SetCursor(wx.StockCursor(wx.CURSOR_HAND))
				self.sketch_dragged_objects(Point().set_xy(0, 0))
			elif evt.button == 3:
				self.cb_on_right.invoke(o, coords = coords)

	#-- pick

	def pick(self, coords):
		p0 = Point().set_xy(0, 0)
		p1 = Point().set_xy(10, 10)
		epsilon = _dist2(self.view.screen2data(p0), self.view.screen2data(p1))
		mindist = None
		for p in self.model.plottables:
			d = p.pick(coords, self.view)
			if d != None and (mindist == None or d < mindist):
				mindist = d
				picked = p
		if mindist != None and mindist <= epsilon:
			return picked
		return None

	#-- skicovani objektu

	def create_backup_dc(self):
		"""Vytvori novy DC a nakopiruje tam akt. graf"""
		l, b, w, h = self.view.wx_bounds().get_lbwh()
		backup_bitmap = wx.EmptyBitmap(w, h)
		backup_dc = wx.MemoryDC()
		backup_dc.SelectObject(backup_bitmap)
		backup_dc.Blit(0, 0, w, h, wx.ClientDC(self.view.get_canvas()), l, b)
		return backup_dc

	def blit_backup_dc(self, target_dc, backup_dc):
		"""blitne backup_dc do target_dc"""
		l, b, w, h = self.view.wx_bounds().get_lbwh()
		target_dc.Blit(l, b, w, h, backup_dc, 0, 0)

	def sketch_dragged_objects(self, offset):
		"""Naskicuje prave tazeny objekt.
		
		offset: instance Point"""
		canvas = self.view.get_canvas()
		dc = wx.ClientDC(canvas)
		self.blit_backup_dc(dc, self.dragged.backup_dc)
		for d in self.model.find_plottables_in_group_with(self.dragged.dragged):
			d.sketch(offset, self.view)

	#-- popisky objektu kvuli statusbaru

	@staticmethod
	def plottable_description(o, offset = None):
		"""Vrati popis objektu i s pripadnym popisem posunuti.
		
		o: instance Plottable
		offset: Point(), pripadne dalsi posunuti k objektu (pricita se k o.offset)
		"""
		descr = o.get_description()
		if offset == None:
			total_offset = o.offset
		else:
			total_offset = offset + o.offset
		ox, oy = total_offset.get_xy()
		if misc_num.vec_equals((ox, oy), (0.0, 0.0), 0.001):
			return descr
		return u"; ".join((descr, u"Posunut: %.3f, %.3f" % (ox, oy)))

	def get_coords_str(self, coords):
		x, y = coords.get_xy()
		ff = format_float
		if self.model.polar.value:
			t = misc_num.rad2deg(misc_num.atan2(y, x))
			r = self.model.extent.h * misc_num.vec_len((x, y))
			return '%s = %s, %s = %s' % (self.xlabel, ff(t, 3), self.ylabel, ff(r, 3))
		else:
			i = self.model.mesh.get_val(x, y)
			if abs(i < 10):
				i_str = ff(i, 3)
			else:
				i_str = '%.0f'%i
			return '%s = %s, %s = %s, I = %s' % (self.xlabel, ff(x, 3), self.ylabel, ff(y, 3), i_str)
