# -*- encoding: cp1250 -*-
"""Modul pro reprezentaci dat mereni"""

import wx
import numpy

from mvc import *
import misc, misc_num
from error_logging import *

import debug

dtype = 'd' # Typecode numpy poli pouzivanych pro vypocty
epsilon = 0.005 # Max. odchylka kazde jedne souradnice

# TODO: V ascii reportu je uveden soubor s merenymi daty: neznamy

# SensorPosition ############################################################

class SensorPosition(CompositeModel):
	"""Trida popisujici pozici cidla"""
	_models = {
		'description': StrModel("Popis"),
		'x': FloatModel(0.0),
		'y': FloatModel(0.0),
		'z': FloatModel(0.0),
	}
	_models_order = ('description', 'x', 'y', 'z')

	def validator(self):
		for m in 'x', 'y', 'z':
			model = getattr(self, m)
			if not -100 <= model.value <= 100:
				msg = "Hodnota mus bt v rozsahu -100 .. 100 m."
				self.validator_failure(msg, model)
				return

	def get_xyz(self):
		return (self.x.value, self.y.value, self.z.value)

	def set_xyz(self, x, y, z):
		self.freeze_changes()
		self.x.value, self.y.value, self.z.value = x, y, z
		self.thaw_changes()
		return self

	def as_numarray(self):
		return numpy.array((self.x.value, self.y.value, self.z.value), dtype)

# PositionModel #############################################################

class PositionModel(FloatContainerModel):
	"""Trida popisujici obecnou polohu v prostoru"""
	_float_attrs = ('x', 'y', 'z')

	def _get_values(self):
		return (self.x, self.y, self.z)
	values = property(_get_values)

	def set_xyz(self, x, y, z):
		self.x, self.y, self.z = x, y, z
		self.changed()
		return self

	def as_numarray(self):
		return numpy.array((self.x, self.y, self.z), dtype)

# MeasuredPoint #############################################################

class MeasuredPoint(FloatContainerModel):
	"""Jeden bod mereni

	zahrnuje jak velicinu v souradnicich stolu
	natoceni/naklopeni/luxmetr (rot, tilt, illuminance[lux]), 
	tak velicinu v souradnicich svetla
	omega/sigma/intensity[cd]

	navic jeste obsahuje dve souradnice omega_def, sigma_def, coz jsou
	souradnice, kde jsem *chtel* namerit. To ze se lisi od omega, sigma, nemusi
	nic znamenat, pokud je matematika v poradku, v nekterych krajnejsich
	pripadech mohou vyustit ruzne hodnoty omega, sigma do fyzicky stejne polohy
	stolu (vlastnost souradne soustavy).
	"""
	_float_attrs = ('rot', 'tilt', 'arm', 'illuminance', 
	                'omega', 'sigma', 'intensity',
					'omega_def', 'sigma_def')

	def get_osi(self):
		return self.omega, self.sigma, self.intensity

	def get_osi_def(self):
		return self.omega_def, self.sigma_def, self.intensity

	def set_osi(self, omega, sigma, intensity):
		self.omega, self.sigma, self.intensity = omega, sigma, intensity
		self.changed()
		return self

	def get_rtai(self):
		return self.rot, self.tilt, self.arm, self.illuminance

	def set_rtai(self, rot, tilt, arm, illuminance):
		self.rot, self.tilt, self.arm, self.illuminance = rot, tilt, arm, illuminance
		self.changed()
		return self

# MeasInfo ###################################################################

class MeasInfo(CompositeModel):
	"""Kontejner pro obecn informace o men.
	
	Vsechny uhly jsou ve stupnich, vsechny delky v metrech.
	"""
	# meas_type
	NOWEDGE = 0
	WEDGE = 1
	ARM = 2
	
	_models = {
		'version': VersionModel('1.0'),
		# casova znamka
		'timestamp': TimeModel(), # Zacatek mereni
		# popisky
		'person': StrModel(u"Mi"),
		'description': StrModel(u"Popis men"),
		'bulb1': StrModel(u"rovka 1"),
		'bulb2': StrModel(u"rovka 2"),
		# zpusob mereni
		# Asi by bylo hezci mit tu StrEnumModel, ale takhle mi to zadarmo
		# funguje s mvc.BookEditor.
		'meas_type': IntModel(), # 0..bez pripravku, 1..s pripravkem, 2..s ramenem
		'photometric_system': StrModel('C-gamma'),   # nebo A-alfa
		# ckdm
		'ckdm': StrModel(u"ID dc jednotky"),
		# luxmetr
		'luxmeter': StrModel(u"ID luxmetru"),
		# stul
		'table': StrModel(u"ID stolu"),
		'table_rot_zero': FloatModel(), # Nula stolu
		'table_tilt_zero': FloatModel(),
		'light_rot_zero': FloatModel(), # Poloha stolu pri svetle v nulove poloze
		'light_tilt_zero': FloatModel(),# (pripocitava se k nule stolu)
		'f': PositionModel(), # Poloha ohniska vzhledem k upinaci plose
		# rameno
		'arm': StrModel(u"ID ramene"),
		'arm_zero': FloatModel(),     # Nulova poloha ramene
		# cidlo (nema smysl u mereni s ramenem)
		'sensor': SensorPosition(),
		}
	_models_order = ('version', 'timestamp', 'person', 'description', 'bulb1', 'bulb2', 'meas_type',
	                 'photometric_system', 'ckdm', 'luxmeter', 'table', 'table_rot_zero', 
					 'table_tilt_zero', 'light_rot_zero', 'light_tilt_zero', 'arm', 'arm_zero', 'sensor')

	def ascii_report(self):
		"""Vrati ascii protokol o informacich o mereni

		protokol je str, radky popisujici jednotlive parametry meas_info
		"""
		return u"\n".join((
			u"Datum men: %s" % str(self.timestamp),
			u"Mi: %s" % self.person.value,
			u"Popis: %s" % self.description.value,
			u"Nulov poloha stolu: natoen = %.3f deg, naklopen = %.3f deg" % \
				(self.table_rot_zero.value, self.table_tilt_zero.value),
			u"Poloha ohniska: x = %.3f m, y = %.3f m, z = %.3f m" % self.f.values,
			u"Poloha stolu pi startu men: natoen = %.3f deg, naklopen = %.3f deg" % \
				(self.light_rot_zero.value, self.light_tilt_zero.value),
			u"Poloha idla: x = %.3f m, y = %.3f m, z = %.3f m" % self.sensor.get_xyz(),
			u"Mic stl: " + self.table.value,
			u"Luxmetr: " + self.luxmeter.value,
			u"Rameno: " + self.arm.value,
			u"dc jednotka: " + self.ckdm.value,
			))

	def validator(self):
		return # Neplatne hodnoty budou matikami ignorovany.
		if self.meas_type.value == self.NOWEDGE:
			return
		if self.meas_type.value == self.WEDGE:
			mt = u"Pi men s ppravkem"
		else:
			mt = u"Pi men s ramenem"
		if self.light_tilt_zero.value != 0.0:
			self.validator_failure(u"%s mus bt naklopen stolu nulov" % mt, self.light_tilt_zero)
		if self.f.x != 0.0:
			self.validator_failure(u"%s mus bt x-ov poloha ohniska nulov" % mt, self.f.float_model('x'))
		if self.f.y != 0.0:
			self.validator_failure(u"%s mus bt y-ov poloha ohniska nulov" % mt, self.f.float_model('y'))

	def html_report(self):
		from registry import appname
		return u"""
			<b>Datum men: </b>%s<br>
			<b>Poznmka: </b>%s<br>
			<p>
			<b>Mic stl: </b>%s<br>
			<b>Luxmetr: </b>%s<br>
			<b>Rameno: </b>%s<br>
			<b>dc jednotka: </b>%s<br>
			<b>Software: </b>%s<br>
			<p>
			<b>rovka 1: </b>%s<br>
			<b>rovka 2: </b>%s<br>
			""" % (
				misc.native2html(str(self.timestamp)),
				misc.native2html(self.description.value),

				misc.native2html(self.table.value),
				misc.native2html(self.luxmeter.value),
				misc.native2html(self.arm.value),
				misc.native2html(self.ckdm.value),
				misc.native2html(appname),

				misc.native2html(self.bulb1.value),
				misc.native2html(self.bulb2.value),
			) 

# MeasData ##################################################################

class MeasData(CompositeModel):
	"""Kontejner namerenych dat"""
	_models = {
		'meas_info': MeasInfo(),
		'points': ArrayModel(MeasuredPoint()),
	}
	_models_order = ('meas_info', 'points')

	def extend(self, other_meas_data):
		self.points.extend(other_meas_data.points)

	def empty(self):
		"""Vyprazdni seznam odmerenych bodu."""
		self.points.empty()

	def is_empty(self):
		return len(self.points) == 0

	def as_mesh(self):
		"""Vytvori z dat Mesh

		viz Mesh.read_from_meas_data
		"""
		try:
			m = Mesh()
			m.read_from_meas_data(self)
		except MeshError:
			return None
		return m

	def ascii_report_rtai(self):
		""" Vrati ascii report, radky popisujici jednotliva namerena data
		"""
		def str_point(p):
			return "%9.3f %14.3f %12.3f %14.3f" % p.get_rtai()

		return u"\n".join((
			u"Namen data v souadnicch stolu.",
			u"",
			u"natoen[deg] naklopen[deg] rameno[deg] osvtlenost",
			u"\n".join([str_point(p) for p in self.points])
			))

	def ascii_report_osi(self):
		"""Vrati ascii report, radky popisujici jednotlive namerene hodnoty"""
		def str_point(p):
			return u"%8.3f %11.3f %10.3f" % p.get_osi_def()

		return u"\n".join((
			u"Namen data v souadnicch svtla",
			u"",
			u"omega[deg] sigma[deg] svtivost[cd]",
			u"\n".join([str_point(p) for p in self.points])
			))

	def ascii_report_body(self, include_data = True, include_translated_data = True):
		"""Vytvori protokol o merenych datech pro ascii export a vrati jej jako str.

		include_data: Vcetne hodnot namerenych dat (natoceni a naklopeni stolu)
		include_translated_data: Vcetne hodnot namerenych dat (omega, sigma)
		"""
		if not include_data and not include_translated_data:
			return u""
		if include_data and include_translated_data:
			return u"\n\n".join((self.ascii_report_rtai(), self.ascii_report_osi()))
		if include_data:
			self.ascii_report_rtai()
		if include_translated_data:
			self.ascii_report_osi()

	def ascii_report_header(self):
		"""Hlavicka ascii reportu"""
		from registry import appname
		sf = self.source_file.value
		return u"\n".join((
			u"Soubor s daty men: " + misc.ternary(sf, sf, u"neznm"),
			u"",
			self.meas_info.ascii_report(),
			u"Software: " + appname,
			))

	def ascii_report(self, include_data = True, include_translated_data = True):
		return '\n\n'.join((self.ascii_report_header(), 
		                    self.ascii_report_body(include_data, include_translated_data)))

	def find_minmax(self, n_points):
		"""V prvnich n_points bodech nalezne bod s nejmensi a nejvetsi intenzitou.

		Vraci tuple (MeasuredPoint, MeasuredPoint), tedy (min, max)
		"""
		min_idx = max_idx = 0
		mp = self.points
		for p_idx in xrange(1, n_points):
			p = mp[p_idx]
			if p.intensity < mp[min_idx].intensity:
				min_idx = p_idx
			if p.intensity > mp[max_idx].intensity:
				max_idx = p_idx
		return (mp[min_idx], mp[max_idx])

	def find_point(self, omega, sigma):
		"""Nalezne bod, ktery je od daneho omega, sigma blize nez epsilon.

		Vraci MeasuredPoint, nebo None pokud se nepodarilo zadny najit.
		"""
		maxdist2 = epsilon**2
		for mp in self.points:
			dist2 = (mp.omega_def - omega)**2 + (mp.sigma_def - sigma)**2
			if dist2 <= maxdist2:
				return mp
		return None

# Mesh #######################################################################

class MeshError(Exception):
	"""Chyba pri prevodu dat do mrizky"""
	pass

class MeshScaleError(Exception):
	"""Chyba pri zmene velikosti mrizky"""
	pass

class Mesh(CompositeModel):
	"""Kontejner pro data usporadana do mrizky.
	
	Pokud jsou data omega/sigma/intenzita, je konvence takova,
	ze omegy se meni po sloupcich a sigmy po radcich.
	"""
	_models = {
		'rows': NumpyArray(dtype = dtype), # 1D, pozice radku
		'cols': NumpyArray(dtype = dtype), # 1D, pozice sloupcu
		'mesh': NumpyArray(dtype = dtype), # [rows][cols]
	}
	_models_order = ('rows', 'cols', 'mesh')

	def __str__(self):
		return "\n".join((
			"Mesh rows: " + str(self.rows), 
			"Mesh cols: " + str(self.cols),
			str(self.mesh) ))

	def _get_regular_array_step(self, array, arrname):
		if len(array) < 2:
			warn(u"Pole %s m mn ne dv hodnoty." % arrname)
			return 1.0
		return (array[-1] - array[0]) / (len(array) - 1)

	def get_omega_step(self):
		return self._get_regular_array_step(self.cols.value, 'cols')

	def get_sigma_step(self):
		return self._get_regular_array_step(self.rows.value, 'rows')

	def is_empty(self):
		return len(self.rows.value) == 0 or len(self.cols.value) == 0
	
	def find_minmax(self):
		return self.mesh.find_minmax()

	def point_coords(self, col, row):
		"""vraci (omega, sigma) souradnice bodu [col, row]"""
		return (self.cols.value[col], self.rows.value[row])

	def get_extent(self):
		r = self.rows.value
		c = self.cols.value
		if len(r) < 2 or len(c) < 2:
			return Rect().set_lbwh(0.,0.,0.,0.)
		return Rect().set_lrbt(c[0], c[-1], r[0], r[-1])

	def contains_rect(self, rect, epsilon = 0.0):
		"""Ujisti se ze bounds(instance Rectangle) je v mesh.
		
		Vraci seznam hlasek, kde vsude je mesh presazeno rect-em. Pokud je
		seznam prazdny, je cely rect uvnitr mesh.
		"""
		extent = self.get_extent().expand(epsilon, epsilon)
		crossings = extent.contains_rect(rect)
		overlaps = []
		if crossings != True:
			for c in crossings:
				overlaps.append(u"Pesah na hranici: %s" % 
						{
							'left': u"omega min.", 'right': u"omega max.", 
							'bottom': u"sigma min.", 'top': u"sigma max.",
						}[c])
		return overlaps

	def has_same_points_as(self, other_mesh, epsilon = 0.0):
		"""Test, zda rows a cols teto mrizky a other_mesh jsou stejne

		epsilon: maximalni povolena odchylka 
		vraci: True/False
		"""
		return (misc_num.vec_equals(self.rows.value, other_mesh.rows.value, epsilon) and
		        misc_num.vec_equals(self.cols.value, other_mesh.cols.value, epsilon))

	def multiply(self, x):
		"""Hodnotu kazdeho bodu mrizky vynasobi skalarem x."""
		self.mesh.value = self.mesh.value*x
		self.mesh.changed()

	def multiply_by_mesh(self, other_mesh):
		"""Vynasobi mrizku mrizkou other_mesh

		Pozor!!! Nasobi bod po bodu body ktere si odpovidaji indexem do rows,
		cols. Pokud ma vysledek k necemu byt, je dulezite aby mrizky mely
		stejne rozmery a stejne pozice bodu!
		"""
		self.mesh.value = self.mesh.value * other_mesh.mesh.value
		self.mesh.changed()

	def contains_invalid_point(self):
		return self.mesh.contains_invalid_point()

	def divide_by_mesh(self, other_mesh):
		"""Vydeli mrizku mrizkou other_mesh

		Pozor!!! Deli bod po bodu body ktere si odpovidaji indexem do rows,
		cols. Pokud ma vysledek k necemu byt, je dulezite aby mrizky mely
		stejne rozmery a stejne pozice bodu!

		Hazi MeshError, pokud doslo k deleni nulou.
		"""
		self.mesh.value = self.mesh.value / other_mesh.mesh.value
		self.mesh.changed()
		if not self.contains_invalid_point():
			return
		log_error(u"V dlic matici pravdpodobn byly nulov hodnoty.")
		log_error(u"Ve vsledn maticky jsou nkter hodnoty nekonen.")
		log_error(u"Neplatn vsledn matice.")
		raise MeshError

	def multiply_bottom(self, c, sigma = 0):
		"""Vynasobi body nize nez je sigma koeficientem c

		Tato funkce je zde kvuli vyhodnocovani smerovych navestidel.  Presne
		funguje tak, ze si nejprve naleznu radek nejblizsi hodnote sigma, ten
		povazuju za nulovy. Vsechny body na radcich s nizsi hodnotou sigma
		potom vynasobim koeficientem c. Nakonec spocitam hodnoty na nulovem
		radku jako prumer hodnot na nejblizsim vyssim a nizsim radkem
		"""
		rows, cols = self.rows.value, self.cols.value
		disthalf = misc_num.avg_dist(rows)/2
		if sigma - disthalf <= rows[0]:
			return	# vsechny nad
		n_rows, n_cols = len(rows), len(cols)
		zero_row = numpy.searchsorted(rows, sigma - disthalf)
		if zero_row == n_rows: # vsechny pod
			self.multiply(c)
			self.mesh.changed()
			return

		multmatrix = numpy.array(
			[c * numpy.ones(n_cols, dtype)] * zero_row + 
			[numpy.ones(n_cols, dtype)] * (n_rows - zero_row),
			dtype)
		
		new_mesh = self.mesh.value * multmatrix
		self.mesh.value = new_mesh
		#Jeste prumer na nulovem radku, jen pokud nejde o uplne krajni radek.
		if zero_row == 0 or zero_row == n_rows - 1:
			return
		mesh = self.mesh.value
		for col in xrange(n_cols):
			up = mesh[zero_row - 1][col]
			bo = mesh[zero_row + 1][col]
			mesh[zero_row][col] = 0.5 * (up + bo)

	def find_symmetric_points(self, sigma = 0):
		"""Nalezne dvojice uzlu mrizky symetricke podle osy sigma

		sigma musi primo lezet v nejakem radku mrizky, pokud tomu tak neni,
		vybere se nejblizsi radek a symetrie se dela podle tohoto radku!

		vraci: seznam [((col_idx0, row_idx0), (col_idx1, row_idx1)), ...], kde
			row != sigma, col_idx0 == col_idx1, row0 == -row1
			radky s indexem 0 jsou v bile oblasti, radky s indexem 1 v cervene.
		"""
		pts = []
		rows, cols = self.rows.value, self.cols.value
		delta = (rows[-1] - rows[0])/(2.0 * (len(rows) - 1))
		if sigma - delta <= rows[0]:
			return pts # vsechny nad
		n_rows, n_cols = len(rows), len(cols)
		zero_row = numpy.searchsorted(rows, sigma - delta)
		if zero_row == n_rows: 
			return pts # vsechny pod

		for ridx0 in xrange(zero_row):
			ridx1 = zero_row + (zero_row - ridx0)
			if ridx1 >= n_rows:
				continue
			pts.extend([((c, ridx1), (c, ridx0)) for c in xrange(n_cols)])

		return pts

	def indices2meas_points(self, indices):
		"""Prevede seznam tuplu(col_idx, row_idx) na [MeasuredPoint, ...]

		seznam MeasuredPoint ma vyplnene pouze osi!
		"""
		mesh = self.mesh.value
		l = []
		for p in indices:
			mp = MeasuredPoint()
			mp.omega, mp.sigma = self.point_coords(*p)
			mp.intensity = mesh[p[1]][p[0]]
			l.append(mp)
		return l

	def _interpolate(self, left, leftval, right, rightval, pos):
		"""Linearni inter/extrapolace

		dva body primky: y(left) = leftval, y(right) = rightval
		vraci y(pos).
		"""
		return leftval + (rightval - leftval)*(pos - left)/(right - left)

	def _get_hval(self, row_idx, left_idx, right_idx, x):
		"""Nalezne hodnotu na radku row_idx, na pozici x.
		
		Hodnota je interpolovana z hodnot ve sloupcich left_idx a right_idx.
		"""
		mesh, cols = self.mesh.value, self.cols.value
		return self._interpolate(
			cols[left_idx], mesh[row_idx][left_idx], 
			cols[right_idx], mesh[row_idx][right_idx], x)

	def _get_vval(self, col_idx, top_idx, bot_idx, y):
		"""Nalezne hodnotu na sloupci col_idx, na pozici y.
		
		Hodnota je interpolovana z hodnot v radcich top_idx a bot_idx.
		"""
		mesh, rows = self.mesh.value, self.rows.value
		return self._interpolate(
			rows[top_idx], mesh[top_idx][col_idx], 
			rows[bot_idx], mesh[bot_idx][col_idx], y)

	def get_val(self, x, y):
		"""Najde hodnotu mrizky v bode x (omega, po sloupcich), y(sigma, po radcich).

		souradnice omega, sigma nemusi lezet primo na nekterem z bodu mrizky,
		intenzita se pocita interpolaci. Pokud lezi vne mrizky, hodnota
		intenzity se spocte tak jako by bod lezel prave na kraji mrizky.
		"""
		rows, cols = self.rows.value, self.cols.value

		if not len(rows) or not len(cols):
			return 0.0

		lidx, ridx, x = misc_num.vec_find_neighbourgs(cols, x)
		bidx, tidx, y = misc_num.vec_find_neighbourgs(rows, y)
		botval = self._get_hval(bidx, lidx, ridx, x)
		topval = self._get_hval(tidx, lidx, ridx, x)
		return self._interpolate(rows[bidx], botval, rows[tidx], topval, y)

	def find_intersections(self, line):
		"""Nalezne pruseciky cary line se sloupci a radky mesh.

		line: instance Line
		vraci: [MeasuredPoint, ...]
		seznam MeasuredPoint ma vyplnene pouze osi!
		"""
		pts = []
		epsilon = 0.001
		rows, cols = self.rows.value, self.cols.value
		#Nejprve zjistim, ktere sloupce a radky jsou perspektivni
		xmin, xmax = cols[0], cols[-1]
		ymin, ymax = rows[0], rows[-1]
		x0, y0, x1, y1 = line.x0, line.y0, line.x1, line.y1
		xmin, xmax = min(x0, x1), max(x0, x1)
		ymin, ymax = min(y0, y1), max(y0, y1)
		first_col = numpy.searchsorted(cols, xmin - epsilon)
		first_row = numpy.searchsorted(rows, ymin - epsilon)
		#Nevcetne posledniho radku
		last_col = numpy.searchsorted(cols, xmax + epsilon)
		last_row = numpy.searchsorted(rows, ymax + epsilon)
		#Radky (horizontaly)
		if abs(y0 - y1) > epsilon:
			k = (x1 - x0) / (y1 - y0)
			for row in xrange(first_row, last_row):
				y = rows[row]
				x = x0 + k * (y - y0)
				pts.append(MeasuredPoint().set_osi(x, y, self.get_val(x, y)))
		#Sloupce (vertikaly)
		if abs(x0 - x1) > epsilon:
			k = (y1 - y0) / (x1 - x0)
			for col in xrange(first_col, last_col):
				x = cols[col]
				y = y0 + k * (x - x0)
				pts.append(MeasuredPoint().set_osi(x, y, self.get_val(x, y)))
		return pts

	def create_from_function(self, rows, cols, f):
		"""vytvori mrizku z funkce f(omega, sigma)

		rows: seznam souradnic radku
		cols: seznam souradnic sloupcu
		f: funkce f(omega, sigma)
		"""
		self.rows.value, self.cols.value = rows, cols
		self.mesh.resize((len(rows), len(cols)))
		m = self.mesh.value
		for rowidx, rowcoord in enumerate(rows):
			for colidx, colcoord in enumerate(cols):
				m[rowidx][colidx] = f(colcoord, rowcoord)
		self.mesh.changed()

	def read_from_meas_data(self, data):
		"""prevede data z instance MeasData na mrizku self.mesh

		v datech musi byt pocet radku > 1 a pocet sloupcu > 1.  Metoda
		predpoklada, ze data jsou pomerne rovnomerne rozptylena do mrizky, a
		souradnice jednotlivych bodu se neodchyluji vice nez nekolik procent od
		uzlu mrizky. Body nicmene nemusi byt nijak setrizeny.

		Nyni se mrizka pocita z bodu definovanych souradnicemi omega_def,
		sigma_def. Drive se pocitalo primo z omega, sigma, ale tento postup mel
		jednu sice malou, ale podstatnou a tezko odstranitelnou nevyhodu.

		Predstavme si, ze nas zajima naklopeni a natoceni stolu pro omega_def =
		0, sigma_def = 90. Zjistim polohu stolu, a zpetne z polohy stolu
		spocitam omega, sigma. Pokud neslo o nejaky singularni pripad, bude
		omega velmi podobne omega_def a sigma sigma_def. Nejake drobne
		nepresnosti tu vsak prece jenom budou (i kdyz mensi nez v radu
		tisicin), takze sigma_def muze byt napr. 90.00003 nebo 89.999995. To je
		prave ta neprijemnost, protoze 90.000003 uz je MIMO povoleny rozsah
		<-90, 90> a proto se jeste v table_math prepocita na sigma =
		-89.999997, omega = 180 (Pozorovatel - snimac pozoruje navestidlo ze
		stejneho pozorovaciho uhlu, ale ,,visi vzhuru nohama''), coz jiz je v
		poradku. Udela mi to nicmene poradnou rotyku v read_from_meas_data,
		protoze v podstate nelze nijak jednoduse dat do mrizky uhly sigma
		[70.000003, 80.0000003, -89.999997, ...]. Jednodussi je proto pocitat s
		definovanymi souradnicemi a doufat pri tom, ze odchylka od nalezenych
		neni prilis velka. Radost z toho nemam, interpolace tu temer ztraci
		smysl (ale nevadi), protoze mrizka v definovanych bodech je v podstate
		uplne presna, az na opravdu zanedbatelne chyby nasobeni. Neda se nic
		delat, pokud se nevymysli nic lepsiho.......
		"""
		def cmp_floats(x, y):
			if x < y:
				return -1
			elif x > y:
				return 1
			return 0

		def find_step(points, idx):
			"""
			Nalezne krok mezy body v setrizenem poli points.  Nejprve naleznu
			nejvetsi rozdil mezi dvema body, pak rozdelim body do skupin ve
			kterych se od sebe vzdalenosti bodu lisi o vice nez TRESHOLD *
			nejvetsi krok, nakonec spocitam prumernou vzdalenost mezi temito
			skupinami.
			"""
			TRESHOLD = 0.5
			maxstep = 0
			for i in xrange(len(points) - 1):
				maxstep = max(points[i+1][idx] - points[i][idx], maxstep)
			treshold = TRESHOLD * maxstep #0.5 should work fine

			num_groups = 0
			group_size = cur_group_size = 0
			cur_group_sum = last_group_sum = 0
			diff_sum = 0
			#Add a marker at the end to allow the loop to run one more time
			points.append([0, 0, 0])
			points[-1][idx] = points[-2][idx] + 2*treshold
			lastval = points[0][idx]
			for p in points:
				if p[idx] > lastval + treshold:
					if num_groups == 0:
						group_size = cur_group_size
						last_group_sum = cur_group_sum / group_size
					elif cur_group_size != group_size:
						points.pop() #remove marker
						return None #Got different num of items each time!
					else:
						cur_group_sum /= group_size
						diff_sum += cur_group_sum - last_group_sum
						last_group_sum = cur_group_sum
					#prepare for the next group
					num_groups += 1
					lastval = p[idx]
					cur_group_size = cur_group_sum = 0
				cur_group_size += 1
				cur_group_sum += p[idx]

			points.pop() #remove marker
			return diff_sum / (num_groups - 1)

		if not len(data.points):
			log_error(u"Seznam odmench bod je przdn.")
			raise MeshError

		points = [p.get_osi_def() for p in data.points]

		points = list(points)
		points.sort(lambda p1, p2: cmp_floats(p1[1], p2[1]))

		sigma_step = find_step(points, 1)
		if not sigma_step:
			log_error(u"Data nejsou v mrizce(sigma_step)")
			raise MeshError
		
		sigma_treshold = sigma_step * 0.5 #Should work fine
		rows = []
		row = []
		num_rows = num_cols = cur_num_cols = 0
		last_s = points[0][1]
		# Add a marker at the end to allow the loop to run one more time
		points.append([0, 0, 0])
		points[-1][1] = points[-2][1] + 2*sigma_treshold
		omega_step = 0
		for p in points:
			if p[1] > last_s + sigma_treshold:
				if (num_rows == 0):
					num_cols = cur_num_cols
				elif (cur_num_cols != num_cols):
					log_error(u"Vsechny radky musi mit stejne sloupcu")
					raise MeshError
				row.sort()
				cur_omega_step = find_step(row, 0)
				if not cur_omega_step:
					log_error(u"Data nejsou v mrizce(omega_step)")
					raise MeshError
				num_rows += 1
				omega_step += cur_omega_step
				rows.append(row)
				row = []
				last_s = p[1]
				cur_num_cols = 0
			cur_num_cols += 1
			row.append(p)

		omega_step /= num_rows
		self.rows.value = misc_num.arange_nelems(rows[0][0][1], sigma_step, num_rows, dtype)
		self.cols.value = misc_num.arange_nelems(rows[0][0][0], omega_step, num_cols, dtype)

		#Az doted jsem data nemenil, pouze tridil.
		self.mesh.freeze_changes()

		def bound(val, a, b):
			if val < a: return a
			if val > b: return b
			return val
		def point(r, c):
			r = bound(r, 0, num_rows - 1)
			c = bound(c, 0, num_cols - 1)
			return rows[r][c]
		rs, cs, mesh = self.rows.value, self.cols.value, self.mesh.value
		# tady koliduje rows(nahore, fakt tam jsou) s self.rows, vymysli lepsi jmeno.
		self.mesh.resize((num_rows, num_cols))
		mesh = self.mesh.value
		for row in xrange(0, num_rows):
			for col in xrange(0, num_cols):
				lp = point(row, col - 1)
				rp = point(row, col + 1)
				up = point(row - 1, col)
				bp = point(row + 1, col)
				p = point(row, col)
				dr = (rs[row] - p[1]) * (up[2] - bp[2]) / (up[1] - bp[1])
				dc = (cs[col] - p[0]) * (rp[2] - lp[2]) / (rp[0] - lp[0])
				mesh[row][col] = p[2] + dr + dc

		self.mesh.changed()
		self.mesh.thaw_changes()


	def scale_to_mesh(self, mesh):
		new_cols, new_rows = mesh.cols.value, mesh.rows.value
		return self.scale_to_cols_rows(new_cols, new_rows)

	def scale_to_cols_rows(self, new_cols, new_rows):
		"""Zmena velikosti, bikubicka interpolace
				
		new_rows: numpy.array, setrizene pole souradnic novych radku
		new_cols: numpy.array, setrizene pole souradnic novych sloupcu
		pri chybe hazi MeshScaleError
		"""
		epsilon = 0.5 
		# Maximalne muze byt new_rows (cols) vetsi o epsilon stupnu nez
		# puvodni, extrapolace velmi rychle diverguje.
		if (new_cols[0] < self.cols.value[0] - epsilon or
		   new_cols[-1] > self.cols.value[-1] + epsilon or
		   new_rows[0] < self.rows.value[0] - epsilon or
		   new_rows[-1] > self.rows.value[-1] + epsilon):
			log_error(u"Star mka nepokrv pln novou mku.")
			log_error(u"Chyba pi zmn velikosti mky")
			raise MeshScaleError

		import bicubic
		b = bicubic.BicubicMesh(self.mesh.value, self.cols.value, self.rows.value)
		try:
			self.mesh.value = b.interpolate_mesh(new_cols, new_rows)
			self.rows.value, self.cols.value = new_rows, new_cols
		except bicubic.BicubicError, e:
			log_error(u"Chyba pi zmn velikosti mky")
			raise MeshScaleError

		self.freeze_changes()
		self.mesh.changed()
		self.rows.changed()
		self.cols.changed()
		self.thaw_changes()

	def points_get(self):
		"""Vraci seznam vsech bodu

		kazdy bod je trojice (omega, sigma, intenzita).  Seznam je setrizeny po
		radcich (podle sigmy), kazdy radek je setrizeny (podle omegy)
		"""
		points = []
		rows, cols, mesh = self.rows.value, self.cols.value, self.mesh.value
		for ridx, r in enumerate(rows):
			points.extend([(c, r, mesh[ridx][cidx]) 
			               for cidx, c in enumerate(cols)])
		return points

	def ascii_report(self):
		"""Vrati ascii report, radky popisujici jednotlive namerene hodnoty"""
		def _print_point(floats):
			return u"%8.3f %14.3f %14.3f" % floats

		return u"\n".join((
			u"Namen data v souadnicch svtla",
			u"",
			u"omega[deg]    sigma[deg]     svtivost[cd]",
			u"\n".join([_print_point(p) for p in self.points_get()])
			))
