# -*- encoding: cp1250 -*-
"""Matika pro prevod pozorovacich uhlu svetla a polohy stolu.

Matika je oddelena zvlast pro mereni s pripravkem, mereni s ramenem a obecna
mereni bez pripravku (ta je zdaleka nejslozitejsi).
"""
# TODO pouzit jednodussi vzorce z diplomky?

from numpy import *
from misc_num import deg2rad, rad2deg, safe_asin, safe_acos, atan2
import amoeba
import debug
from sys import maxint
from typecheck import *
from storage import MeasuredPoint, MeasInfo
import debug

dtype = 'd'

# vzdalenost roviny stolu od osy naklapeni
table_board_distance = 0.136

# Udava pozici bodu, ktery zaroven lezi v ose otaceni pripravku a zaroven lezi
# na upinaci rovine pripravku v pripade ze naklopeni i natoceni stolu jsou
# nulove. wedge_pos je vztazena k _upinaci_ plose stolu.  Vyse popsany bod ma
# tedy v absolutni souradne soustave (vzhledem ke stredu polohovani stolu) tyto
# souradnice:
# (wedge_pos[0], wedge_pos[1], wedge_pos[2] + table_board_distance)
wedge_pos = array((0.0, 0.1635, 0.131), dtype) # [m]


# Uhel pripravku
#           \
#            \
#            /\ <- upinaci rovina pripravku
#           |  \
#  wedge_angle  \    upinaci rovina stolu (tilt = 0)
#    ===========================       -----> y
wedge_angle = deg2rad(44.825) # [rad]

# Vzdalenost osy otaceni ramene od pruseciku os polohovani stolu ve smeru osy z
# (predpoklada se, ze osa otaceni ramene lezi primo na ose z).
arm_dist = 0.128 # [m]

# Polomer ramene
arm_radius = 1.70 # [m]

# mimochodem, podle stareho GFM je cidlo pro vsesmeraky umisteno
# 0.0732 m ve smeru osy z nad stredem polohovani.

# Spolecne matematicke funkce ###############################################

def rotate_x(v, fi):
	""" rotace okolo osy x
	
	vraci vektor vznikly rotaci vektoru v okolo primky y = z = 0 o uhel fi
	ve smeru hodinovych rucicek pri pohledu y doprava/z nahoru"""
	cosfi, sinfi = cos(fi), sin(fi)
	m = array(((1, 0, 0), (0, cosfi, sinfi), (0, -sinfi, cosfi)), dtype)
	return dot(m, v)

def rotate_y(v, fi):
	""" rotace okolo osy y
	
	vraci vektor vznikly rotaci vektoru v okolo primky x = z = 0 o uhel fi
	ve smeru hodinovych rucicek pri pohledu x doleva/z nahoru """
	cosfi, sinfi = cos(fi), sin(fi)
	m = array(((cosfi, 0, -sinfi), (0, 1, 0), (sinfi, 0, cosfi)), dtype)
	return dot(m, v)

def rotate_z(v, fi):
	""" rotace okolo osy z
	
	vraci vektor vznikly rotaci vektoru v okolo primky x = y = 0 o uhel fi
	ve smeru hodinovych rucicek pri pohledu x dolu/y doprava """
	cosfi, sinfi = cos(fi), sin(fi)
	m = array(((cosfi, sinfi, 0), (-sinfi, cosfi, 0), (0, 0, 1)), dtype)
	return dot(m, v)

def rotate_free(vec, axis, fi):
	"""Rotuje vektor vec okolo osy axis (ve 3d, prochazejici pocatkem) o fi rad.

	Kdyz se divam tak, ze osa ukazuje primo na me, rotace se deje proti smeru
	hodinovych rucicek.
	
	http://www.mines.edu/~gmurray/ArbitraryAxisRotation/ArbitraryAxisRotation.html
	"""
	x, y, z = vec[0], vec[1], vec[2]
	u, v, w = axis[0], axis[1], axis[2]
	u2, v2, w2 = u**2, v**2, w**2
	ux, vy, wz = u*x, v*y, w*z
	s = ux + vy + wz
	cosfi = cos(fi)
	l2 = u2 + v2 + w2
	lsin = sqrt(l2) * sin(fi)
	result = array((0, 0, 0), dtype)
	result[0] = u*s + (x*(v2+w2) - u*(vy+wz))*cosfi + lsin*(-w*y+v*z)
	result[1] = v*s + (y*(u2+w2) - v*(ux+wz))*cosfi + lsin*( w*x-u*z)
	result[2] = w*s + (z*(u2+v2) - w*(ux+vy))*cosfi + lsin*(-v*x+u*y)
	return result/l2
		
@returns(float)
def vec_len(v):
	"""delka vektoru"""
	return sqrt(reduce(lambda x, y: x + y*y, v, 0))

def vec_normalize(v):
	"""Prepocte vektor na jednotkovy"""
	return v / vec_len(v)

def v3d_cross(u, v):
	"""cross product dvou 3d vektoru"""
	return array((u[1]*v[2] - v[1]*u[2],
	              v[0]*u[2] - u[0]*v[2], 
	              u[0]*v[1] - v[0]*u[1]), dtype)

@returns(float)
def v3d_angle(u, v):
	"""Uhel mezi u, v. u i v musi byt jednotkove!"""
	fi = sum(u*v)
	if (fi <= -1):
		return pi
	if (fi >= 1):
		return float(0)
	return arccos(fi)

@returns(float)
def v3d_axisangle(u, v, axis):
	"""spocte uhel od u k v okolo osy axis

	Kdyz se divam tak, ze osa ukazuje primo na me, rotace se deje proti smeru
	hodinovych rucicek od u k v.

	Hazi OverflowError v degenerovanem pripade.
	"""

	a = axis/vec_len(axis)
	#Zjistim kolmice k p a kazdemu z u,v
	nu, nv = v3d_cross(u, a), v3d_cross(v, a)
	try:
		nu /= vec_len(nu)
		nv /= vec_len(nv)
	except OverflowError:
		# Hodime pokud u nebo v splyva s axis.
		raise

	#Spoctu uhel
	fi = v3d_angle(nu, nv)
	#Kvuli znamenku (+/-)
	n = v3d_cross(a, nu)
	if sum(n * nv) > 0:
		return fi
	return -fi

# TableMathError ############################################################

class TableMathError(Exception):
	pass

# TableMath #################################################################

class TableMath(object):
	"""Predek vsech matik
	
	Pouze zde je poskytovano verejne rozhrani k matice, tj. get_rot_tilt_arm,
	get_omega_sigma atd.. Tyto metody "pro styk s uzivatelem" jeste prijimaji a
	vraceji udaje ve stupnich. Volaji pretizene metody svych potomku
	do_get_rot_tilt_arm, do_get_omega_sigma, ... , ktere naopak jiz pracuji
	pouze v radianech. V potomcich TableMath by se stupne vubec nemely objevit.
	"""

	TILT_LIMIT = (deg2rad(-46.5), deg2rad(46.5))
	ROT_LIMIT =  (deg2rad(-190.0),deg2rad(190.0))
	ARM_LIMIT =  (deg2rad(-5.0), deg2rad(95.0))

	#-- verejny protokol

	def __init__(self, meas_info):
		self.light_rot_zero = deg2rad(meas_info.light_rot_zero.value)
		self.light_tilt_zero = deg2rad(meas_info.light_tilt_zero.value)
		self.sensor = meas_info.sensor.as_numarray()

	@takes("TableMath", MeasuredPoint)
	@returns(MeasuredPoint)
	def compute_measpoint_osi(self, meas_point):
		"""U MeasuredPoint vypocte osi """
		rot_deg, tilt_deg, arm_deg, illuminance = meas_point.get_rtai()
		meas_point.omega, meas_point.sigma = self.get_omega_sigma(rot_deg, tilt_deg, arm_deg)
		dist = self.dist_f_sensor(meas_point.rot, meas_point.tilt, meas_point.arm)
		meas_point.intensity = meas_point.illuminance * (dist**2)
		return meas_point

	def get_rot_tilt_arm(self, omega_deg, sigma_deg):
		"""Vraci rot, tilt, arm pro dane omega_deg, sigma_deg. [deg]

		Vraci tuple (rot, tilt, arm) [deg].

		Tato funkce jeste bere parametry ve stupnich, ale vola funkci
		do_get_rot_tilt_arm, ktera jiz pracuje v radianech.
		"""

		self.assert_measinfo_valid()
		omega, sigma = [self._deg2rad(x) for x in (omega_deg, sigma_deg)]
		self._assert_nontext(self.allowed_omega_sigma(omega, sigma))
		rta = self.do_get_rot_tilt_arm(omega, sigma)
		rot_deg, tilt_deg, arm_deg = [self._rad2deg(x) for x in rta]

		rot_deg = rot_deg % 360.0
		if rot_deg > 180.0:
			rot_deg -= 360.0

		rta = tuple(self._deg2rad(x) for x in (rot_deg, tilt_deg, arm_deg))
		self._assert_nontext(self.allowed_rot_tilt_arm(*rta))
		return (rot_deg, tilt_deg, arm_deg)
	
	def get_omega_sigma(self, rot_deg, tilt_deg, arm_deg):
		"""Vraci omegu a sigmu pro dane rot, tilt, arm. [deg]

		Omegu vraci v rozsahu <-180, +180>, sigmu v rozsahu <-90, 90> [deg].
		Vraci tuple (omega, sigma) [deg].
		Tato funkce jeste bere parametry ve stupnich, ale vola funkci
		do_get_omega_sigma, ktera jiz pracuje v radianech.
		"""

		self.assert_measinfo_valid()
		rot, tilt, arm = [self._deg2rad(x) for x in (rot_deg, tilt_deg, arm_deg)]
		self._assert_nontext(self.allowed_rot_tilt_arm(rot, tilt, arm))
		os = self.do_get_omega_sigma(rot, tilt, arm)
		omega_deg, sigma_deg = [self._rad2deg(x) for x in os]

		# Zde vyuzivam toho, ze pokud zvysim(snizim) sigmu o pi a zaroven
		# provedu omega = pi - omega, dostanu identicke postaveni (jsem v te
		# same fotometricke rovine ale vzhuru nohama). Timto vzorcem tedy muzu
		# snadno prevest sigma do rozsahu povolenych hodnot, tedy <-90, 90>.
		# Omega se potom jiz prevadi snadno, rozsah <-180, 180> je periodicky s
		# periodou 360.

		while(sigma_deg < -90.0): 
			omega_deg, sigma_deg = 180.0 - omega_deg, sigma_deg + 180.0
		while(sigma_deg > +90.0): 
			omega_deg, sigma_deg = 180.0 - omega_deg, sigma_deg - 180.0

		omega_deg = omega_deg % 360.0
		if omega_deg > 180.0:
			omega_deg -= 360.0
		os = tuple(self._deg2rad(x) for x in (omega_deg, sigma_deg))
		self._assert_nontext(self.allowed_omega_sigma(*os))
		return (omega_deg, sigma_deg)

	def dist_f_sensor(self, rot_deg, tilt_deg, arm_deg):
		""" Vzdalenost cidla od zdroje. rot, tilt, arm [deg]. """
		self.assert_measinfo_valid()
		rta = tuple(self._deg2rad(x) for x in (rot_deg, tilt_deg, arm_deg))
		self._assert_nontext(self.allowed_rot_tilt_arm(*rta))
		return self.do_get_dist_f_sensor(*rta)

	#-- private

	def assert_measinfo_valid(self):
		"""Pokud neni meas_info v poradku, necht hodi TableMathError"""
		pass

	@classmethod
	def allowed_rot_tilt_arm(cls, rot, tilt, arm):
		if rot != None and (rot < cls.ROT_LIMIT[0] or rot > cls.ROT_LIMIT[1]):
			return u"Otoen (%.3f) stolu mimo povolen rozsah" % rad2deg(rot)
		if tilt != None and (tilt < cls.TILT_LIMIT[0] or tilt > cls.TILT_LIMIT[1]):
			return u"Naklopen (%.3f) stolu mimo povolen rozsah" % rad2deg(tilt)
		if arm != None and (arm < cls.ARM_LIMIT[0] or arm > cls.ARM_LIMIT[1]):
			return u"Poloha ramene (%.3f) stolu mimo povolen rozsah" % rad2deg(arm)

	@classmethod
	def allowed_omega_sigma(cls, omega, sigma):
		if omega < -pi or omega > pi:
			return u"Omega (%.3f) stolu mimo povolen rozsah" % rad2deg(omega)
		if sigma < -pi/2 or sigma > pi/2:
			return u"Sigma (%.3f) stolu mimo povolen rozsah" % rad2deg(sigma)

	def _assert_nontext(self, nontext):
		if nontext != None:
			raise TableMathError(nontext)

	def _rad2deg(self, x):
		if x != None:
			return rad2deg(x)
		return None

	def _deg2rad(self, x):
		if x != None:
			return deg2rad(x)
		return None

	def _str_vec(self, vec):
		# myslim ze je to nejkratsi mozny zapis - nicmene mu to neubira na
		# citelnosti :)
		return u"[%s]" % (', '.join('%.3f'%v for v in vec))

# TableMathWedge ############################################################

class TableMathWedge(TableMath):
	""" Matika pro mereni s priravkem
	
	Tato matika predpoklada, ze svetlo je umistene na pripravku a ohnisko lezi
	v ose otaceni pripravku. Poloha ohniska (meas_info.f) se zadava vztazena k
	upinaci plose pripravku
	"""
	@takes("TableMathWedge", MeasInfo)
	def __init__(self, meas_info):
		TableMath.__init__(self, meas_info)
		# f je poloha ohniska vzhledem k absolutnim souradnicim v pripade rot =
		# tilt = 0, nikoliv k upinaci rovine pripravku! pouzivam ho v teto
		# matice pouze k pocitani vzdalenosti pripravku od cidla.
		self.f = array((0, 0, table_board_distance), dtype)
		self.f += wedge_pos
		self.f += rotate_x(array((0.0, 0.0, meas_info.f.z), dtype), wedge_angle)

		psi = atan2(self.f[2], self.f[1])
		self.ksi1 = pi/2 - psi
		self.ksi2 = pi - wedge_angle - psi

	def str_settings(self):
		return (u"sensor: %s, f: %s, ksi1: %f deg, ksi2: %f deg" % (
		        self._str_vec(self.sensor), self._str_vec(self.f), 
		        rad2deg(self.ksi1), rad2deg(self.ksi2)))

	def assert_measinfo_valid(self):
		if self.sensor[0] != 0:
			raise TableMathError(u"\n".join((
				u"Snma mus mt nulovou x-ovou polohu."
				u"Chybn nastaven pracovit."
				)))

	def do_get_omega_sigma(self, rot, tilt, arm):
		omega = rot - self.light_rot_zero
		f_ = rotate_x(self.f, -tilt)
		fi = atan2(self.sensor[2] - f_[2], self.sensor[1] - f_[1])
		sigma = wedge_angle + fi - tilt
		return (omega, sigma)

	def do_get_rot_tilt_arm(self, omega, sigma):
		rot = self.light_rot_zero + omega

		delta = sigma + self.ksi2
		a, b, d = self.sensor[1], self.sensor[2], vec_len(self.f)
		alpha1 = atan2(b, a)
		p = vec_len((a, b))
		gamma1 = safe_asin(d*sin(delta)/p)
		alpha2 = pi - delta - gamma1
		alpha = alpha1 + alpha2
		tilt = alpha + self.ksi1 - pi/2
		return (rot, tilt, None)

	def do_get_dist_f_sensor(self, rot, tilt, arm):
		new_f = rotate_x(self.f, tilt + wedge_angle)
		retval = vec_len(new_f - self.sensor)
		return retval

# TableMathArm ############################################################

class TableMathArm(TableMath):
	""" Matika pro rameno
	
	Tato matika predpoklada, ze svetlo je umistene na stole, okolo ktereho
	krouzi rameno se snimacem ktery se neustale diva doprostred upinaci plochy
	stolu pri stole v nulove poloze. 

	Jediny parametr podstatny pro matiku, ktery se z meas_info bere, je poloha
	ohniska, a to pouze pro zjisteni vzdalenosti mezi ohniskem a cidlem pro
	prepocet osvetlenost -> svitivost. Poloha ohniska se jako vzdy stanovuje od
	upinaci plochy.

	Predpokladam, ze kdyz se koukam primo do cidla, a nastavim souradnici rot =
	0, je svetlo na stul namontovane tak aby se omega koukala primo ve smeru
	osy y.
	"""

	@takes("TableMathArm", MeasInfo)
	def __init__(self, meas_info):
		TableMath.__init__(self, meas_info)
		self.f = array((0, 0, table_board_distance), dtype)
		self.f += meas_info.f.as_numarray()
		self.light_rot_zero = meas_info.light_rot_zero.value

	def assert_measinfo_valid(self):
		# V soucasne implementaci musi ohnisko lezet v kruznici, po ktere se
		# pohybuje rameno.
		epsilon = 0.1 # [m]
		f = array(self.f)
		f = f - array((0, 0, arm_dist), dtype)
		if vec_len(f) + epsilon >= arm_radius:
			raise TableMathError(u"\n".join((
				u"Ohnisko le mimo krunici vymezenou pohybem ramene."
				u"Chybn nastaven pracovit."
				)))

	def str_settings(self):
		return u"f: %s, light_rot_zero: %.1f" % (self._str_vec(self.f), self.light_rot_zero)

	def do_get_omega_sigma(self, rot, tilt, arm):
		arm = pi/2 - arm # viz README k dokumentaci matiky z 2008-05-15
		omega = self.light_rot_zero - rot
		s_y = arm_radius*sin(arm)
		s_z = arm_radius*cos(arm) + arm_dist
		sigma = atan2(s_z - self.f[2], s_y)
		return (omega, sigma)

	def do_get_rot_tilt_arm(self, omega, sigma):
		rot = self.light_rot_zero - omega
		sigma_ = pi/2 - sigma
		a, c = arm_dist - self.f[2],  arm_radius
		gamma = arcsin(a*sin(sigma_)/c)
		arm_ = pi - gamma - sigma_
		arm = pi - arm_
		arm = pi/2 - arm # viz README k dokumentaci matiky z 2008-05-15
		return (rot, 0.0, arm)

	def do_get_dist_f_sensor(self, rot, tilt, arm):
		""" Vzdalenost cidla od zdroje. rot, tilt, arm [deg]. """
		arm = pi/2 - arm # viz README k dokumentaci matiky z 2008-05-15
		arm_ = pi - arm
		a, c = arm_dist - self.f[2],  arm_radius
		b = sqrt(c*c + a*a - 2*a*c*cos(arm_))
		return b

# TableMathArbitrary #########################################################

class TableMathArbitrary(TableMath):
	"""Trida pro prevod souradnic stolu a svetla

	Interne jsou vsechny uhly v radianech, ale do metod z esterniho protokolu
	vstupuji a vystupuji ve stupnich.
	"""
	# max. odchylka v hledani souradnic amoeby (?)
	epsilon = deg2rad(0.005)
	# max. pocet iteraci v amoebe
	maxiters = 100
	# koef. simplex. algoritmu (?)
	amoeba_scale = [0.1, 0.1]
	# hodnota vyskove fce pro amoebu v nedosazitelnych polohach stolu
	BOUNDARY_HEIGHT = 5.0 
		
	@takes("TableMathArbitrary", MeasInfo)
	def __init__(self, meas_info):
		""" 
		
		Vse je v kartezskych souradnicich, osa z smeruje nahoru, osa y od stolu
		smerem k cidlu a osa x smerem k pozorovateli ktery stoji pred stolem
		celem ke stolu a cidlo ma po prave ruce. Nulovy bod souradne soustavy
		se nachazi v pruseciku os nataceni a naklapeni. Vsechny vzdalenosti
		jsou uvadeny v metrech. 
		"""

		TableMath.__init__(self, meas_info)

		self.f = meas_info.f.as_numarray()
		self.f[2] += table_board_distance
		# f je poloha ohniska vzhledem k absolutnim souradnicim v poloze rot = tilt = 0

		self.photometric_system = meas_info.photometric_system.value
		if self.photometric_system == 'C-gamma':
			self.o = array((0, 1, 0), dtype)
			self.p = array((0, 0, 1), dtype)
			self.a = array((1, 0, 0), dtype)
		else:
			assert(None == "Bad photometric system")
		if self.light_rot_zero != 0.0 or self.light_tilt_zero != 0.0:
			self.o = array(self.move_object(self.light_rot_zero, self.light_tilt_zero, self.o))
			self.p = array(self.move_object(self.light_rot_zero, self.light_tilt_zero, self.p))
			self.a = array(self.move_object(self.light_rot_zero, self.light_tilt_zero, self.a))

	#-- verejny protokol

	def do_get_omega_sigma(self, rot, tilt, arm):
		"""Zjisteni omegy a sigmy po pootoceni a naklopeni stolu"""
		return self.rot_tilt_to_plane_angle(rot, tilt)

	def do_get_rot_tilt_arm(self, omega, sigma):
		"""Jak musim natocit a naklopit stul abych ziskal pozadovane omega, sigma

		omega: pozadovane omega
		sigma: pozadovane sigma
		"""
		return self.plane_angle_to_rot_tilt_arm(omega, sigma)

	def do_get_dist_f_sensor(self, rot, tilt, arm):
		"""Pro dane natoceni a naklopeni spocte vzdalenost ohniska od cidla"""
		f = self.move_object(rot, tilt, self.f)
		dist = vec_len(self.sensor - f)
		return dist

	#--

	def move_object(self, rot, tilt, v):
		"""Zjisti polohu bodu v po pootoceni stolu o rot a naklopeni o tilt
		
		Predpokladam, ze nez jsem zacal bodem hybat, byl stul v poloze 
		rot = tilt = 0. Vse je v radianech. 
		"""
		return rotate_x(rotate_z(v, rot), -tilt)

	def find_plane_angle(self, f, o, p, a):
		""" Nalezne uhly plane a angle pro dane f, o, p, a.

		Viz dokumentace k matice. Vse je v radianech.
		"""
		fc = self.sensor - f
		try:
			plane = v3d_axisangle(o, fc, p)
		except OverflowError: # degenerovany pripad, o nebo fc // p
			plane = 0.0

		new_o = rotate_free(o, p, plane)
		new_a = rotate_free(a, p, plane)
		try:
			angle = v3d_axisangle(new_o, fc, new_a)
		except OverflowError: # degenerovany pripad ? mozna nemuze nastat.
			angle = 0.0
		return (plane, angle)

	def rot_tilt_to_plane_angle(self, rot, tilt):
		"""Zjisti plane a angle pri nastavenem rot, tilt stolu 
		
		rot, tilt jsou v radianech, vraci vysledek v radianech.
		"""
		f = self.move_object(rot, tilt, self.f)
		o = self.move_object(rot, tilt, self.o)
		p = self.move_object(rot, tilt, self.p)
		a = self.move_object(rot, tilt, self.a)
		return self.find_plane_angle(f, o, p, a)

	def str_settings(self):
		return (u"sensor: %s, f: %s, o: %s, a: %s, p: %s" % (
		        self._str_vec(self.sensor), self._str_vec(self.f), self._str_vec(self.o), 
				self._str_vec(self.a), self._str_vec(self.p)))

	def height(self, wanted_plane, wanted_angle, rot, tilt):
		"""Funkce definujici vysku pro amoebu."""
		# Mam zadane dve dvojice (plane, angle). Prvni, parametr
		# plane_angle_to_rot_tilt_arm, zadal uzivatel a je to takove plane a
		# angle, pro ktere hledam vhodne natoceni a naklopeni stolu. Druha
		# dvojice se da spocitat z nejakeho natoceni a naklopeni stolu new_rot,
		# new_tilt. Nyni budu predpokladat, ze osa p je rovnobezna s osou z a
		# osa a je rovnobezna s osou x. Podle techto os natocim vektor (0, 1) o
		# hodnoty plane, angle kazde z dvojic a spoctu uhel mezi vyslednymi
		# vektory. To je zaroven vysledek funkce height. Pokud zjistim, ze
		# new_rot, new_tilt nelze na stole realizovat, vratim nejakou vyssi
		# hodnotu pro indikaci nedosazitelneho stavu amoebe. Protoze amoeba
		# funkci maximalizuje, vracim zapornou hodnotu uhlu.
		#
		# v1 by mohlo mit smysl cachovat, pokud by to bylo nutne. wanted_plane,
		# wanted_angle je stejne pro cele jedno hledani
		# plane_angle_to_rot_tilt_arm().

		if self.allowed_rot_tilt_arm(rot, tilt, None) != None:
			return -self.BOUNDARY_HEIGHT 
		new_plane, new_angle = self.rot_tilt_to_plane_angle(rot, tilt)
		v1 = rotate_z(rotate_x(array((0, 1, 0), dtype), wanted_angle), wanted_plane)
		v2 = rotate_z(rotate_x(array((0, 1, 0), dtype), new_angle), new_plane)
		retval = -v3d_angle(v1, v2)
		return retval

	def plane_angle_to_rot_tilt_arm(self, plane, angle):
		"""Zjisti jak musim otocit a naklopit stul abych dosahl pozadovaneho plane, angle

		Hazi TableMathError pokud se mu nepovedlo nalezt pozadovane natoceni a
		naklopeni v self.maxiters iteracich.

		plane, angle jsou v radianech. Vraci rot, tilt v radianech.
		"""

		def height(new_rot_tilt, data = None):
			return self.height(plane, angle, new_rot_tilt[0], new_rot_tilt[1])

		# Udelam si seznam vychozich bodu a setridim je od nejlepsiho
		STEP = deg2rad(30)
		points = []

		for r in arange(0, self.ROT_LIMIT[1] + STEP, STEP):
			for t in arange(0, self.TILT_LIMIT[1] + STEP, STEP):
				points.append((-height((+r, +t)), +r, +t))
				points.append((-height((-r, +t)), -r, +t))
				points.append((-height((+r, -t)), +r, -t))
				points.append((-height((-r, -t)), -r, -t))
		points.pop(0); points.pop(0); points.pop(0) # (mam tam 4x (0, 0))
		points.sort()
		while points[-1][0] == self.BOUNDARY_HEIGHT:
			points.pop(-1)
		
		# Ted uz "jen" najdu minimum
		for point, iter in zip(points, xrange(maxint)):
			(rot, tilt), h, iters = amoeba.amoeba([point[1], point[2]], self.amoeba_scale, height, 
			                                      self.epsilon, self.epsilon, self.maxiters)

			setting = 'try: %i, rot[deg]: %.3f, tilt[deg]: %.3f, f: %f, iters: %i' % \
			          (iter, rad2deg(rot), rad2deg(tilt), h, iters)
			if (-h <= self.epsilon):
				return (rot, tilt, None)

		# nepovedlo se rot, tilt nalezt..
		raise TableMathError(u"\n".join((
			"f: %s, o: %s, p: %s, a: %s, c: %s, plane: %f, angle: %f" % (
				self._str_vec(self.f), self._str_vec(self.o), self._str_vec(self.p),
				self._str_vec(self.a), self._str_vec(self.sensor), rad2deg(plane), rad2deg(angle)),
			u"Nepodailo se nalzt odpovdajc polohu stolu.",
			u"Pravdpodobn nedosaiteln stav.")))

# funkce ####################################################################

def create_table_math(meas_info):
	tmclasses = {
			meas_info.NOWEDGE: TableMathArbitrary,
			meas_info.WEDGE: TableMathWedge,
			meas_info.ARM: TableMathArm,
	}
	tmclass = tmclasses[meas_info.meas_type.value]
	return tmclass(meas_info)
