#!/usr/bin/env python
# -*- coding: iso-8859-1 -*-
################################################################################
#
# Method call parameters/return value type checking decorators.
# Dmitry Dvoinikov <dmitry@targeted.org>, Oct 2006
#
# Samples:
#
# from typecheck import *
#
# @takes(int, str) # takes int, str, upon a problem throws InputParameterError
# @returns(int)    # returns int, upon a problem throws ReturnValueError
# def foo(i, s): 
#     return i + len(s)
#
# @takes((int, long), (str, unicode)) # int or long, str or unicode (aka basestring)
# def foo(i, s, anything):            # and the third parameter is not checked
#     ...
#
# class foo(object):
#     @takes("foo", optional(int))  # foo, maybe int, but foo is yet incomplete
#     def __init__(self, i = None): # and is thus specified by name
#         ...
#     @takes("foo", int)            # foo, and int if presents in args,
#     def bar(self, *args):         # if args is empty, the check passes ok
#         ...
#     @takes("foo")                 
#     @returns(object)              # returns foo which is fine, because
#     def biz(self):                # foo is an object
#         return self
#     @classmethod                  # classmethod's and staticmethod's
#     @takes(type)                  # go same way
#     def baz(cls):
#         ...
#    
# @takes(int)
# @returns(optional("int", foo))    # returns either int, foo or NoneType
# def bar(i):                       # "int" (rather than just int) is for fun
#     if i > 0: 
#         return i
#     elif i == 0:
#         return foo()              # otherwise returns NoneType
#
# @takes(callable)                  # built-in functions are treated as predicates
# @returns(lambda x: x == 123)      # and so do user-defined functions or lambdas
# def execute(f, *args, **kwargs):
#     return f(*args, **kwargs)
#
# assert execute(execute, execute, execute, lambda x: x, 123) == 123
#
# def readable(x):                  # user-defined type-checking predicate
#     return hasattr(x, "read")
#
# anything is an alias for predicate lambda: True,
# nothing is an alias for NoneType, as in:
#
# @takes(callable, readable, optional(anything), optional(int))
# @returns(nothing)
# def foo(f, r, x = None, i = None):
#     ...
#
# @takes(with_attr("read", "write")) # another way of protocol checking
# def foo(pipe):
#     ...
#
# @takes(list_of(int))              # list of ints
# def foo(x):
#     print x[0]
#
# @takes(tuple_of(callable))        # tuple of callables
# def foo(x):
#     print x[0]()
#
# @takes(dict_of(str, list_of(int))) # dict mapping strs to lists of int
# def foo(x):
#     print sum(x["foo"])
#
# @takes(by_regex("^[0-9]{1,8}$"))  # integer-as-a-string regex
# def foo(x):
#     i = int(x)
#
# The (3 times longer) source code with self-tests is available from:
# http://www.targeted.org/python/recipes/typecheck.py
#
################################################################################

__all__ = [ "takes", "InputParameterError", "returns", "ReturnValueError", 
            "optional", "nothing", "anything", "list_of", "tuple_of", "dict_of",
            "by_regex", "with_attr" ]

no_check = False # set this to True to turn all checks off

################################################################################

from inspect import getargspec, isfunction, isbuiltin, isclass
from types import NoneType
from re import compile as regex

################################################################################

def base_names(C):
    "Returns list of base class names for a given class"
    return [ x.__name__ for x in C.__mro__ ]
    
################################################################################

def type_name(v):
    "Returns the name of the passed value's type"
    return type(v).__name__

################################################################################

class Checker(object):

    def __init__(self, reference):
        self.reference = reference

    def check(self, value): # abstract
        pass

    _registered = [] # a list of registered descendant class factories

    @staticmethod
    def create(value): # static factory method
        for f, t in Checker._registered:
            if f(value):
                return t(value)
        else:
            return None

################################################################################

class TypeChecker(Checker):

    def check(self, value):
        return isinstance(value, self.reference)

Checker._registered.append((isclass, TypeChecker))

nothing = NoneType

################################################################################

class StrChecker(Checker):

    def check(self, value):
        value_base_names = base_names(type(value))
        return self.reference in value_base_names or "instance" in value_base_names
   
Checker._registered.append((lambda x: isinstance(x, str), StrChecker))

################################################################################

class TupleChecker(Checker):

    def __init__(self, reference):
        self.reference = map(Checker.create, reference)

    def check(self, value):
        return reduce(lambda r, c: r or c.check(value), self.reference, False)

Checker._registered.append((lambda x: isinstance(x, tuple) and not
                                      filter(lambda y: Checker.create(y) is None,
                                             x), 
                            TupleChecker))

optional = lambda *args: args + (NoneType, )

################################################################################

class FunctionChecker(Checker):

    def check(self, value):
        return self.reference(value)

Checker._registered.append((lambda x: isfunction(x) or isbuiltin(x), 
                            FunctionChecker))

anything = lambda *args: True

################################################################################

class ListOfChecker(Checker):

    def __init__(self, reference):
        self.reference = Checker.create(reference)

    def check(self, value):
        return isinstance(value, list) and \
               not filter(lambda e: not self.reference.check(e), value)

list_of = lambda *args: lambda value: ListOfChecker(*args).check(value)

################################################################################

class TupleOfChecker(Checker):

    def __init__(self, reference):
        self.reference = Checker.create(reference)

    def check(self, value):
        return isinstance(value, tuple) and \
               not filter(lambda e: not self.reference.check(e), value)

tuple_of = lambda *args: lambda value: TupleOfChecker(*args).check(value)

################################################################################

class DictOfChecker(Checker):

    def __init__(self, key_reference, value_reference):
        self.key_reference = Checker.create(key_reference)
        self.value_reference = Checker.create(value_reference)

    def check(self, value):
        return isinstance(value, dict) and \
               not filter(lambda e: not self.key_reference.check(e), value.iterkeys()) and \
               not filter(lambda e: not self.value_reference.check(e), value.itervalues())

dict_of = lambda *args: lambda value: DictOfChecker(*args).check(value)

################################################################################

class RegexChecker(Checker):

    def __init__(self, reference):
        self.reference = regex(reference)

    def check(self, value):
        return isinstance(value, basestring) and self.reference.match(value)

by_regex = lambda *args: lambda value: RegexChecker(*args).check(value)

################################################################################

class AttrChecker(Checker):

    def __init__(self, *attrs):
        self.attrs = attrs

    def check(self, value):
        return reduce(lambda r, c: r and c, map(lambda a: hasattr(value, a), self.attrs), True)

with_attr = lambda *args: lambda value: AttrChecker(*args).check(value)

################################################################################

def takes(*args):
    "Method signature checking decorator"

    # convert decorator arguments into a list of checkers

    checkers = []
    for i, arg in enumerate(args):
        checker = Checker.create(arg)
        if checker is None:
            raise TypeError("@takes decorator got parameter %d of unsupported "
                            "type %s" % (i + 1, type_name(arg)))
        checkers.append(checker)

    if no_check: # no type checking is performed, return decorated method itself

        def takes_proxy(method):
            return method        

    else:

        def takes_proxy(method):
            
            method_args, method_defaults = getargspec(method)[0::3]

            def takes_invocation_proxy(*args, **kwargs):
    
                # append the default parameters

                if method_defaults is not None and len(method_defaults) > 0 \
                and len(method_args) - len(method_defaults) <= len(args) < len(method_args):
                    args += method_defaults[len(args) - len(method_args):]

                # check the types of the actual call parameters

                for i, (arg, checker) in enumerate(zip(args, checkers)):
                    if not checker.check(arg):
                        raise InputParameterError("%s() got invalid parameter "
                                                  "%d of type %s" %
                                                  (method.__name__, i + 1, 
                                                   type_name(arg)))

                return method(*args, **kwargs)

            takes_invocation_proxy.__name__ = method.__name__
            return takes_invocation_proxy
    
    return takes_proxy

class InputParameterError(TypeError): pass

################################################################################

def returns(sometype):
    "Return type checking decorator"

    # convert decorator argument into a checker

    checker = Checker.create(sometype)
    if checker is None:
        raise TypeError("@returns decorator got parameter of unsupported "
                        "type %s" % type_name(sometype))

    if no_check: # no type checking is performed, return decorated method itself

        def returns_proxy(method):
            return method        

    else:

        def returns_proxy(method):
            
            def returns_invocation_proxy(*args, **kwargs):
                
                result = method(*args, **kwargs)
                
                if not checker.check(result):
                    raise ReturnValueError("%s() has returned an invalid "
                                           "value of type %s" % 
                                           (method.__name__, type_name(result)))

                return result
    
            returns_invocation_proxy.__name__ = method.__name__
            return returns_invocation_proxy
        
    return returns_proxy

class ReturnValueError(TypeError): pass

################################################################################

if __name__ == '__main__': # run self-tests
    
    print "self-testing module typecheck.py:"

    # decorators arguments

    # test

    try:
        @takes(int, 10, bool)
        def poor(): 
            pass
    except TypeError, e:
        assert str(e) == "@takes decorator got parameter 2 of unsupported type int"
    else:
        assert False
    
    # test

    try:
        @returns(10)
        def poor(): 
            pass
    except TypeError, e:
        assert str(e) == "@returns decorator got parameter of unsupported type int"
    else:
        assert False

    # simple passthrough checking

    # test

    class biz: pass
    class baz(object): pass

    class foo(object): # for a new-style class

        @takes("foo", int)
        @returns("int")
        def a(self, i):
            return i

        @classmethod
        @takes(type, "str")
        @returns(str)
        def b(cls, s):
            return s

        @staticmethod
        @takes(biz)
        @returns(biz)
        def c(b):
            return b

        @staticmethod
        @takes(baz)
        @returns(baz)
        def d(b):
            return b

    assert foo().a(1) == 1
    assert foo.b("bar") == "bar"
    b = biz(); assert foo.c(b) is b
    b = baz(); assert foo.d(b) is b

    # test

    from new import classobj

    class bar: # for an old-style class

        @takes("bar", int)
        @returns("int")
        def a(self, i):
            return i

        @takes("anything_goes_for_old_style_class", int)
        @returns("int")
        def b(self, i):
            return i

        @classmethod
        @takes(classobj, "str")
        @returns(str)
        def c(cls, s):
            return s

        @staticmethod
        @takes(biz)
        @returns(biz)
        def d(b):
            return b

    assert bar().a(1) == 1
    assert bar().b(2) == 2
    assert bar().c("123") == "123"
    b = biz(); assert bar().d(b) is b

    # tuples + failure checking

    # test

    @takes((int, str))
    @returns((int, unicode))
    def foo(s):
        return s

    assert foo(1) == 1

    try:
        foo("str")
    except ReturnValueError, e:
        assert str(e) == "foo() has returned an invalid value of type str"
    else:
        assert False

    try:
        foo(u"unicode")
    except InputParameterError, e:
        assert str(e) == "foo() got invalid parameter 1 of type unicode"
    else:
        assert False

    # nested tuples + failure checking

    # test

    class foo(object):

        @takes("foo", (("str", "unicode"), (int, long)))
        @returns(((int, long), (bool, )))
        def f(self, x):
            return x

    x = 1; assert foo().f(x) is x
    x = 1L; assert foo().f(x) is x 
    x = False; assert foo().f(x) is x # because bool is a subclass of int

    try:
        foo().f(bool)
    except InputParameterError, e:
        assert str(e) == "f() got invalid parameter 2 of type type"
    else:
        assert False

    try:
        foo().f(u"unicode")
    except ReturnValueError, e:
        assert str(e) == "f() has returned an invalid value of type unicode"
    else:
        assert False

    # number of parameters checking

    # test

    @takes(int)
    def foo(i):
        pass

    try:
        foo() # less parameters provided than required
    except TypeError, e:
        assert str(e) == "foo() takes exactly 1 argument (0 given)"
    else:
        assert False

    # test

    @takes(int)
    def foo(i = None):
        pass

    try:
        foo() # less parameters provided than checked
    except InputParameterError, e:
        assert str(e) == "foo() got invalid parameter 1 of type NoneType"
    else:
        assert False

    # test

    @takes(optional(int))
    def foo(i = None):
        pass

    foo() # less parameters provided than checked, but check is optional

    # test

    @takes(optional(int))
    def foo(i = "123"):
        pass

    try:
        foo() # less parameters provided than checked, invalid default type
    except InputParameterError, e:
        assert str(e) == "foo() got invalid parameter 1 of type str"
    else:
        assert False

    # test

    class foo(object):
        @takes("foo", optional(int, long))
        def __init__(self, i = None):
            self.i = i

    assert foo(1).i == 1
    assert foo(1L).i == 1L
    foo()
    assert foo().i is None

    # test

    @takes(int, int)
    def foo(i, j = None):
        pass

    try:
        foo(1)
    except InputParameterError, e:
        assert str(e) == "foo() got invalid parameter 2 of type NoneType"
    else:
        assert False

    # test

    @takes(int, optional(int))
    def foo(i, j = None):
        pass

    foo(1)
    foo(1, 2)

    try:
        foo(1, "str")
    except InputParameterError, e:
        assert str(e) == "foo() got invalid parameter 2 of type str"
    else:
        assert False

    # test

    @takes(int)
    def foo(i, j):
        pass

    foo(1, 2)
    foo(1, "str")

    try:
        foo("str", 2)
    except InputParameterError, e:
        assert str(e) == "foo() got invalid parameter 1 of type str"
    else:
        assert False

    # test

    @takes(int)
    def foo(a = 123):
        return a

    assert foo(456) == 456
    assert foo() == 123
    try:
        foo(123, 456) == 456
    except TypeError, e:
        assert str(e) == "foo() takes at most 1 argument (2 given)"
    else:
        assert False

    @takes(int, str)
    def foo(i = 123, s = "123"):
        return str(i) == s

    assert foo()
    assert foo(123)
    assert foo(123, "123")
    assert foo(456, "456")
    assert not foo(1234)
    assert not foo(123, "1234")

    # test

    @takes(int, optional("int")) # note, k is not checked, and k is optional, and *args present
    def foo(i, j = 0, k = 0, *args, **kwargs):
        return (i + j + k) in args or ("sum" in kwargs and kwargs["sum"] == i + j + k)

    try:
        foo()
    except TypeError, e:
        assert str(e) == "foo() takes at least 1 argument (0 given)"
    else:
        assert False

    assert not foo(0)
    assert foo(0, sum = 0)
    assert not foo(1, 2)
    assert not foo(1, 2, 3, "any", "thing")
    assert foo(1, 2, 3, "any", 6, "thing")
    assert foo(1, 2, 3, "any", "thing", sum = 6)

    # test

    class foo(object):
        @takes("foo", int, "int") # note, k is not checked, and k is optional, and *args present
        def bar(self, i, j = 0, k = 0, *args, **kwargs):
            return (i + j + k) in args or ("sum" in kwargs and kwargs["sum"] == i + j + k)

    try:
        foo().bar()
    except TypeError, e:
        assert str(e) == "bar() takes at least 2 arguments (1 given)"
    else:
        assert False

    assert not foo().bar(0)
    assert foo().bar(0, sum = 0)
    assert not foo().bar(1, 2)
    assert not foo().bar(1, 2, 3, "any", "thing")
    assert foo().bar(1, 2, 3, "any", 6, "thing")
    assert foo().bar(1, 2, 3, "any", "thing", sum = 6)

    # test

    @takes(int, str) # wow ! it does optional checking too !
    def foo(i, *args):
        return i + len(args)

    assert foo(1) == 1
    assert foo(1, "str") == 2
    assert foo(1, "str", 2, 3) == 4

    try:
        foo(1, 2L, "what", "ever")
    except InputParameterError, e:
        assert str(e) == "foo() got invalid parameter 2 of type long"
    else:
        assert False

    # complex types checking

    # test

    @takes(tuple, list, dict)
    @returns(tuple)
    def foo(t, l, d): 
        return (t, l, d)
    
    assert foo((1, 2), [3, 4], { 5: 6 }) == ((1, 2), [3, 4], { 5: 6 })
    assert foo(foo((1, 2), [3, 4], {5: 6}), [7, 8], { 9: 10 }) == (((1, 2), [3, 4], {5: 6}), [7, 8], { 9: 10 })
    
    try:
        foo(foo, [3, 4], { 5: 6 })
    except InputParameterError, e:
        assert str(e) == "foo() got invalid parameter 1 of type function"
    else:
        assert False

    try:
        foo((1, 2), foo, { 5: 6 })
    except InputParameterError, e:
        assert str(e) == "foo() got invalid parameter 2 of type function"
    else:
        assert False

    try:
        foo((1, 2), [3, 4], foo)
    except InputParameterError, e:
        assert str(e) == "foo() got invalid parameter 3 of type function"
    else:
        assert False

    # test

    class foo(object):
        @takes("foo", int)
        def __init__(self, i):
            self.i = i
        @returns("foo")
        def clone(self):
            return foo(self.i)

    assert isinstance(foo(1).clone(), foo)
    assert foo(1).clone().i == 1

    class bar(foo):
        @returns("bar")
        def clone(self):
            return bar(self.i)

    assert isinstance(bar(1).clone(), bar)
    assert isinstance(foo.clone(bar(1)), foo)

    # test

    class foo(object):
        @takes("mumbo-jumbo")
        def __init__(self):
            pass
    
    try:
        foo()
    except InputParameterError, e:
        assert str(e) == "__init__() got invalid parameter 1 of type foo"
    else:
        assert False

    # test

    class foo(object):
        @takes("foo")
        def __init__(self):
            pass

    foo() 

    class bar(object):
        @takes("bar")
        def __init__(self):
            foo.__init__(123) # this is not checked with @takes

    try:
        bar()
    except TypeError, e:
        assert str(e) == "unbound method __init__() must be called with foo instance as first argument (got int instance instead)"
    else:
        assert False

    # test

    class foo(object):
        @returns("foo")
        def foo(self):
            return foo()

    class bar(foo):
        @returns("bar")
        def foo(self):
            return foo.foo(self)

    assert isinstance(foo().foo(), foo)
    try:
        bar().foo()
    except ReturnValueError, e:
        assert str(e) == "foo() has returned an invalid value of type foo"
    else:
        assert False

    # predicates checking

    # test

    @takes(callable)
    @returns(lambda x: x == 1)
    def execute(f, *args, **kwargs):
        return f(*args, **kwargs)

    assert execute(execute, execute, execute, lambda x: x, 1) == 1

    try:
        execute("yikes", 123)
    except InputParameterError, e:
        assert str(e) == "execute() got invalid parameter 1 of type str"
    else:
        assert False

    # test

    line_readable = lambda x: hasattr(x, "readline")

    @takes((line_readable, list), int)
    @returns(list)
    def read_lines(f, n):
        if isinstance(f, list):
            return f[:n]
        else:
            return [ f.readline().strip() for i in range(n) ]

    from sys import modules
    f = open(modules[__name__].__file__)
    try:
        a = read_lines(f, 2)
        assert a == [ "#!/usr/bin/env python", "# -*- coding: iso-8859-1 -*-" ]
        a = read_lines(a, 1)
        assert a == [ "#!/usr/bin/env python" ]
    finally:
        f.close()

    try:
        read_lines(read_lines, 0)
    except InputParameterError, e: 
        assert str(e) == "read_lines() got invalid parameter 1 of type function"
    else:
        assert False

    # test

    class foo(object):
        @returns(lambda x: hasattr(x, "bar"))
        def bar(self, x = None):
            return x or self

    foo().bar()
    foo().bar(foo())
    try:
        foo().bar(123)
    except ReturnValueError, e: 
        assert str(e) == "bar() has returned an invalid value of type int"
    else:
        assert False

    # test
    
    @takes(long)
    @returns(nothing)
    def foo(i):
        if i > 0:
            return i

    foo(0L)

    try:
        foo(1L)
    except ReturnValueError, e: 
        assert str(e) == "foo() has returned an invalid value of type long"
    else:
        assert False

    # test

    @takes((int, long), optional(str))
    def foo(i, s = None):
        pass

    foo(*(1, ))
    foo(*(2L, "str"))

    try:
        foo(*())
    except TypeError, e:
        assert str(e) == "foo() takes at least 1 argument (0 given)"
    else:
        assert False

    try:
        foo(*(3L, 4))
    except InputParameterError, e:
        assert str(e) == "foo() got invalid parameter 2 of type int"
    else:
        assert False

    try:
        foo(*(3L, "str", "oops"))
    except TypeError, e:
        assert str(e) == "foo() takes at most 2 arguments (3 given)"
    else:
        assert False

    # test

    class foo(object):
        @takes("foo", "foo", anything, "foo")
        def biz(self, a, x, b):
            return x in [ a, b ]

    assert not foo().biz(foo(), foo(), foo())
    f = foo()
    assert foo().biz(f, f, f)

    # test

    @takes(optional(anything))
    @returns(nothing)
    def foo(x = None):
        return x

    assert foo() is None

    try:
        foo(foo)
    except ReturnValueError, e: 
        assert str(e) == "foo() has returned an invalid value of type function"
    else:
        assert False

    # test

    def readable(x):
        return hasattr(x, "read")

    class bar(object):
        def read(self):
            pass

    @takes(callable, readable, optional(anything), optional(int))
    @returns(nothing)
    def foo(f, r, x = None, i = None):
        return x

    assert foo(lambda: True, bar()) is None
    assert foo(lambda: True, bar(), None) is None
    assert foo(lambda: True, bar(), None, 123) is None

    try:
        foo(lambda: True, bar(), None, "str")
    except InputParameterError, e:
        assert str(e) == "foo() got invalid parameter 4 of type str"
    else:
        assert False
    
    try:
        foo(lambda: True, bar(), bar())
    except ReturnValueError, e: 
        assert str(e) == "foo() has returned an invalid value of type bar"
    else:
        assert False

    # test

    @takes(list_of(int))
    def foo(x):
        pass

    foo([1])
    
    try:
        foo((1,))
    except InputParameterError, e:
        assert str(e) == "foo() got invalid parameter 1 of type tuple"
    else:
        assert False

    try:
        foo([1L])
    except InputParameterError, e:
        assert str(e) == "foo() got invalid parameter 1 of type list"
    else:
        assert False

    # test

    @returns(list_of(int))
    def foo(x):
        return x 

    foo([1])

    try:
        foo((1,))
    except ReturnValueError, e: 
        assert str(e) == "foo() has returned an invalid value of type tuple"
    else:
        assert False

    try:
        foo([1L])
    except ReturnValueError, e: 
        assert str(e) == "foo() has returned an invalid value of type list"
    else:
        assert False

    # test

    @takes(tuple_of(int))
    def foo(x):
        pass

    foo((1,))
    
    try:
        foo([1])
    except InputParameterError, e:
        assert str(e) == "foo() got invalid parameter 1 of type list"
    else:
        assert False

    try:
        foo((1L,))
    except InputParameterError, e:
        assert str(e) == "foo() got invalid parameter 1 of type tuple"
    else:
        assert False

    # test
    
    @returns(tuple_of(int))
    def foo(x):
        return x

    foo((1,))
    
    try:
        foo([1])
    except ReturnValueError, e:
        assert str(e) == "foo() has returned an invalid value of type list"
    else:
        assert False

    try:
        foo((1L,))
    except ReturnValueError, e:
        assert str(e) == "foo() has returned an invalid value of type tuple"
    else:
        assert False

    # test

    @takes(dict_of(int, str))
    def foo(x):
        pass

    foo({ 1: "foo" })
    
    try:
        foo((1, "foo"))
    except InputParameterError, e:
        assert str(e) == "foo() got invalid parameter 1 of type tuple"
    else:
        assert False

    try:
        foo({ "foo": "bar" })
    except InputParameterError, e:
        assert str(e) == "foo() got invalid parameter 1 of type dict"
    else:
        assert False

    try:
        foo({ 1: 2 })
    except InputParameterError, e:
        assert str(e) == "foo() got invalid parameter 1 of type dict"
    else:
        assert False

    # test

    @returns(dict_of(int, str))
    def foo(x):
        return x

    foo({ 1: "foo" })
    
    try:
        foo((1, "foo"))
    except ReturnValueError, e:
        assert str(e) == "foo() has returned an invalid value of type tuple"
    else:
        assert False

    try:
        foo({ "foo": "bar" })
    except ReturnValueError, e:
        assert str(e) == "foo() has returned an invalid value of type dict"
    else:
        assert False

    try:
        foo({ 1: 2 })
    except ReturnValueError, e:
        assert str(e) == "foo() has returned an invalid value of type dict"
    else:
        assert False

    # test

    @takes(list_of(tuple_of(dict_of((str, int), callable))))
    def foo(x):
        for t in x:
            for d in t:
                for k, v in d.iteritems():
                    assert k == v()

    foo([({ "foo": lambda: "foo" }, { 1: lambda: 1 }), ({ "bar": lambda: "bar" }, { 2: lambda: 2 })])

    try:
        foo([({ "foo": lambda: "foo" }, { 1: lambda: 1 }), "FAIL"])
    except InputParameterError, e:
        assert str(e) == "foo() got invalid parameter 1 of type list"
    else:
        assert False

    try:
        foo([({ "foo": lambda: "foo" }, { 1: lambda: 1 }), ({ "bar": lambda: "bar" }, "FAIL")])
    except InputParameterError, e:
        assert str(e) == "foo() got invalid parameter 1 of type list"
    else:
        assert False

    try:
        foo([({ "foo": lambda: "foo" }, { 1: lambda: 1 }), ({ "bar": lambda: "bar" }, { u"FAIL": lambda: u"FAIL" })])
    except InputParameterError, e:
        assert str(e) == "foo() got invalid parameter 1 of type list"
    else:
        assert False

    # test

    @takes(by_regex("^fo+$"))
    def foo(x):
        pass

    foo("fo")
    foo("foo")
    foo("fooo")

    try:
        foo("bar")
    except InputParameterError, e:
        assert str(e) == "foo() got invalid parameter 1 of type str"
    else:
        assert False

    try:
        foo(10)
    except InputParameterError, e:
        assert str(e) == "foo() got invalid parameter 1 of type int"
    else:
        assert False

    # test

    @returns(by_regex(u"^\u0410.{1,2}$"))
    def foo(x):
        return x

    foo(u"\u0410\u0411\u0412")

    try:
        foo(u"\u0411\u0412")
    except ReturnValueError, e:
        assert str(e) == "foo() has returned an invalid value of type unicode"
    else:
        assert False

    # test

    class foo(object):
        def bar(self):
            pass
        @classmethod
        def biz(cls):
            pass
        @staticmethod
        def baz():
            pass

    @takes(with_attr("bar"))
    def test(x):
        pass

    test(foo())
    
    @takes(with_attr("biz"))
    def test(x):
        pass

    test(foo())

    @takes(with_attr("baz"))
    def test(x):
        pass

    test(foo())

    @takes(with_attr("bar", "biz", "baz"))
    def test(x):
        pass

    test(foo())

    @takes(with_attr("notthere"))
    def test(x):
        pass

    try:
        test(foo())
    except InputParameterError, e:
        assert str(e) == "test() got invalid parameter 1 of type foo"
    else:
        assert False

    @takes(with_attr("bar", "notthere"))
    def test(x):
        pass

    try:
        test(foo())
    except InputParameterError, e:
        assert str(e) == "test() got invalid parameter 1 of type foo"
    else:
        assert False

    # test

    class old_foo:
        def bar(self):
            pass
        def biz(cls):
            pass
        biz = classmethod(biz)
        def baz():
            pass
        baz = staticmethod(baz)

    @takes(with_attr("bar"))
    def test(x):
        pass

    test(old_foo())
    
    @takes(with_attr("biz"))
    def test(x):
        pass

    test(old_foo())

    @takes(with_attr("baz"))
    def test(x):
        pass

    test(old_foo())

    @takes(with_attr("bar", "biz", "baz"))
    def test(x):
        pass

    test(old_foo())

    @takes(with_attr("notthere"))
    def test(x):
        pass

    try:
        test(old_foo())
    except InputParameterError, e:
        assert str(e) == "test() got invalid parameter 1 of type instance"
    else:
        assert False

    @takes(with_attr("bar", "notthere"))
    def test(x):
        pass

    try:
        test(old_foo())
    except InputParameterError, e:
        assert str(e) == "test() got invalid parameter 1 of type instance"
    else:
        assert False

    # test

    @returns(with_attr("__init__"))
    def foo():
        return 1

    foo()

    @returns(with_attr("__del__"))
    def bar():
        return 1

    try:
        bar()
    except ReturnValueError, e:
        assert str(e) == "bar() has returned an invalid value of type int"
    else:
        assert False

    @returns(with_attr("__init__", "__del__"))
    def bar():
        return 1

    try:
        bar()
    except ReturnValueError, e:
        assert str(e) == "bar() has returned an invalid value of type int"
    else:
        assert False

    # test

    class foo(object):
        def __init__(self):
            self.x = 1

    @takes(with_attr("x"))
    def test(x):
        pass

    test(foo())

    @takes(with_attr("y"))
    def test(x):
        pass

    try:
        test(foo())
    except InputParameterError, e:
        assert str(e) == "test() got invalid parameter 1 of type foo"
    else:
        assert False

    print "ok"

################################################################################
# EOF
