# -*- encoding: cp1250 -*-

""" Kontejnery pro definice typovych zkousek a izokandel """

from numpy import sin, cos, pi, arange, searchsorted
from misc_num import safe_asin, safe_acos, safe_sqrt, is_float_modulo

from mvc import *
import registry
import misc

epsilon = 0.001
dtype = 'd'

#############################################################################
# Validatory ################################################################
#############################################################################

def validator_omega(pt):
	if not -180.0 <= pt <= 180:
		return u"Platn rozsah je -180 .. +180 stup"

def validator_sigma(pt):
	if not -90.0 <= pt <= 90:
		return u"Platn rozsah je -90 .. +90 stup"

def validate_omega_range(x):
	if x.omega_max.value <= x.omega_min.value:
		msg = u"Omega do mus bt vt ne omega od"
		x.validator_failure(msg, x.omega_min)
		return False
	return True

def validate_sigma_range(x):
	if x.sigma_max.value <= x.sigma_min.value:
		msg = u"Sigma do mus bt vt ne sigma od"
		x.validator_failure(msg, x.sigma_min)
		return False
	return True

def validate_omega_step(x):
	if not validate_omega_range(x):
		return False

	delta = x.omega_max.value - x.omega_min.value
	if 2*x.omega_step.value > delta:
		x.validator_failure(u"Krok omegy se musi vejt do rozsahu od - do nejmn dvakrt.", x.omega_step)
	if not is_float_modulo(delta, x.omega_step.value, epsilon):
		x.validator_failure(u"Krok omegy mus bt celoselnm nsobkem rozsahu od - do.", x.omega_step)
		return False
	return True

def validate_sigma_step(x):
	if not validate_sigma_range(x):
		return False

	delta = x.sigma_max.value - x.sigma_min.value
	if 2*x.sigma_step.value > delta:
		x.validator_failure(u"Krok sigmy se musi vejt do rozsahu od - do nejmn dvakrt.", x.sigma_step)
	if not is_float_modulo(delta, x.sigma_step.value, epsilon):
		x.validator_failure(u"Krok sigmy mus bt celoselnm nsobkem rozsahu od - do.", x.sigma_step)
		return False
	return True

def validator_circle(x):
	if not -360 - epsilon <= x <= 360 + epsilon:
		return u"Platn rozsah je -360 .. +360 stup"

def validator_nonnegative(x):
	if x <= 0 - epsilon:
		return u"Hodnota mus bt nezporn"

def validator_positive(x):
	if x < 0 - epsilon:
		return u"Hodnota mus bt kladn"

def validator_at_least_005(x):
	if x < 0.005 - epsilon:
		return u"Hodnota mus bt alespo 0.005"

def validator_at_least_1(x):
	if x < 1.0 - epsilon:
		return u"Hodnota mus bt alespo 1"

def validate_list_nonempty(model, list, list_description = None):
	if len(list) == 0:
		if not list_description:
			msg = u"Seznam nesm bt przdn"
		else:
			msg = u"Seznam %s nesm bt przdn" % list_description
		model.validator_failure(msg, list)
		return False
	return True

#############################################################################
# Definice izokandel ########################################################
#############################################################################

class Isocandel(CompositeModel):
	_models = {
		'omega_min': FloatModel(0.0, validator_omega),
		'omega_max': FloatModel(0.0, validator_omega),
		'sigma_min': FloatModel(0.0, validator_sigma),
		'sigma_max': FloatModel(0.0, validator_sigma),
	}
	_models_order = ('omega_min', 'omega_max', 'sigma_min', 'sigma_max')

	def move(self, offset):
		"""Destruktivne pohne izokandelou o zadany offset(instanci Point)"""
		oo, os = offset.get_xy()
		self.omega_min.value += oo
		self.omega_max.value += oo
		self.sigma_min.value += os
		self.sigma_max.value += os
		return self

	def get_lines(self):
		"""Vraci list s carami izokandely.
		
		prvky (cary) jsou ve tvaru ((omega, sigma), (omega, sigma))"""
		raise NotImplementedError

	def get_bbox(self):
		return Rect().set_lrbt(
				self.omega_min.value, self.omega_max.value,
				self.sigma_min.value, self.sigma_max.value)

	def get_omega_span(self, sigma):
		"""Vraci rozpeti omegy ktere izokandela pokryva v urovni sigma.

		(min, max) nebo None pokud nepokryva nic.
		"""
		raise NotImplementedError
			
	def get_sigma_span(self, omega):
		"""Vraci rozpeti sigmy ktere izokandela pokryva v urovni omega.

		(min, max) nebo None pokud nepokryva nic.
		"""
		raise NotImplementedError

	@staticmethod
	def _points2lines(points, ox = 0, oy = 0):
		"""Prevede seznam bodu na seznam car

		napr. z bodu points = [(0, 0), (1, 1), (2, 2)]
		vraci [((0, 0), (1, 1)), ((1, 1), (2, 2))]
		"""
		lines = ArrayModel(Line())
		for p1, p2 in misc.pairwise(points):
			lines.append(Line().set(p1[0] + ox, p1[1] + oy, p2[0] + ox, p2[1] + oy))
		return lines

	def find_inner_points_indices(self, mesh):
		"""Nalezne body mrizky mesh, ktere lezi uvnitr izokandely.

		mesh necht je instance storage.Mesh.
		vraci tuple dvojic indexu do mesh ((col, row), (col, row), ...)
		"""
		def find_first(array, n):
			return min(len(array) - 1, searchsorted(array, n - epsilon))
		def find_last(array, n):
			return max(0, searchsorted(array, n + epsilon) - 1)

		inner_points = []

		rowmin_idx = find_first(mesh.rows.value, self.sigma_min.value)
		rowmax_idx = find_last(mesh.rows.value, self.sigma_max.value)
		for r in xrange(rowmin_idx, rowmax_idx+1):
			try:
				col_min, col_max = self.get_omega_span(mesh.rows.value[r])
			except TypeError: #vratilo None
				continue
			colmin_idx = find_first(mesh.cols.value, col_min)
			colmax_idx = find_last(mesh.cols.value, col_max)
			inner_points.extend([(c, r) for c in xrange(colmin_idx, colmax_idx + 1)])
		return inner_points

	def find_inner_points(self, mesh):
		"""Nalezne vsechny body uvnitr izokandely.

		Nalezne vsechny body uvnitr izokandely, tj. vsechny pruseciky radku
		a sloupcu uvnitr mrizky mesh ktere lezi uvnitr izokandely.

		mesh necht je instance storage.Mesh
		Vraci [storage.MeasuredPoint, ...]
		"""
		return mesh.indices2meas_points(self.find_inner_points_indices(mesh))

	def find_mesh_intersections(self, mesh):
		"""Nalezne vsechny pruseciky izokandely s mesh.

		Viz storage.Contour.find_intersections()
		"""
		pts = []
		for l in self.get_lines():
			pts.extend(mesh.find_intersections(l))
		return pts
	
	def find_contour_minmax(self, mesh):
		"""Nalezne minimum a maximum izokandely na mrizce mesh.

		Hleda na prusecicich izokandely s jednotlivymi radky a sloupci mrizky
		Nejprve posune izokandelu o offset_omega, offset sigma

		mesh necht je instance storage.Mesh
		Vraci tuple (MeasuredPoint min, MeasuredPoint max)
		pokud se nepodarilo nalezt, vraci None.
		"""
		pts = self.find_mesh_intersections(mesh)
		if len(pts) == 0:
			return None
		return (
			reduce(lambda a, b: misc.ternary(a.intensity < b.intensity, a, b), pts),
			reduce(lambda a, b: misc.ternary(a.intensity > b.intensity, a, b), pts)
		)

	def validator(self):
		return validate_omega_range(self) and validate_sigma_range(self)

# IsocandelRectangle ########################################################

class IsocandelRectangle(Isocandel):
	_models = {
		'intensity': FloatModel(1.0, validator_positive),
	}
	_models.update(Isocandel._models)

	def get_description(self):
		return (u"Obdlnk: rove %.1f, [%.3f, %.3f] - [%3f, %3f]" %
			(self.intensity.value, self.omega_min.value, self.omega_max.value, 
			 self.sigma_min.value, self.sigma_max.value))

	def get_lines(self):
		om, ox, sm, sx = self.get_bbox().get_lrbt()
		points = [(om, sm), (om, sx), (ox, sx), (ox, sm), (om, sm)]
		return self._points2lines(points)

	def get_omega_span(self, sigma):
		if not self.sigma_min.value - epsilon <= sigma <= self.sigma_max.value + epsilon:
			return None
		return (self.omega_min.value, self.omega_max.value)

	def get_sigma_span(self, omega):
		if not self.omega_min.value - epsilon <= omega <= self.omega_max.value + epsilon:
			return None
		return (self.sigma_min.value, self.sigma_max.value)

# IsocandelEllipse ##########################################################

class IsocandelEllipse(Isocandel):
	_models = {
		'intensity': FloatModel(1.0, validator_positive),
	}
	_models.update(Isocandel._models)
		
	def get_description(self):
		return (u"Elipsa: rove %.1f, [%.3f, %.3f] - [%3f, %3f]" %
			(self.intensity.value, self.omega_min.value, self.omega_max.value, 
			 self.sigma_min.value, self.sigma_max.value))

	def _get_ellipse_params(self):
		om, ox, sm, sx = self.get_bbox().get_lrbt()
		a, b = (ox - om) / 2, (sx - sm) / 2
		offset_x, offset_y = (om + ox)/2, (sm + sx)/2
		return (a, b, offset_x, offset_y)

	def get_lines(self):
		step = 0.1
		a, b, offset_x, offset_y = self._get_ellipse_params()
		points = [(a*cos(t), b*sin(t)) for t in arange(0, 2*pi + step, step)]
		return self._points2lines(points, offset_x, offset_y)

	def get_omega_span(self, sigma):
		if not self.sigma_min.value - epsilon <= sigma <= self.sigma_max.value + epsilon:
			return None
		a, b, offset_x, offset_y = self._get_ellipse_params()
		span = a*cos(safe_asin((sigma - offset_y)/b))
		return (-span + offset_x, +span + offset_x)
		
	def get_sigma_span(self, omega):
		if not self.omega_min.value - epsilon <= omega <= self.omega_max.value + epsilon:
			return None
		a, b, offset_x, offset_y = self._get_ellipse_params()
		span = b*sin(safe_acos((omega - offset_x)/a))
		return (-span + offset_y, +span + offset_y)

# IsocandelOval #############################################################
class IsocandelOval(Isocandel):
	"""Oval (polokruhy nalevo a napravo).

	Urcen rozsahem sigma a omega, vzdy se cely tesne vejde do obdelniku
	omega_min, sigma_min .. omega_max, sigma_max
	"""
	_models = {
		'white_intensity': FloatModel(1.0, validator_positive),
		'red_intensity':   FloatModel(1.0, validator_positive),
	}
	_models.update(Isocandel._models)

	def get_description(self):
		return (u"Ovl: rove %.1f/%.1f, [%.3f, %.3f] - [%3f, %3f]" %
			(self.white_intensity.value, self.red_intensity.value,
			self.omega_min.value, self.omega_max.value, 
			self.sigma_min.value, self.sigma_max.value))

	def get_lines(self):
		step = 0.1
		om, ox, sm, sx = self.get_bbox().get_lrbt()
		a, b = (ox - om)/2 - (sx - sm)/2, (sx - sm)/2
		oo, so = (om+ox)/2, (sm+sx)/2
		lines = ArrayModel(Line())
		lines.extend(self._points2lines(((-a, -b), (a, -b)), oo, so))
		lines.extend(self._points2lines(((-a, +b), (a, +b)), oo, so))
		l = [(b*cos(t)-a, b*sin(t)) for t in arange(pi/2, 3*pi/2+step, step)]
		r = [(b*cos(t)+a, b*sin(t)) for t in arange(-pi/2, pi/2+step, step)]
		lines.extend(self._points2lines(l, oo, so))
		lines.extend(self._points2lines(r, oo, so))
		return lines

	def get_omega_span(self, sigma):
		if not self.sigma_min.value - epsilon <= sigma <= self.sigma_max.value + epsilon:
			return None
		om, ox, sm, sx = self.get_bbox().get_lrbt()
		ol, sl = ox - om, sx - sm
		oc, sc = (om + ox)/2, (sm + sx)/2
		a = safe_sqrt((sl/2)**2 - (sigma - sc)**2)
		return (oc - a, oc + a)
		
	def get_sigma_span(self, omega):
		if not self.omega_min.value - epsilon <= omega <= self.omega_max.value + epsilon:
			return None
		om, ox, sm, sx = self.get_bbox().get_lrbt()
		ol, sl = ox - om, sx - sm
		oc, sc = (om + ox)/2, (sm + sx)/2
		if (om + sl/2 <= omega <= ox - sl/2):
			return (sm, sx)
		d = abs(oc - omega) - (ol - sl/2)
		a = safe_sqrt((sl/2)**2 - d**2)
		return (sc - a, sc + a)

	def validator(self):
		Isocandel.validator(self)
		if not self.validation.valid:
			return False

		do = self.omega_max.value - self.omega_min.value
		ds = self.sigma_max.value - self.sigma_min.value
		if do + epsilon < ds:
			msg = (u"Ovl mus mt vt ku ne vku.")
			self.validator_failure(msg, self.sigma_min)
			return False
		return True

isocandel_templates = [IsocandelRectangle(), IsocandelEllipse(), IsocandelOval()]

#############################################################################
#############################################################################
# Definice typovych zkousek #################################################
#############################################################################
#############################################################################

# Typetest ##################################################################

class Typetest(CompositeModel):
	"""Rodic vsech typovych zkousek"""
	dir_def = '' # Defaultni adresar s definici typove zkousky
	_models = {
		'version': VersionModel('1.0'),
		'name': StrModel(u"Nzev zkouky"),
		'description': StrModel(u"Popis zkouky"),
	}
	_models_order = ('version', 'name', 'description')

#############################################################################
# Izokandelove typove zkousky ###############################################
#############################################################################

# TypetestIso ###############################################################

class TypetestIso(Typetest):
	"""Typova zkouska pro izokandelova mereni (smerova, sestupova) """

	models_iso_range = {
		'omega_min': FloatModel(-1.0, validator_omega),
		'omega_max': FloatModel(1.0, validator_omega),
		'omega_step': FloatModel(1.0, validator_at_least_005),
		'sigma_min': FloatModel(-1.0, validator_sigma),
		'sigma_max': FloatModel(1.0, validator_sigma),
		'sigma_step': FloatModel(1.0, validator_at_least_005),
	}
	_models = {
		'isocandels': ListModel(isocandel_templates),
	}
	_models.update(models_iso_range)
	_models.update(Typetest._models)
	_models_order = Typetest._models_order + ('omega_min', 'omega_max', 'omega_step', 
	                'sigma_min', 'sigma_max', 'sigma_step', 'brightness', 'uniformity', 
					'sym_min', 'sym_max', 'isocandels')

	def get_bbox(self):
		"""Vraci Rect, minimalni obdelnik ohranicujici vsechny izokandely"""
		if not len(self.isocandels):
			return Rect().set_lbwh(0, 0, 0, 0)
		return Rect().set_union([i.get_bbox() for i in self.isocandels])

	def get_extent(self):
		return Rect().set_lbrt(self.omega_min.value, self.sigma_min.value,
		                       self.omega_max.value, self.sigma_max.value)

	@staticmethod
	def _coords(min, max, step):
		return arange(min.value, max.value + epsilon, step.value)
	
	def get_rows(self):
		"""Vraci numpy.array, setrizene pole souradnic radku

		radky zacinaji od sigma_min, po kroce sigma_step dokud nejsou
		vyssi nez sigma_max.
		"""
		return self._coords(self.sigma_min, self.sigma_max, self.sigma_step)

	def get_cols(self):
		"""Vraci numpy.array, setrizene pole souradnic sloupcu

		radky zacinaji od omega_min, po kroce omega_step dokud nejsou
		vyssi nez omega_max.
		"""
		return self._coords(self.omega_min, self.omega_max, self.omega_step)

	def validator(self):
		if (not validate_omega_step(self)) or (not validate_sigma_step(self)):
			return False
		if not validate_list_nonempty(self, self.isocandels, "izokandel"):
			return False

		def bbox_failure(msg, item):
			bbox_def = u"\n".join((
				u"",
				u"Rozsahy men omega a sigma v sob mus zahrnovat vechny",
				u"definovan izokandely.",
				u"Podle souasn definice izokandel mus bt",
				u"Omega od <= %f",
				u"Omega do >= %f",
				u"Sigma od <= %f",
				u"Sigma do >= %f")) % self.get_bbox().get_lrbt()
			self.validator_failure(msg + bbox_def, item)
		
		om, ox, sm, sx = self.get_bbox().get_lrbt()
		if self.omega_min.value > om:
			bbox_failure(u"Omega min. pli vysok", self.omega_min)
			return False
		if self.omega_max.value < ox:
			bbox_failure(u"Omega max. pli nzk", self.omega_max)
			return False
		if self.sigma_min.value > sm:
			bbox_failure(u"Sigma min. pli vysok", self.sigma_min)
			return False
		if self.sigma_max.value < sx:
			bbox_failure(u"Sigma max. pli nzk", self.sigma_max)
			return False

	def move_isocandels(self, offset):
		"""Destruktivne pohne typovou zkousku o zadany offset
		
		Tato metoda slouzi pro zmenu zamereni, pohne se vsemi izokandelami a
		typovka se opet vyhodnoti. Aby vse dobre fungovalo, musim pouzit
		nasledujici postup: Nejprve posunu vsechny izokandely o zadany offset.
		To by ovsem nestacilo, protoze se muze stat ze prekroci hranici danou
		omega/sigma min/max. V tom pripade by ale nemusela projit validace,
		protoze ta kontroluje zda izokandely nekde nepresahuji. Proto pak musim
		pripadne roztahnout tuto hranici tak, aby stale jeste zahrnovala
		vsechny izokandely, ale ne vic nez je nezbytne nutne. Musim roztahovat
		vzdy o krok, aby mi body ve kterych provadim vyhodnoceni posunute
		typovky padly do stejnych bodu jako vyhodnoceni puvodni typovky.

		offset: instance Point(), x ~ omega, y ~ sigma
		"""
		for i in self.isocandels:
			i.move(offset)
		bbox = self.get_bbox()

		while bbox.l < self.omega_min.value:
			self.omega_min.value -= self.omega_step.value
		while bbox.r > self.omega_max.value:
			self.omega_max.value += self.omega_step.value
		while bbox.b < self.sigma_min.value:
			self.sigma_min.value -= self.sigma_step.value
		while bbox.t > self.sigma_max.value:
			self.sigma_max.value += self.sigma_step.value

# TypetestDirectional #######################################################

class TypetestDirectional(TypetestIso):
	"""Typova zkouska, smerova navestidla
	
	Zde mimo jine rozlisuji zapustena a nezapustena navestidla. Zapustena
	navestidla, kterych je vetsina, nezari ve smerech sigma < 0, ovsem
	elipsovite izokandely jsou definovany tak, ze do teto oblasti zasahuji.
	Musim tedy zaridit, aby se tato oblast nevyhodnocovala, a aby v teto
	oblasti nevadilo, ze zde sice jsou definovany izokandely, ale nemeri se tu
	zadna data."""

	dir_def = registry.paths.typetests_def_directional
	dir_eval = registry.paths.typetests_eval_directional
	dir_data = registry.paths.typetests_data_directional
	dir_proto = registry.paths.typetests_protocols_directional

	_models = {
		'brightness': FloatModel(0.0, validator_nonnegative),
		'uniformity': FloatModel(1.0, validator_nonnegative),
		'inset': BoolModel(True), # True pokud jde o zapustene navestidlo
	}
	_models.update(TypetestIso._models)

	def get_bbox(self):
		bbox = TypetestIso.get_bbox(self)
		if self.inset.value:
			l, b, r, t = bbox.get_lbrt()
			bbox.set_lbrt(l, max(0, b), r, t)
		return bbox

# TypetestApproach #######################################################

class TypetestApproach(TypetestIso):
	"""Typova zkouska, sestupova navestidla"""
	
	dir_def = registry.paths.typetests_def_approach
	dir_eval = registry.paths.typetests_eval_approach
	dir_data = registry.paths.typetests_data_approach
	dir_proto = registry.paths.typetests_protocols_approach

	_models = {
		'sym_min': FloatModel(2.0, validator_positive),
		'sym_max': FloatModel(6.5, validator_positive),
	}
	_models.update(TypetestIso._models)

	def validator(self):
		return TypetestIso.validator(self) and self.validate_symmetry()

	@staticmethod
	def validate_symmetry(symmetry_specs_model):
		if symmetry_specs_model.sym_min.value >= symmetry_specs_model.sym_max.value:
			msg = u"Symetrie od mus bt men ne symetrie do."
			symmetry_specs_model.validator_failure(msg, symmetry_specs_model.sym_max)
			return False
		return True

##############################################################################
# Polarni typove zkousky (vsesmeraky, svitidla) ##############################
##############################################################################

# OmegaSpan #################################################################

class OmegaSpan(CompositeModel):
	"""Rozsah pro konstantni sigma (omega od do s krokem pri danem sigma)"""
	_models = {
		'sigma': FloatModel(0.0, validator_sigma),
		'omega_min': FloatModel(-1.0, validator_omega),
		'omega_max': FloatModel(1.0, validator_omega),
		'omega_step': FloatModel(1.0,validator_at_least_005),
		'intensity': FloatModel(1.0, validator_positive),
	}
	_models_order = ('sigma', 'omega_min', 'omega_max', 'omega_step', 'intensity')

	def points(self):
		"""Vraci seznam dvojic (omega, sigma), bodu ktere se maji odmerit"""

		def yield_sigma():
			while True: yield self.sigma.value
		return zip(
		           arange(self.omega_min.value,
		                  self.omega_max.value + epsilon, 
		                  self.omega_step.value, dtype),
		           yield_sigma(),
				   )

	def points_with_intensity(self):
		"""Vraci seznam dvojic (omega, sigma), bodu ktere se maji odmerit"""

		def yield_sigma():
			while True: yield self.sigma.value

		def yield_intensity():
			while True: yield self.intensity.value

		return zip(
		           arange(self.omega_min.value,
		                  self.omega_max.value + epsilon, 
		                  self.omega_step.value, dtype),
		           yield_sigma(),
				   yield_intensity(),
				   )

	def __len__(self):
		return len(self.points())

	def __getitem__(self, idx):
		return (self.omega_min.value + idx*self.omega_step.value, self.sigma.value)

	def validator(self):
		return validate_omega_step(self)

# SigmaSpan #################################################################

class SigmaSpan(CompositeModel):
	"""Rozsah pro konstantni omega (sigma od do s krokem pri danem omega)
	
	Narozdil od OmegaSpan zde pribyl jeste atribut parametr. Je urcen pro mereni
	typovych zkousek, ktere probiha nasledujicim zpusobem:
	1) Odmerim rozsah -180..180 omega pro nejake dane sigma
	2) Najdu minimalni intenzitu na rozsahu odmerenem v 1)
	3) Pro omega, ktere odpovidalo tomuto bodu minimalni intenzity, odmerim
		nejaky rozsah sigma.
	Atribut parameter pak rika, zda se ma merit v minimu ('min'),
	maximu('max'), nebo v hodnote dane atributem omega ('value').
	"""
	_models = {
		'omega': FloatModel(0.0, validator_omega),
		'sigma_min': FloatModel(-1.0, validator_sigma),
		'sigma_max': FloatModel(1.0, validator_sigma),
		'sigma_step': FloatModel(1.0, validator_at_least_005),
		'intensity': FloatModel(1.0, validator_positive),
		'parameter': StrEnumModel(('value', 'min', 'max'), 'value'),
	}
	_models_order = ('omega', 'sigma_min', 'sigma_max', 'sigma_step', 'intensity', 'parameter')

	def points(self):
		"""Vraci seznam dvojic (omega, sigma), bodu ktere se maji odmerit"""

		def yield_omega():
			while True: yield self.omega.value
		return zip(yield_omega(),
		           arange(self.sigma_min.value,
		                  self.sigma_max.value + epsilon, 
		                  self.sigma_step.value, dtype))
	def points_with_intensity(self):
		"""Vraci seznam dvojic (omega, sigma), bodu ktere se maji odmerit"""

		def yield_omega():
			while True: yield self.omega.value
		def yield_intensity():
			while True: yield self.intensity.value
		return zip(yield_omega(),
		           arange(self.sigma_min.value,
		                  self.sigma_max.value + epsilon, 
		                  self.sigma_step.value, dtype),
				   yield_intensity(),
				   )

	def __len__(self):
		return len(self.points())

	def __getitem__(self, idx):
		return (self.omega.value, self.sigma_min.value + idx*self.sigma_step.value)

	def validator(self):
		return validate_sigma_step(self)

# SpansMixin ################################################################

class SpansMixin(object):
	_models = {
		'omega_spans': ArrayModel(OmegaSpan()), # rozsahy pro konst. sigma
		'sigma_spans': ArrayModel(SigmaSpan()), # rozsahy pro konst. omega
	}

	def validator(self):
		os, ss = self.omega_spans, self.sigma_spans
		if len(os) == 0 and len(ss) == 0:
			self.validator_failure(u"Alespo jeden seznam mus bt neprzdn", os)
			return False
		if len(os) == 0: # Pokud merim min/max, potrebuju alespon jeden OmegaSpan
			if misc.any(span.parameter.value != 'value' for span in ss):
				self.validator_failure(u"Je poteba definovat alespo jeden rozsah pro konstatn omega", os)
				return False
		return True

# TypetestPolar #############################################################

class TypetestPolar(Typetest, SpansMixin):
	_models = {}
	_models.update(Typetest._models)
	_models.update(SpansMixin._models)

	def get_extent(self):
		r = t = -100000
		l = b =  100000
		for os in self.omega_spans:
			for pt in os.points():
				l, r = min(l, pt[0]), max(r, pt[0])
				b, t = min(b, pt[1]), max(t, pt[1])
		for ss in self.sigma_spans:
			for pt in ss.points():
				l, r = min(l, pt[0]), max(r, pt[0])
				b, t = min(b, pt[1]), max(t, pt[1])
		return Rect().set_lbrt(l, b, r, t)

# TypetestOmni ##############################################################

class TypetestOmni(TypetestPolar):
	"""Typova zkouska vsesmerovych navestidel"""
	dir_def = registry.paths.typetests_def_omni
	dir_eval = registry.paths.typetests_eval_omni
	dir_data = registry.paths.typetests_data_omni
	dir_proto = registry.paths.typetests_protocols_omni
	
# TypetestLight #############################################################

class TypetestLight(TypetestPolar):
	"""Typova zkouska svitidel"""
	dir_def = registry.paths.typetests_def_light
	dir_eval = registry.paths.typetests_eval_light
	dir_data = registry.paths.typetests_data_light
	dir_proto = registry.paths.typetests_protocols_light

#############################################################################
# Tabulkove typove zkousky ##################################################
#############################################################################

# TablePoint ################################################################

class TablePoint(CompositeModel):
	"""Jeden bod definice tabulkoveho mereni"""
	_models = {
		'omega': FloatModel(0.0, validator_omega),
		'sigma': FloatModel(0.0, validator_sigma),
		'intensity': FloatModel(0.0, validator_nonnegative),
	}

# TypetestTable ##############################################################

class TypetestTable(Typetest):
	"""Typova zkouska pro tabulkova mereni"""

	dir_def = registry.paths.typetests_def_table
	dir_eval = registry.paths.typetests_eval_table
	dir_data = registry.paths.typetests_data_table
	dir_proto = registry.paths.typetests_protocols_table
	
	_models = {
		'points': ArrayModel(TablePoint()),
	}
	_models.update(Typetest._models)

	def validator(self):
		return validate_list_nonempty(self, self.points, u"bod")

	def get_extent(self):
		l, b, r, t = None, None, None, None
		for pt in self.points:
			l, r = min(l, pt.omega.value), max(r, pt.omega.value)
			b, t = min(b, pt.sigma.value), max(t, pt.sigma.value)
		return Rect().set_lbrt(l, b, r, t)
