# -*- encoding: cp1250 -*-

import wx

from mvc import *
from plot import PlotModel, PlotView, PlotController
import registry
import storage
import misc
import pdf
import gfm_common_gui

resources = registry.xresources.viewdata

# LuminogramDlg #############################################################

class LuminogramDlg(Dialog):
	def __init__(self, model, parent = None):
		Dialog.__init__(self, model, resources, 'Dlg_Luminogram', parent)

		self.settings = registry.settings.value
		m, f = self.model, self.frame

		self.add_view(CheckEditor(m.lumi_on, f, 'Check_enabled'))
		self.add_view(CheckEditor(m.use_specified_lumi_range, f, 'Check_limit_intensity'))

		cmp_func = (lambda a,b: a.get_colormap() == b.get_colormap())
		try:
			idx = self.settings.lumi_themes.index(m.lumi_theme, cmp_func = cmp_func)
		except ValueError:
			idx = 0
		focused_theme = IntModel(idx)
		display_defs = {'name': (u"nzev", 100, list_text('%s', 'name'))}
		self.add_view(MulticolArrayView(self.settings.lumi_themes, f, 'List_themes', 
		                                ('name',), display_defs, focused_theme))
		self._copy_theme = ArrayItemCopier(self.settings.lumi_themes, focused_theme, m.lumi_theme)

# ViewdataFrame #############################################################

class ViewdataFrame(Frame):
	"""prohlizeni dat. Model je instance MeasData"""
	# Neni DocumentFrame, protoze
	# 1) Nema ukladani (to by tak nevadilo)
	# 2) Ma 2x open, jednou pro sestupaky a jednou pro smerova nasvestidla
	def __init__(self, model, parent = None):
		Frame.__init__(self, model, resources, 'Frame_ViewIso', parent)

		m, f = self.model, self.frame

		self.set_menu('MB_View')
		self.bind(wx.EVT_MENU, self.close, 'MI_Quit')
		self.bind(wx.EVT_MENU, self.on_open_directional, 'MI_OpenDirectional')
		self.bind(wx.EVT_MENU, self.on_open_approach, 'MI_OpenApproach')
		self.bind(wx.EVT_MENU, self.on_lumi, 'MI_Lumi')

		self.show_info = BoolModel(True)
		self.show_info.add_change_observer(self.update_text)
		self.add_view(CheckEditor(self.show_info, f, 'Check_info'))
		self.show_mesh = BoolModel(True)
		self.show_mesh.add_change_observer(self.update_text)
		self.add_view(CheckEditor(self.show_mesh, f, 'Check_mesh'))
		self.show_raw = BoolModel(False)
		self.show_raw.add_change_observer(self.update_text)
		self.add_view(CheckEditor(self.show_raw, f, 'Check_raw'))

		self.fname = FilenameModel('')

		self.plot = PlotModel()
		self.update_plot()

		plot_view = PlotView(self.plot, self.frame, resources, 'Plot_plot')
		self.add_view(plot_view)
		self.plot_controller = PlotController(self.plot, plot_view, 'omega', 'sigma')

		self.update_text()

	def open(self, dir):
		fname = open_dialog(dir, '', u"Vyberte soubor s namenmi daty",
		                    registry.wildcards.xml, self.frame)
		if not fname:
			return
		try:
			self.model.xml_load(fname)
		except XMLModelLoadError:
			return
		m = self.model.as_mesh()
		if not m:
			return
		self.fname.value = fname
		self.plot.mesh.copy_from(m)
		self.update_plot()
		self.update_text()

	def update_plot(self):
		gfm_common_gui.prepare_isoplot(self.plot, u"Namen data: %s" % self.fname.basename())

	def update_text(self):
		info = raw = mesh = u""
		md, mi = self.model, self.model.meas_info
		if self.show_info.value:
			info = self.model.meas_info.ascii_report()
		if self.show_mesh.value:
			mesh = self.plot.mesh.ascii_report()	
		if self.show_raw.value:
			raw = self.model.ascii_report_rtai()
		report = u"\n\n".join([i for i in (info, mesh, raw) if i]) 
		self.xrcctrl('Text_Report').SetValue(report)

	def done(self):
		self.show_info.del_change_observer(self.update_text)
		self.show_mesh.del_change_observer(self.update_text)
		self.show_raw.del_change_observer(self.update_text)

	def on_open_directional(self, evt):
		self.open(registry.paths.typetests_data_directional)

	def on_open_approach(self, evt):
		self.open(registry.paths.typetests_data_approach)

	def on_lumi(self, evt):
		LuminogramDlg.run_dialog(self.plot, self.frame)

# ExportData ################################################################

class ExportData(CompositeModel):
	"""Trida, ktera urcuje ktera data se maji exportovat (raw/meas)"""
	_models = {
		'raw': BoolModel(True),
		'meas': BoolModel(True),
	}

class ExportDataDlg(Dialog):
	"""Dialog nad ExportData"""
	def __init__(self, model, parent = None):
		Dialog.__init__(self, model, resources, 'Dlg_ChooseData', parent)
		m, f = self.model, self.frame
		self.add_view(CheckEditor(m.raw,  f, 'Check_raw'))
		self.add_view(CheckEditor(m.meas, f, 'Check_meas'))

# ExportPrintMixin ##########################################################

class ExportPrintMixin(object):
	"""Mixin do DocumentFrame, umoznuje reagovat na nabidky z menu tiskem/exportem

	Umi tisknout/exportovat ascii data mereni, html report a obrazky.
	Potrebuje od instance tridy do ktere se mixuje nasledujici atributy:

	frame: instanci wx.Window()
	plot_views: sekvence PlotView(), ktere se maji exportovat/tisknout
	model: instanci s nasledujicimi atributy
	 - protocol(): Cesta, kam se maji ukladat exportovane protokoly
	 - html_report() - html report
	 - get_protocol_title(): Titulek, ktery se bude psat nad protokoly a obrazky
	 - ascii_report(Bool export_raw, Bool export_meas) - ascii report

	 Inicializuje se jako ExportPrintMixin.bind_menu(self) ve tride do ktere se
	 vklada.
	"""

	def bind_menu(self):
		self.bind(wx.EVT_MENU, self.on_export_pdf, 'MI_ExportPDF'),
		self.bind(wx.EVT_MENU, self.on_export_img, 'MI_ExportImage'),
		self.bind(wx.EVT_MENU, self.on_export_ascii, 'MI_ExportAscii'),
		self.bind(wx.EVT_MENU, self.on_print_pdf, 'MI_PrintPDF'),
		self.bind(wx.EVT_MENU, self.on_print_img, 'MI_PrintImage'),
		self.bind(wx.EVT_MENU, self.on_print_ascii, 'MI_PrintAscii'),

	def _get_images_dpi(self):
		return registry.settings.value.images_dpi.value

	def _save_dlg(self, msg_tail, wildcard):
		return save_dialog(
				self.model.get_protocol_path(), 
				u"", 
				"Vyberte soubor do kterho se " + msg_tail, 
				wildcard, 
				self.frame)

	def export_pdf(self, fname = None, success_mesage = None):
		"""Exportuje PDF report.

		fname: Pokud je None, zepta se na jmeno, jinak exportuje do fname
		success_mesage: Co se ma ukazat uzivateli po skonceni exportu za
			hlasku. Pokud je None, neukaze se nic.
		vraci: True kdyz se vsecho povedlo, False pri chybe
		"""
		if not fname:
			path = self._save_dlg(u"m protokol uloit", registry.wildcards.pdf)
			if not path:
				return False
		else:
			path = fname

		progress = wx.ProgressDialog(u"Exportuji do PDF...", "_"*40, 
		                             maximum = len(self.plot_views) + 1, 
		                             parent = self.frame)
		
		image_fnames = []
		for idx, pv in enumerate(self.plot_views):
			progress.Update(0, u"Vytvm obrzek z grafu %i" % (idx+1))
			try:
				img_fname = misc.tempnam('.png')
				image_fnames.append(img_fname)
				pv.print_figure(img_fname, dpi = self._get_images_dpi())
			except EnvironmentError, e:
				log_environment_error(e)
				log_error(u"Chyba pi vytven obrzku")
				progress.Destroy()
				return False

		try:
			progress.Update(len(self.plot_views), u"Vytvm protokol ve formtu PDF")
			html_report = self.model.html_report()
			pdf.write_pdf_report(path, self.model.get_protocol_title(), html_report, image_fnames)
		finally:
			progress.Destroy()
			
		for i in image_fnames:
			misc.safe_remove(i)
		if success_mesage:
			msgbox_info(success_mesage)
		return True

	def export_img(self, plot, fname = None, success_mesage = None):
		"""Exportuje obrazek do souboru.

		fname: Pokud je None, zepta se na jmeno, jinak exportuje do fname
		success_mesage: Co se ma ukazat uzivateli po skonceni exportu
			za hlasku. Pokud je None, neukaze se nic.
		vraci: True kdyz se vsecho povedlo, False pri chybe
		"""
		if not fname:
			path = self._save_dlg(u"m obrzek uloit", registry.wildcards.pictures)
			if not path:
				return False
		else:
			path = fname

		try:
			plot.print_figure(path, dpi = self._get_images_dpi())
		except EnvironmentError, e:
			log_environment_error(e)
			log_error(u"Chyba pi exportu obrzku")
			return False

		if success_mesage:
			msgbox_info(success_mesage)
		return True
		
	def export_ascii(self, fname = None, success_mesage = None):
		"""Exportuje ascii data do souboru.

		fname: Pokud je None, zepta se na jmeno, jinak exportuje do fname
		success_mesage: Co se ma ukazat uzivateli po skonceni exportu
			za hlasku. Pokud je None, neukaze se nic.
		vraci: True kdyz se vsecho povedlo, False pri chybe
		"""
		export_data = ExportData()
		if wx.ID_OK != ExportDataDlg.run_dialog(export_data):
			return False

		if not fname:
			path = self._save_dlg(u"maj ASCII data men uloit", registry.wildcards.txt)
			if not path:
				return False
		else:
			path = fname

		report = self.model.ascii_report(export_data.raw.value, export_data.meas.value)

		try:
			f = open(path, 'w')
			f.write(report)
			f.close()
		except EnvironmentError, e:
			log_environment_error(e)
			log_error(u"Chyba pi ukldn dat do souboru " + path)
			return False

		if success_mesage:
			msgbox_info(success_mesage)
		return True

	def on_export_pdf(self, evt):
		self.export_pdf(success_mesage = u"Protokol byl spn exportovn.")

	def on_export_img(self, evt):
		for p in self.plot_views:
			self.export_img(p, success_mesage = u"Obrzek byl spn exportovn.")

	def on_export_ascii(self, evt):
		self.export_ascii(success_mesage = u"ASCII data byla spn exportovna.")

	def on_print_pdf(self, evt):
		busy = wx.BusyInfo(u"Moment prosm, pipravuji protokol k tisku...")
		tmp_name = misc.tempnam('.pdf')
		if not self.export_pdf(tmp_name):
			return
		pdf.pdf_print(tmp_name)

	def on_print_img(self, evt):
		for p in self.plot_views:
			busy = wx.BusyInfo(u"Moment prosm, pipravuji obrzek k tisku...")
			img_tmp_name = misc.tempnam('.png')
			if not self.export_img(p, img_tmp_name):
				return
			tmp_name = misc.tempnam('.pdf')
			pdf.write_pdf_image(tmp_name, self.model.get_protocol_title(), img_tmp_name)
			pdf.pdf_print(tmp_name)
			misc.safe_remove(img_tmp_name)

	def on_print_ascii(self, evt):
		ascii_tmp_name = misc.tempnam('.txt')
		if not self.export_ascii(ascii_tmp_name):
			return
		tmp_name = misc.tempnam('.pdf')
		pdf.write_pdf_ascii(tmp_name, self.model.get_protocol_title(), ascii_tmp_name)
		pdf.pdf_print(tmp_name)
		misc.safe_remove(ascii_tmp_name)

# funkce ####################################################################

def gui_viewdata(parent):
	ViewdataFrame(storage.MeasData()).run()

